#!/usr/bin/lua
local socket
local encapGRE
local decapGRE
local createGRE
local createTun
local tunnels
local int32ToStr
local strToInt32
local int16ToStr
local strToInt16
local sendKeepAlive
local getPath
local setPath

socket=require"socket"
nsg = require"nsg"

tunnels={
   max = 0,
   hash = {},
   KAs = {},
   KATime = 0,
}

local greConfig = {

}
controlCommands = {
   
}
sendKeepAlive = function(tun)
                   local tunCfg = tunnels[tun]
                   local gre = tunCfg.link
                   --local greCfg = tunnels[gre]
                   --io.write("KAE:"..tunCfg.KAError.."KAW:"..tostring(tunCfg.KAWait).."\n")
                   if tunCfg.KAWait then
                      tunCfg.KAError = tunCfg.KAError + 1
                      if (tunCfg.KAError > tunCfg.KARetry) and tunCfg.KAUP then
                         --io.write("keepalive!!! DOWN IFACE:"..tunCfg.name.."\n")
                         tunCfg.KAUP = false
                         io.write(tunCfg.name.." down\n")
                         os.execute("ifconfig "..tunCfg.name.." down")
                      end
                   end
                   local ret, err = gre:sendto(tunCfg.KAPacket, tunCfg.destination, 0)
                   if not ret then
                      io.write("err".."\n")
                   end
                   tunCfg.KAWait = true
                end

int16ToStr = function(int)
                return string.char(math.floor(int/2^8) % 256, int % 256)
             end

int32ToStr = function(int)
                return string.char(
                   math.floor(int/2^24) % 256,
                   math.floor(int/2^16) % 256,
                   math.floor(int/2^8) % 256,
                   int % 256)
             end

strToInt16 = function(str)
                b1,b2 = str:byte(1,2)
                return b2 + b1 * 2^8
             end
strToInt32 = function(str)
                b1,b2,b3,b4 = str:byte(1,4)
                return b4 + b3 * 2^8 + b2 * 2^16 + b1 * 2^24
             end

encapGRE = function (buf, tun)
              local res
              res = table.concat({tun.GREHeader.head, buf})

              --io.write("head crc:"..table.concat({tun.GREHeader.head:byte(1,#tun.GREHeader.head)}," ")..":"..(nsg.ipsum(tun.GREHeader.head, #tun.GREHeader.head)).."\n")

              if tun.checksum then
                 local crc = nsg.ipsum(res, #res)
                 res =  table.concat({
                                        tun.GREHeader.flagsVer,
                                        tun.GREHeader.proto,
                                        int16ToStr(crc),
                                        int16ToStr(0), --reserved
                                        tun.GREHeader.key,
                                        buf,
                                     })
                 --io.write("crc:"..(nsg.ipsum(res, #res)).."\n")
              end
              return res
           end

decapGRE = function (buf, size)
              if not buf or #buf < 24 then
                 io.write("Error receiving data, data is too short.")
                 return nil
              end
              local IPVersion = math.floor(buf:byte(1)/16)
              local IPHeaderLen
              local key = ""
              local checksum
              local encap = 0

              if IPVersion == 4 then
                 IPHeaderLen = 4 * (buf:byte(1) - IPVersion * 16)
              elseif IPVersion ==  6 then
                 IPHeaderLen = 40
              end
              local GREFlags = buf:byte(1 + IPHeaderLen)
              encap = strToInt16(buf:sub(1 + IPHeaderLen + 2, 1 + IPHeaderLen + 3))
              local GREHeaderLen = 4
              if math.floor(GREFlags/128) == 1 then  -- checksum
                 local grePacket = buf:sub(1 + IPHeaderLen)
                 if 0 ~=  nsg.ipsum(grePacket, #grePacket) then
                    io.write("Bad GRE checksum, drop packet")
                    return nil
                 end
                 checksum = strToInt16(buf:sub(1 + IPHeaderLen + GREHeaderLen ,3 + IPHeaderLen + GREHeaderLen))
                 GREHeaderLen = GREHeaderLen + 4
                 GREFlags = GREFlags - 128
              end
              if math.floor(GREFlags/32) == 1 then --key
                 key = strToInt32(buf:sub(1 + IPHeaderLen + GREHeaderLen, 5 + IPHeaderLen + GREHeaderLen ))
                 GREHeaderLen = GREHeaderLen + 4
                 GREFlags = GREFlags - 32
              end
              buf=buf:sub(1 + IPHeaderLen + GREHeaderLen) or ""
              return buf, encap, checksum , key
           end

createCTL = function (addr, port)
               local sockCTL, err = socket.udp()
               if err then
                  io.write(err)
               else
                  sockCTL:settimeout(0)
                  sockCTL:setsockname(addr or "127.0.0.1", port or 50009)
                  tunnels[sockCTL] = {
                     socket = sockCTL,
                     type = "CTL",
                     name = "CTL",
                     index = tunnels.max + 1,
                  }
                  tunnels.max = tunnels.max + 1
                  tunnels[#tunnels+1] = sockCTL
                  end
                  return sockCTL, err
               end

getPath = function (root, first, ...)
                return type(root)=="table" and first and getPath(root[first], ...) or 
             type(root)~="table" and root
             end

setPath = function (root, value, ...)
             local prevRoot
             local index
             for _, v in ipairs({...}) do
                prevRoot = root
                index = v
                root[v] = root[v] or {}
                root = root[v]
             end
             prevRoot[index] = value
          end

deleteGRE = function (source, device)
               device = device or ""
               source = source or "0.0.0.0"
               local sockGRE, err
               sockGRE = getPath(tunnels.hash, device, source, "GRE")
               if sockGRE then
                  tunnels[sockGRE].usage =  tunnels[sockGRE].usage - 1
                  if tunnels[sockGRE].usage == 0 then
                     setPath(tunnels.hash, nil, device, source, "GRE")
                     local GRECfg = tunnels[sockGRE]
                     tunnels[GRECfg.index] =  tunnels[#tunnels]
                     tunnels[tunnels[GRECfg.index]].index = GRECfg.index
                     tunnels[sockGRE] = nil
                     tunnels[#tunnels] = nil
                     sockGRE:close()
                     tunnels.max = tunnels.max - 1
                  end
               end
            end

createGRE = function (source, device)
               device = device or ""
               source = source or "0.0.0.0"
               local sockGRE, err
               sockGRE = getPath(tunnels.hash, device, source, "GRE")
               if sockGRE then
                  tunnels[sockGRE].usage =  tunnels[sockGRE].usage + 1
                  return sockGRE
               end
               sockGRE, err = socket.raw(47) --47 for gre
               if err then
                  return nil, err
               else
                  setPath(tunnels.hash, sockGRE, device, source, "GRE")
                  sockGRE:settimeout(0)
                  sockGRE:setsockname(source)
                  tunnels[sockGRE] = {
                     socket = sockGRE,
                     type = "GRE",
                     name = "GRE:"..source,
                     source = source,
                     device = device,
                     index = tunnels.max + 1,
                     usage = 1,
                  }
                  tunnels.max = tunnels.max + 1
                  tunnels[#tunnels+1] = sockGRE
               end
               return sockGRE
            end

deleteTun = function (device, source, destination, key)
               local sockTun = getPath(tunnels.hash, device, source, destination, key)
               setPath(tunnels.hash, nil, device, source, destination, key)
               if sockTun then
                  local TunCfg = tunnels[sockTun]
                  tunnels[TunCfg.index] =  tunnels[#tunnels]
                  tunnels[tunnels[TunCfg.index]].index = TunCfg.index
                  tunnels[sockTun] = nil
                  tunnels[#tunnels] = nil
                  sockTun:close()
                  tunnels.max = tunnels.max - 1
               end
            end

createTun = function (name, link, source, destination, encap, key , checksum, KAPeriod, KARetry)
               if  getPath(tunnels.hash, tunnels[link].device, tunnels[link].source, destination, key or "") then
                  return nil, "Tunnel with such config already exists."
               end
               local  tun, err = socket.tun(name, encap=="eth" and 1 or nil)
               if err then
                  return nil, err
               else
                  tun:settimeout(0)
                  tunnels[tun] = {
                     socket = tun,
                     type = "tun",
                     link = link,
                     name = name,
                     encap = encap=="eth" and 25944 or 2048,
                     destination = destination,
                     key = key,
                     checksum = (checksum=="yes") and true or false,
                     KAPeriod = KAPeriod,
                     KARetry = KARetry,
                     index = tunnels.max + 1,
                  }
                  local tunCfg = tunnels[tun]
                  local greCfg = tunnels[link]
                  tunCfg.GREHeader = {
                     flagsVer = string.char((tunCfg.checksum and 128 or 0) + (key and 32 or 0), 0),
                     proto = int16ToStr(encap=="eth" and 25944 or 2048) ,
                     checksum = tunCfg.checksum and int32ToStr(0) or "",
                     key = key and int32ToStr(tonumber(key)) or "",
                  }
                  
                  tunCfg.GREHeader.head = table.concat({
                            tunCfg.GREHeader.flagsVer,
                            tunCfg.GREHeader.proto,
                            tunCfg.GREHeader.checksum,
                            tunCfg.GREHeader.key,
                                                       })
                  tunCfg.GREHeader.headLen = #tunCfg.GREHeader.head

                  setPath(tunnels.hash, tun, greCfg.device, greCfg.source, destination, key or "" )
                  
                  tunnels.max = tunnels.max + 1
                  tunnels[#tunnels+1] =  tun
                  if KAPeriod then
                     tunnels.KAs[tun] = tun
                     tunCfg.KAUP = false
                     tunCfg.KAWait = false
                     tunCfg.KAError = 0
                     tunCfg.KALastTime = 0
                     tunCfg.KASource = strToInt32(string.char(socket.skip(2,(socket.dns.toip(destination) or destination ):find("(%d+)%.(%d+)%.(%d+)%.(%d+)"))))
                     tunCfg.KADestination = strToInt32(string.char(socket.skip(2,(socket.dns.toip(source) or source):find("(%d+)%.(%d+)%.(%d+)%.(%d+)"))))

--using of KEY in KA packet is non standard!!!
                     local KAGRE  =  string.char((key and 32 or 0),0,0,0)..(key and int32ToStr(key) or "")
                     local KAIPHead =  table.concat({"\69\0",--size ver + headLen
                                                    int16ToStr(20+#KAGRE), --total size
                                                    "\0\0", --ident 
                                                    "\0\0", --flags + frag
                                                    "\64\47", --ttl + proto 
                                                    "\0\0",  --check sum
                                                    int32ToStr(tunCfg.KASource),--source
                                                    int32ToStr(tunCfg.KADestination),--destination
                                                 })
                     tunCfg.KAPacket = table.concat({
                                                             string.char((key and 32 or 0),0), -- gre flags + ver 
                                                             int16ToStr(2048), -- encap proto IP
                                                             (key and int32ToStr(key) or ""), -- GRE key
                                                             "\69\0",--size ver + headLen
                                                             int16ToStr(20+#KAGRE), --total size
                                                             "\0\0", --ident
                                                             "\0\0", --flags + frag
                                                             "\64\47", --ttl + proto
                                                             int16ToStr(nsg.ipsum(KAIPHead, #KAIPHead)),  --check sum
                                                             int32ToStr(tunCfg.KASource),--source
                                                             int32ToStr(tunCfg.KADestination),--destination
                                                             KAGRE,})
                  else
                     tunCfg.KAUP = true
                     io.write  (tunCfg.name.." up\n")
                     os.execute("ifconfig "..tunCfg.name.." up")
                  end
               end
               return tun, err
            end

controlCommands.config = function()
                            return ser.serialize(greConfig).."\n"
                      end
controlCommands.addTunnel = function(name, cfg)
                               if not name or not cfg then
                                  return "Error: wrong argument list for addTunnel"
                               else
                                  if greConfig[name] then
                                     return "Error: tunnel "..name.." already exists"
                                  end
                                  if cfg.device and #cfg.device>0 and 0~=os.execute("ifconfig "..cfg.device.." 2>/dev/null | grep -o 'UP' >/dev/null") then
                                     return "Warn: no active master device for "..name
                                  end
                                  local gre, tun, err
                                  gre, err = createGRE(cfg.source, cfg.device)
                                  if gre then
                                     tun, err = createTun(name, gre, cfg.source or "0.0.0.0", cfg.destination or "0.0.0.0", cfg.encapsulation, cfg.key, cfg.checksum, cfg.keepalive=="yes" and cfg["keepalive-interval"] or nil,cfg["keepalive-retry"])
                                     if err then
                                        deleteGRE(cfg.source, cfg.device)
                                        return "Error:"..err
                                     else
                                        tunnels[gre].tunCount = (tunnels[gre].tunCount or 0) + 1
                                        greConfig[name] = cfg
                                        --return "OK"
                                     end
                                  else
                                     return"Error:"..err
                                  end
                               end
                               return "OK"
                            end

controlCommands.delTunnel = function(name)
                               local cfg = greConfig[name]
                               if not name or not cfg then
                                  return "Error: no such gre tunnel"
                               else
                                  greConfig[name] = nil
                                  deleteTun(cfg.device or "",
                                            cfg.source or "0.0.0.0",
                                            cfg.destination or "0.0.0.0",
                                            cfg.key or "")
                                  deleteGRE(cfg.source, cfg.device)
                               end
                               return "OK"
                            end

function main(conf, port)
   ser = require"serialize"
   local gre, tun, err
   createCTL("127.0.0.1", port)
   io.write("conf="..ser.serialize(conf).."\n")
   for k,v in pairs(conf) do
      if (tostring(k)):find("^gre%d+$") then
         io.write("Creating tunnel:"..k.."; ")
         err = controlCommands.addTunnel(k, v)
         io.write("result:"..err.."\n")
      end
   end
   local stats={rx=0, tx=0, rx10=0, tx10=0, prx = 0 , ptx = 0, prx10 = 0 , ptx10 = 0}
   local time
   local TA = 1
   time = socket.gettime()

   while true do
      local rcv,send,err
      --io.write(tostring(time).."\n")
      --      io.write("TA:"..TA.."\n")
      rcv, send, err= socket.select(tunnels, {} , TA)
      current = socket.gettime()
      if current > tunnels.KATime then
         local next = 60
         local tunCfg
         for k,v in pairs(tunnels.KAs) do
            local diff
            tunCfg = tunnels[v]
            if current - tunCfg.KALastTime > tunCfg.KAPeriod then
               sendKeepAlive(v)
               tunCfg.KALastTime = current
            end
            diff = tunCfg.KAPeriod - (current - tunCfg.KALastTime)
            next = (next > diff) and diff or next
         end
         TA = next
         tunnels.KATime = current + next
      end
      --[[
      if current - time > 10 then
         io.write("Tx:\t"..(stats.tx).." B\t"..(stats.ptx).." P\t"..((stats.tx-stats.tx10)/10).." B/s\t"..((stats.ptx-stats.ptx10)/10).." P/s\n")
         io.write("Rx:\t"..(stats.rx).." B\t"..(stats.prx).." P\t"..((stats.rx-stats.rx10)/10).." B/s\t"..((stats.prx-stats.prx10)/10).." P/s\n")

         stats.tx10 = stats.tx
         stats.rx10 = stats.rx
         stats.ptx10 = stats.ptx
         stats.prx10 = stats.prx

         time = current
      end
      --]]      local sockCfg
      local sock
      local readTime = 1 / (#rcv + 1) -- we want read all select-ed threads for no more then 1 second 
      for k,v in ipairs(rcv) do 
         local threadReadStart = socket.gettime()
         local threadReadCount = 0
         sock = v
         sockCfg = tunnels[v]
         --       io.write(sockCfg.name)
         if sockCfg.type == "tun" then
            local buf, ip
            local tun = sock
            local tunCfg = sockCfg
            local gre = tunCfg.link
            local greCfg = tunnels[gre]
            ip = greCfg.source
            repeat
               buf = tun:receive()
               threadReadCount = threadReadCount + 1
               if (buf and #buf or 0) > 0 then
                  buf = encapGRE(buf, tunCfg)
                  local ret, err = gre:sendto(buf, tunCfg.destination, 0)
                  if not ret then
                     io.write("err".."\n")
                  end
                  --            io.write(":\t"..#buf.."  "..ip..">"..sockCfg.destination..":"..table.concat({buf:byte(1,10)}).."\n")
                  stats.rx = stats.rx + #buf
                  stats.prx = stats.prx + 1
               end
            until ((buf and #buf or 0) <= 0) or ((socket.gettime()-readTime) > threadReadStart)
         elseif sockCfg.type == "GRE" then
            local buf, encap, ip, key, checksum
            local gre = sock
            local greCfg = sockCfg
            local tun
            local tunCfg
            repeat
               buf, ip = v:receivefrom()
               if (buf and #buf or 0) > 0 then
                  buf, encap, checksum, key = decapGRE(buf)
                  tun = tunnels.hash[greCfg.device]
                  tun = tun and tun[greCfg.source]
                  tun = tun and tun[ip] and tun[ip][key]
                  if tun then
                     tunCfg = tunnels[tun]
                     if encap == 0 then
                        --keepalive
                        tunCfg.KAWait = false
                        tunCfg.KAError = 0
                        --io.write("keepalive received\n")
                        if  not tunCfg.KAUP then
                           tunCfg.KAUP = true
                           io.write(tunCfg.name.." up\n")
                           os.execute("ifconfig "..tunCfg.name.." up")
                           --io.write("KAE:"..tunCfg.KAError.."KAW:"..tostring(tunCfg.KAWait).."\n")
                        end
                     elseif encap == 2048 and buf:byte(10) == 47 and buf:byte(23) == 0 and buf:byte(24) == 0
                     then
                        --io.write("sending keepAlive ECHO\n")
                        buf = encapGRE(decapGRE(buf), tunCfg)
                        local ret,err = gre:sendto(buf, tunCfg.destination, 0)
                        if not ret then
                           io.write("err".."\n")
                        end
                        --tunnels.KAGRE:send(buf)
                     elseif  encap == tunCfg.encap then
                        local ret,err = tun:send(buf)
                        if not ret then
                           io.write("err".."\n")
                        end
                        stats.tx = stats.tx + #buf
                        stats.ptx = stats.ptx + 1
                        --    io.write("HASH found")
                     else
                        io.write("Wrong encapsulation:"..encap.."\n")
                     end
                  else
                     io.write("Tunnel not found for :"..ip..">"..greCfg.source..":"..tostring(key).."\n")
                  end
                  --                  io.write(":\t"..#buf.."  "..ip..">"..sockCfg.source..":"..table.concat({buf:byte(1,10)}).."\n")
               end
            until ((buf and #buf or 0) <= 0) or ((socket.gettime()-readTime) > threadReadStart)
            threadReadStart = socket.gettime()
            threadReadCount = 0
         elseif sockCfg.type == "CTL"  then
            local CTL = sock
            local CTLCfg = sockCfg
            local f, res, err
            local mes, addr, port = CTL:receivefrom()
            if mes then
               if addr == "127.0.0.1" and (
                  mes:find("^config%(%)$") or
                  mes:find("^delTunnel%(['\"]gre%d+['\"]%)$") or
               mes:find("^addTunnel%(['\"]gre%d+['\"][,]{[%[%]-a-zA-Z=0-9\"'_,.{}]+}%)$")
         )
               then
                  f, err = loadstring("return controlCommands."..mes)
                  if f then
                     res, err = f()
                  else
                     err = "Error in command"
                  end
               else
                  err = "Access denied"
               end
               CTL:sendto( tostring( res or err ) , addr, port)
            end
         end
      end
   end
end
io.write("Starting GRE daemon\n")
main(loadfile(arg[1])(), tonumber(arg[2]) or 50009)

