module(..., package.seeall)
local s = require'scheme'
local usb = require'nsgusbconf'

local oneWireDevName = usb.port2tty["1-wire"]

function setDevName()
   oneWireDevName = usb.port2tty["1-wire"]
end

--[[
local arch2oneWireDevName = {
  nsg600 = "/dev/ttyAM1",
  nsg700 = "/dev/nsg/ow1"
}
local oneWireDevName = arch2oneWireDevName[BI.arch]
]]--

local function getDevID(name)
  return name:match"(%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x)$"
end

local function getDevType(name)
  return name:match"(%x%x)%x%x%x%x%x%x%x%x%x%x%x%x%x%x$"
end

local devtype2name = {
  ["10"] = 'temp',
  ["29"] = 'swt8',
  ["3A"] = 'swt2'
}
setmetatable(devtype2name, {__index = function() return 'ow' end})

local function oneWireAutoconfig(path, sch, val)
  path = s.clonePath(path, -1)
  path[#path+1] = 'device'
  if val == 'init' or not C.getConfByPath(path) then
    C.setConfigValue(path, {})
  end
  local devConf = C.getConfByPath(path)
  if not devConf then
    devConf = {}
    C.setConfigValue(path, devConf)
  end
  local str = s.readpopen("nsgow -i "..oneWireDevName)
  local devs = {}
  for id in str:gmatch'(%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x)%s+' do
    devs[id] = true
  end
  if val == 'update' then
    for k, _ in pairs(devConf) do
      if not devs[getDevID(k)] then
        devConf[k] = nil
      end
    end
  end
  local devConfID = {}
  for k, _ in pairs(devConf) do
    devConfID[getDevID(k)] = true
  end
  for k, _ in pairs(devs) do
    if not devConfID[k] then
      devConf[devtype2name[getDevType(k)]..'-'..k] = {_keep=true}
    end
  end
  return "Find 1-Wire devices:\n"..str
end

local function getCircuitState(oneWireDevName, devID)
  local str = s.readpopen(("nsgow %s -d %s -k 2>&1"):format(
  oneWireDevName, devID))
  local circuitState = {}
  for d in str:gmatch'[0,1]' do
    circuitState[#circuitState+1] = d
  end
  return circuitState
end

local function getCircuitStateName(state, inverse)
  return state == '1' and (inverse and 'OFF' or 'ON') 
  or (state == '0' and (inverse and 'ON' or 'OFF') or 'UNKNOWN')
end

local circuitScheme = s.newScheme {
  _H = "oneWireDeviceCircuit",
  on = s.newScheme {
    _H = "oneWireDeviceCircuitOn",
    --_T = 'areyousure',
    --_V = areyousureVal,
    --_A = {"Sircuit will switch ON.\n"},
    _X = function(path, sch, val)
      local inverse = C.getConfByPath(s.clonePath(path, -3, "inverse"))
      s.readpopen(("nsgow %s -d %s -%d %s -k 2>&1"):format(
          oneWireDevName, 
          getDevID(path[#path - 3]), 
          path[#path - 1] - 1,
          inverse and 'open' or 'short'
        )
      )
    end,
    _R = 'user',
    _P = 50,
  },
  off = s.newScheme {
    _H = "oneWireDeviceCircuitOff",
    --_T = 'areyousure',
    --_V = areyousureVal,
    --_A = {"Sircuit will switch OFF.\n"},
    _X = function(path, sch, val)
      local inverse = C.getConfByPath(s.clonePath(path, -3, "inverse"))
      s.readpopen(("nsgow %s -d %s -%d %s -k 2>&1"):format(
          oneWireDevName, 
          getDevID(path[#path - 3]), 
          path[#path - 1] - 1,
          inverse and 'short' or 'open'
        )
      )
    end,
    _R = 'user',
    _P = 40,
  },
  pulse = s.newScheme {
    _H = "oneWireDeviceCircuitPulse",
    _X = function(path, sch, val, sesNum)
      local inverse = C.getConfByPath(s.clonePath(path, -3, "inverse"))
      s.readpopen(("nsgow %s -d %s -%d %s -k -r %d 2>&1"):format(
          oneWireDevName, 
          getDevID(path[#path - 3]), 
          path[#path - 1] - 1,
          inverse and 'drop' or 'pulse',
          C.getConfByPath(s.clonePath(path, -1, "reset-delay"), sesNum) or
          1000
        )
      )
    end,
    _R = 'user',
    _P = 20,
  },
  drop = s.newScheme {
    _H = "oneWireDeviceCircuitDrop",
    _X = function(path, sch, val, sesNum)
      local inverse = C.getConfByPath(s.clonePath(path, -3, "inverse"))
      s.readpopen(("nsgow %s -d %s -%d %s -k -r %d 2>&1"):format(
          oneWireDevName, 
          getDevID(path[#path - 3]), 
          path[#path - 1] - 1,
          inverse and 'pulse' or 'drop',
          C.getConfByPath(s.clonePath(path, -1, "reset-delay"), sesNum) or
          1000
        )
      )
    end,
    _R = 'user',
    _P = 10,
  },
  toggle = s.newScheme {
    _H = "oneWireDeviceCircuitToggle",
    _X = function(path, sch, val)
      s.readpopen(("nsgow %s -d %s -%d toggle -k 2>&1"):format(
          oneWireDevName, getDevID(path[#path - 3]), path[#path - 1] - 1))
    end,
    _R = 'user',
    _P = 30,
  },
  ["reset-delay"] = s.integerType {
    _H = "oneWireDeviceCircuitResetDelay",
    _A = {0, 10000},
    _D = 1000,
    _LOCAL = true,
    _R = 'user',
  },
  show = s.newScheme{
    _H = "oneWireShow",
    _X = function(path)
      local circuitState = getCircuitState(oneWireDevName, getDevID(path[#path - 3]))
      local state = circuitState[path[#path - 1]]
      local inverse = C.getConfByPath(s.clonePath(path, -3, "inverse"))
      return "Circuit "..getCircuitStateName(state, inverse)
    end,
    _R = 'user',
    _P = -10,
  },
}

local ow29Scheme = s.newScheme{
  _H = "oneWireDeviceName",
  ["type"] = s.newScheme{
    _H = "oneWireDeviceType",
    _T = "choice",
    _A = {
      {"spc8", ""},
    },
    _V = s.choiceVal,
    _DEP = 1,
  },
  show = s.newScheme{
    _H = "oneWireShow",
    _X = function(path)
      local circuitState = getCircuitState(oneWireDevName, getDevID(path[#path - 1]))
      local out = {}
      local inverse = C.getConfByPath(s.clonePath(path, -1, "inverse"))
      for k, v in ipairs(circuitState) do
        out[#out+1] = "Circuit "..tostring(k)..' '..getCircuitStateName(v, inverse)
      end
      return table.concat(out, '\n')
    end,
    _R = 'user',
  },
  inverse = s.booleanType{
    _H = "oneWireDeviceInverse",
  },
  circuit = s.newScheme{
    _H = "oneWireDeviceCircuits",
    [1] = circuitScheme,
    [2] = circuitScheme,
    [3] = circuitScheme,
    [4] = circuitScheme,
    [5] = circuitScheme,
    [6] = circuitScheme,
    [7] = circuitScheme,
    [8] = circuitScheme,
  }
}

local function isIOdevice(path)
  return C.getConfByPath(s.clonePath(path, -3, "type")) ~= "IC-2di-220"
end
local circuitScheme3A = s.newScheme { -- 3A
  _H = "oneWireDeviceCircuit",
  on = function(_, path)
    return isIOdevice(path) and
    s.newScheme {
      _H = "oneWireDeviceCircuitOn",
      _X = function(path, sch, val)
        s.readpopen(("nsgow %s -d %s:b -%s short -k 2>&1"):format(
            oneWireDevName, 
            getDevID(path[#path - 3]), 
            path[#path - 1]
          )
        )
      end,
      _R = 'user',
      _P = 50,
    }
  end,
  off = function(_, path)
    return isIOdevice(path) and
    s.newScheme {
      _H = "oneWireDeviceCircuitOff",
      _X = function(path, sch, val)
        s.readpopen(("nsgow %s -d %s:b -%s open -k 2>&1"):format(
            oneWireDevName, 
            getDevID(path[#path - 3]), 
            path[#path - 1]
          )
        )
      end,
      _R = 'user',
      _P = 40,
    }
  end,
  pulse = function(_, path)
    return isIOdevice(path) and
    s.newScheme {
      _H = "oneWireDeviceCircuitPulse",
      _X = function(path, sch, val, sesNum)
        s.readpopen(("nsgow %s -d %s:b -%s pulse -k -r %d 2>&1"):format(
            oneWireDevName, 
            getDevID(path[#path - 3]), 
            path[#path - 1],
            C.getConfByPath(s.clonePath(path, -1, "reset-delay"), sesNum) or
            1000
          )
        )
      end,
      _R = 'user',
      _P = 20,
    }
  end,
  drop = function(_, path)
    return isIOdevice(path) and
    s.newScheme {
      _H = "oneWireDeviceCircuitDrop",
      _X = function(path, sch, val, sesNum)
        s.readpopen(("nsgow %s -d %s:b -%s drop -k -r %d 2>&1"):format(
            oneWireDevName, 
            getDevID(path[#path - 3]), 
            path[#path - 1],
            C.getConfByPath(s.clonePath(path, -1, "reset-delay"), sesNum) or
            1000
          )
        )
      end,
      _R = 'user',
      _P = 10,
    }
  end,
  toggle = function(_, path)
    return isIOdevice(path) and
    s.newScheme {
      _H = "oneWireDeviceCircuitToggle",
      _X = function(path, sch, val)
        s.readpopen(("nsgow %s -d %s:b -%s toggle -k 2>&1"):format(
            oneWireDevName, getDevID(path[#path - 3]), path[#path - 1]))
      end,
      _R = 'user',
      _P = 30,
    }
  end,
  ["reset-delay"] = function(_, path)
    return isIOdevice(path) and
    s.integerType {
      _H = "oneWireDeviceCircuitResetDelay",
      _A = {0, 10000},
      _D = 1000,
      _LOCAL = true,
      _R = 'user',
      _P = -9,
    }
  end,
  show = s.newScheme{
    _H = "oneWireShow",
    _X = function(path)
      local circuitState = getCircuitState(oneWireDevName, getDevID(path[#path - 3]))
      local circuit = path[#path - 1]
      local state = circuitState[circuit == 'A' and 1 or 2]
      return "Circuit "..circuit.." "..(state == '0' and 'open' or 'short')..'\n'
    end,
    _R = 'user',
    _P = -10,
  },
}

local ow3AScheme = s.newScheme{
  _H = "oneWireDeviceName",
  ["type"] = s.newScheme{
    _H = "oneWireDeviceType",
    _T = "choice",
    _A = {
      {"IC-2do-220/0,12", ""},
      {"IC-2do-220/1", ""},
      {"IC-2di-220", ""},
      {"IC-2dio", ""},
    },
    _V = s.choiceVal,
    _D = "IC-2dio",
    _DEP = 1,
  },
  show = s.newScheme{
    _H = "oneWireShow",
    _X = function(path)
      local circuitState = getCircuitState(oneWireDevName, getDevID(path[#path - 1]))
      return (
        "Circuit A "..(circuitState[1] == '0' and 'open' or 'short')..'\n'..
        "Circuit B "..(circuitState[2] == '0' and 'open' or 'short')..'\n'
      )
    end,
    _R = 'user',
  },
  circuit = s.newScheme{
    _H = "oneWireDeviceCircuits",
    ['A'] = circuitScheme3A,
    ['B'] = circuitScheme3A,
  }
}

local ow10Scheme = s.newScheme{
  _H = "oneWireDeviceName",
  ["type"] = s.newScheme{
    _H = "oneWireDeviceType",
    _T = "choice",
    _A = {
      {"temp", ""},
    },
    _V = s.choiceVal,
  },
  show = s.newScheme{
    _H = "oneWireShow",
    _X = function(path)
      local str = s.readpopen(("nsgow %s -d %s -k 2>&1"):format(
          oneWireDevName, getDevID(path[#path - 1])))
      str = str:match"^%s*(.-)%s*$"
      return ("t = %s C"):format(str)
    end,
    _R = 'user',
  },
}

local devtype2scheme = {
  ["10"] = ow10Scheme,
  ["29"] = ow29Scheme,
  ["3A"] = ow3AScheme,
}
setmetatable(devtype2scheme, {__index = function() return s.newScheme{} end})
local function get1wireDeviceScheme(_, path)
  return devtype2scheme[getDevType(path[#path])]
end

local _oneWireScheme = s.newScheme {
  _H = "oneWirePort",
  
  autoconfig = s.newScheme{
    _H = "oneWireAutoconfig",
    _T = "choice",
    _A = {
      {"add", "Add new devices to configuration"},
      {"init", "Set configuration with default value"},
      {"update", "Add new devices and remove nonexistent ones"},
    },
    _D = "update",
    _V = s.choiceVal,
    _X = oneWireAutoconfig,
    _DEP = 1,
  },
  
  device = s.newScheme{
    _H = "oneWireDevice",
    _NAME = get1wireDeviceScheme,
  },
  
  show = s.newScheme{
    _H = "oneWireShow",
    _X = function()
      return s.readpopen("nsgow -s "..oneWireDevName)
    end,
    _R = 'user',
  },
}

oneWireScheme = oneWireDevName and _oneWireScheme
