module(..., package.seeall)
local s = require'scheme'

local function createBirdConfFile(conf)
   local function addConf(table, prefix, value)
      if value ~= nil then
         table[#table + 1] = prefix..tostring(value)..";"
      end
   end

   local function confStr(str, def)
      return str and ("%q"):format(tostring(str)) or def
   end

   local function indentConf(confTable)
      local indent = 0
      for i,s in ipairs(confTable) do
         if s:find("}") then indent = indent - 1 end
         confTable[i] = ("\t"):rep(indent) .. confTable[i]
         if s:find("{") then indent = indent + 1 end
      end
      return confTable
   end

   local birdConf = {
      "# NSG Bird config",
      "# This file was generated automatically by nsgconfd"
   }

   local function genDebug(conf)
      if conf and conf.debug and conf.debug.protocols then
         if conf.debug.protocols.all then
            return "all;"
         else
            local protoBase = {}
            local protocols = {states=1,routes=1,filters=1,interfaces=1,events=1,packets=1}
            for k,v in  pairs(conf.debug.protocols) do
               if v then
                  protoBase[#protoBase+1] = protocols[k] and k
               end
            end
            if #protoBase>0 then
               return "{"..table.concat(protoBase,",").."};"
            end
         end
      end
      return "off;"
   end

   local function genDefProto(conf)
      if conf then
         birdConf[#birdConf+1] = "debug "..genDebug(conf)
         birdConf[#birdConf+1] = conf.extra or ""
         addConf(birdConf,"disabled ",conf.disabled and "on" or "off")
         addConf(birdConf,"preference ",conf.preference)
         addConf(birdConf,"description ",conf.description and ("%q"):format(conf.description))
         addConf(birdConf,"import ",conf.import)
         addConf(birdConf,"export ",conf.export)
         addConf(birdConf,"table ",conf.table)
      end
   end
   local debugCmd = {off="0",medium="1",all="2"}
   if conf.debug and conf.debug.commands then
      birdConf[#birdConf+1] = "debug commands "..debugCmd[conf.debug.commands]..";"
   end
   birdConf[#birdConf+1] = "debug protocols "..genDebug(conf)
   birdConf[#birdConf+1] = conf.extra or ""
   
   if conf["router-id"] then
      addConf(birdConf,"router id ",conf["router-id"])
   end
   
   if conf.table then
      for k,v in ipairs(conf.table) do
         addConf(birdConf,"table ", v )
      end
   end
   
   if conf.rip then
      for k,v in ipairs(conf.rip) do
         birdConf[#birdConf+1] = "protocol rip rip"..tostring(k).." {"
         genDefProto(v)
         addConf(birdConf,"authentication ",v.authentication)
         addConf(birdConf,"honor ",v.honor)
         addConf(birdConf,"port ",v.port)
         addConf(birdConf,"infinity ",v.infinity)
         addConf(birdConf,"period ",v.period)
         addConf(birdConf,"timeout time ",v["timeout-time"])
         addConf(birdConf,"garbage time ",v["garbage-time"])
         for n,c in pairs(v.interface or {}) do
            birdConf[#birdConf+1] = "interface "..(confStr(n, '""')).." {"
            addConf(birdConf,"mode ",c.mode)
            addConf(birdConf,"metric ",c.metric)
            birdConf[#birdConf+1] = "}; #interface "..tostring(n)
         end
         birdConf[#birdConf+1] = "} #protocol rip rip"..tostring(k).."\n"
      end
   end
   if conf.bgp then
      for k,v in ipairs(conf.bgp) do
         birdConf[#birdConf+1] = "protocol bgp bgp"..k.." {"
         genDefProto(v)
         addConf(birdConf,"local as ",v["local-as"])
         birdConf[#birdConf+1] = "neighbor "..v["neighbor-ip"].." as "..v["neighbor-as"]..";"
         if v.multihop.enable == "on" and v.multihop.number and v.multihop.via then   birdConf[#birdConf+1] = "multihop "..v.multihop.number.." via "..v.multihop.via..";" end
         addConf(birdConf,"next hop self ",v["next-hop-self"] and ";")
         addConf(birdConf,"missing lladdr ",v["missing-lladdr"])
         addConf(birdConf,"source address ",v["source-address"])
         addConf(birdConf,"password ",confStr(v["password"]))
         addConf(birdConf,"passive ",v["passive"])
         addConf(birdConf,"rr client ",v["rr-client"]and ";")
         addConf(birdConf,"rr cluster id ",v["rr-cluster-id"])
         addConf(birdConf,"rs client ",v["rs-client"] and ";")
         addConf(birdConf,"enable route refresh ",v["enable-route-refresh"])
         addConf(birdConf,"interpret communities ",v["interpret-communities"])
         addConf(birdConf,"enable as4 ",v["enable-as4"])
         addConf(birdConf,"capabilities ",v["capabilities"])
         addConf(birdConf,"advertise ipv4 ",v["advertise-ipv4"])
         addConf(birdConf,"route limit ",v["route-limit"])
         addConf(birdConf,"disable after error ",v["disable-after-error"])
         addConf(birdConf,"hold time ",v["hold-time"])
         addConf(birdConf,"startup hold time ",v["startup-hold-time"])
         addConf(birdConf,"keepalive time ",v["keepalive-time"])
         addConf(birdConf,"connect retry time ",v["connect-retry-time"])
         addConf(birdConf,"start delay time ",v["start-delay-time"])
         birdConf[#birdConf+1] = "error wait time "..v["error-wait-time-min"]..","..v["error-wait-time-max"]..";"
         
         addConf(birdConf,"error forget time ",v["error-forget-time"])
         addConf(birdConf,"path metric ",v["path-metric"])
         addConf(birdConf,"prefer older ",v["prefer-older"])
         addConf(birdConf,"default bgp_med ",v["default-bgp_med"])
         addConf(birdConf,"default bgp_local_pref ",v["default-bgp_local_pref"])
         birdConf[#birdConf+1] = "} #protocol bgp bgp "..k.."\n"
      end
   end
   if conf.ospf then
      for k,v in ipairs(conf.ospf) do
         birdConf[#birdConf+1] = "protocol ospf ospf"..k.." {"
         genDefProto(v)
         addConf(birdConf,"rfc1583compat ",v.rfc1583compat and "on" or "off")
         addConf(birdConf,"tick ",v.tick)
         for areaid,areadata in pairs(v.area or {}) do
            birdConf[#birdConf+1] = "area "..areaid.." {"
            addConf(birdConf,"stub ", areadata.stub and "yes" or "no")
            addConf(birdConf,"stub cost ", areadata["stub-cost"])
            birdConf[#birdConf+1] = "networks  {"
            for network,netdata in pairs(areadata.networks or {}) do
               birdConf[#birdConf+1] = network..(netdata.hidden and " hidden;" or ";")
            end
            birdConf[#birdConf+1] = "}; #networks"
            for stubnet, stubdata in pairs(areadata.stubnet or {}) do
               birdConf[#birdConf+1] = "stubnet "..stubnet.." {"
               addConf(birdConf,"hidden ",stubdata.hidden and "on" or "off")
               addConf(birdConf,"summary ",stubdata.summary and "on" or "off")
               addConf(birdConf,"cost ",stubdata.cost)
               birdConf[#birdConf+1] = "}; #stubnet "..stubnet
            end
            for iface, ifdata in ipairs(areadata.interface or {}) do
               birdConf[#birdConf+1] = "interface "..confStr(ifdata.name, '""').." {"
               addConf(birdConf,"cost ",ifdata.cost)
               addConf(birdConf,"stub ",ifdata.stub and "yes" or "no")
               addConf(birdConf,"hello ",ifdata.hello)
               addConf(birdConf,"poll ",ifdata.poll)
               addConf(birdConf,"retransmit ",ifdata.retransmit)
               addConf(birdConf,"priority ",ifdata.priority)
               addConf(birdConf,"wait ",ifdata.wait)
               addConf(birdConf,"dead count ",ifdata["dead-count"])
               addConf(birdConf,"dead ",ifdata.dead)
               addConf(birdConf,"rx buffer ",ifdata["rx-buffer"])
               addConf(birdConf,"type ",ifdata.type)
               addConf(birdConf,"strict nonbroadcast ",ifdata["strict-nonbroadcast"] and "on" or "off")
               addConf(birdConf,"authentication ",ifdata.authentication)
               addConf(birdConf,"password ",confStr(ifdata["password"]))
               birdConf[#birdConf+1] = "neighbors  {"
               for neigh,neighdata in pairs(ifdata.neighbors or {}) do
                  birdConf[#birdConf+1] = neigh..(neighdata.eligible and " eligible;" or ";")
               end
               birdConf[#birdConf+1] = "}; #neigborhs"
               birdConf[#birdConf+1] = "}; #interface "..ifdata.name
            end
            for vil, vildata in pairs(areadata["virtual-link"] or {}) do
                birdConf[#birdConf+1] = "virtual link "..vil.." {"
                addConf(birdConf,"hello ",vildata.hello)
                addConf(birdConf,"retransmit ",vildata.retransmit)
                addConf(birdConf,"wait ",vildata.wait)
                addConf(birdConf,"dead count ",vildata["dead-count"])
                addConf(birdConf,"dead ",vildata.dead)
                addConf(birdConf,"authentication ",vildata.authentication)
                addConf(birdConf,"password ",confStr(vildata.password))
                birdConf[#birdConf+1] = "}; #virt link "..vil
            end
            birdConf[#birdConf+1] = "}; #area id "..areaid
         end
         birdConf[#birdConf+1] = "}; #proto ospf "..k.."\n"
      end
   end
   if conf.device then
      birdConf[#birdConf+1] = "protocol device {"
      genDefProto(conf.device)
      addConf(birdConf,"scan time ",conf.device["scan-time"])
      for n,c in pairs(conf.device.primary or {}) do
         birdConf[#birdConf+1] = "primary "..confStr(c.interface, '""').." "..n..";"
      end
      birdConf[#birdConf+1] = "} #device\n"
   end

   if conf.static then
      for k,v in ipairs(conf.static) do
         birdConf[#birdConf+1] = "protocol static static"..tostring(k).." {"
         genDefProto(v)
         for n,c in pairs(v.route or {}) do
            if c.action == "via-interface" and c.interface then
               birdConf[#birdConf+1] = "route "..n.." via "..confStr(c.interface,'""')..";"
            elseif c.action == "via-gateway" and c.gateway then
               birdConf[#birdConf+1] = "route "..n.." via "..c.gateway..";"
            elseif c.action then
               birdConf[#birdConf+1] = "route "..n.." "..c.action..";"
            end
         end
         birdConf[#birdConf+1] = "}#protocol static"..tostring(k).."\n"
      end
   end

   if conf.kernel then
      for k,v in ipairs(conf.kernel) do
         birdConf[#birdConf+1] = "protocol kernel kernel"..k.." {"
         genDefProto(v)
         addConf(birdConf,"scan time ",v["scan-time"])
         addConf(birdConf,"learn ",v["learn"])
         addConf(birdConf,"kernel table ",v["kernel-table"])
         addConf(birdConf,"device routes ",v["device-routes"])
         birdConf[#birdConf+1] = "}#protocol kernel"..k.."\n"
      end
   end
   if conf.direct then
      for k,v in ipairs(conf.direct) do
         birdConf[#birdConf+1] = "protocol direct direct"..k.." {"
         genDefProto(v)
         local iface = ""
         local d=""
         for n,p in ipairs(v.interface) do
            iface = iface .. d..confStr(p, '""')
            d = ","
         end
         addConf(birdConf,"interface ", iface)
         birdConf[#birdConf+1] = "}# direct"..k.."\n"
      end
   end
   if conf.pipe then
      for k,v in ipairs(conf.pipe) do
         birdConf[#birdConf+1] = "protocol pipe pipe"..k.." {"
         genDefProto(v)
         addConf(birdConf,"mode ", v.mode)
         addConf(birdConf,"peer table ", v["peer-table"])
         birdConf[#birdConf+1] = "} #protocol pipe"..k.."\n"
      end
   end

   birdConf[#birdConf+1] = "\n"
   local f = io.open("/etc/bird.conf", 'w')
   if f then
      f:write(table.concat(indentConf(birdConf),"\n"))
      f:close()
   end
end

local function restartBird(path, sch, conf, prevConf)
   if not s.confCmp(conf, prevConf) then
      if conf and conf.enable then
         local conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
         createBirdConfFile(conf)
         if conf.enable ~= (prevConf and prevConf.enable) then
            os.execute("/usr/bin/killall bird")
            C.restartDaemon(
               path,
               {"/usr/sbin/bird", "-d", "-c", "/etc/bird.conf"}, 
               {noRestart = false, atExit = C.increasingSleep}
            )
         else
            os.execute("/usr/bin/killall -SIGHUP bird")
         end
         local msg = s.readpopen("/usr/sbin/bird -p -c /etc/bird.conf 2>&1")
         if #msg > 1 then
            return false, msg
         else
            return true
         end
      else
         C.restartDaemon(path)
      end
   end
   return true
end

local debugScheme = s.newScheme{ 
   _H = "birdDebug",
   protocols = s.newScheme{
      _H = "birdDebugProtocols",
      _N = function (sch, conf)
              return {
                 [1] = conf.all or nil,
                 all = conf.all or nil,
              }
           end,
      all = s.booleanType{
         _H = "birdDebugProtocolsAll",
         _DEP = 1,
      },
      states = s.booleanType{
         _H = "birdDebugProtocolsStates",
      },
      routes = s.booleanType{
         _H = "birdDebugProtocolsRoutes",
      },
      filters = s.booleanType{
         _H = "birdDebugProtocolsFilters",
      },
      interfaces = s.booleanType{
         _H = "birdDebugProtocolsInterfaces",
      },
      events = s.booleanType{
         _H = "birdDebugProtocolsEvents",
      },
      packets = s.booleanType{
         _H = "birdDebugProtocolsPackets",
      },
   },
}

local protcolScheme = s.newScheme{
   debug = debugScheme{},
   preference = s.integerType{
      _H = "birdProtocolPreference",
      _A = {0, 65535},
   },
   disabled = s.booleanType{_H = "birdProtocolDisabled", _P = 95},
   ["router-id"] = s.ipAddressType{_H = "birdProtocolRouterid"},
   description = s.stringType{_H = "birdProtocolDescription", _P = 100},
   import = s.stringType{_H = "birdProtocolImport"},
   export = s.stringType{_H = "birdProtocolExport"},
   table = s.stringType{_H = "birdProtocolTable"},
   interface = s.stringType{_H = "birdProtocolInterface"},
   extra = s.stringType{_H = "birdProtocolExtra"},
}

birdScheme = s.newScheme{
   _H = "bird",
   _Z = restartBird,
   enable = s.booleanType{ _P = 100, _D = false, _H = "birdEnable" },
   show = s.newScheme{
      _H = "birdShow",
      config = s.newScheme{
         _H = "birdShowConfig",
         _X = function()
                 local cfg = s.readpopen("cat /etc/bird.conf 2>&1")
                 local out = {}
                 local i = 0
                 for s in cfg:gmatch("[^\n]*\n") do
                    i = i + 1
                    out[#out +1] = ("%.4d: "):format(i)..s
                 end
                 return('\n'..table.concat(out))
              end,
         _R = 'user',
      },
      log =  s.newScheme{
         _H = "birdLog",
         _X = function(path)
                 path = s.clonePath(path, -2)
                 return C.getDaemonLog(path)
              end,
         _R = 'user',
      },
      interfaces = s.newScheme{
         _H = "birdShowInterfaces",
         _X = function(path)
                 local cfg = s.readpopen("echo show interfaces | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ospf = s.newScheme{
         _H = "birdShowOspf",
         _X = function(path)
                 local cfg = s.readpopen("echo show ospf | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ["ospf-interface"] = s.newScheme{
         _H = "birdShowOspfInterface",
         _X = function(path)
                 local cfg = s.readpopen("echo show ospf interface | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ["ospf-lsadb"] = s.newScheme{
         _H = "birdShowOspfLsadb",
         _X = function(path)
                 local cfg = s.readpopen("echo show ospf lsadb | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ["ospf-neighbors"] = s.newScheme{
         _H = "birdShowOspfNeighbors",
         _X = function(path)
                 local cfg = s.readpopen("echo show ospf neighbors | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ["ospf-state"] = s.newScheme{
         _H = "birdShowOspfState",
         _X = function(path)
                 local cfg = s.readpopen("echo show ospf state | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ["ospf-topology"] = s.newScheme{
         _H = "birdShowOspfTopology",
         _X = function(path)
                 local cfg = s.readpopen("echo show ospf topology | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ["protocols"] = s.newScheme{
         _H = "birdShowProtocols",
         _X = function(path)
                 local cfg = s.readpopen("echo show protocols | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ["route"] = s.newScheme{
         _H = "birdShowRoute",
         _X = function(path)
                 local cfg = s.readpopen("echo show route | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ["static"] = s.newScheme{
         _H = "birdShowStatic",
         _X = function(path)
                 local cfg = s.readpopen("echo show static | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ["status"] = s.newScheme{
         _H = "birdShowStatus",
         _X = function(path)
                 local cfg = s.readpopen("echo show status | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
      ["symbols"] = s.newScheme{
         _H = "birdShowSymbols",
         _X = function(path)
                 local cfg = s.readpopen("echo show symbols | birdc 2>&1 | grep -v 'bird>'")
                 return cfg or ""
              end,
         _R = 'user',
      },
   },
   extra = s.stringType{
       _H = "birdExtra",
   },
   debug = debugScheme{ 
      _H = "birdDebug",
      commands = s.choiceType{
         _H = "birdDebugCommands",
         _A = {
            {"off", ""},
            {"medium", ""},
            {"all", ""},
         },
         _D = "off",
      },
   },
   ["router-id"] = s.ipAddressType{
      _H = "birdRouterid"
   },
   rip = s.newScheme{
      _H = "birdRip",
      _LIST = protcolScheme{
         _H = "birdRipList",
         authentication = s.choiceType{
            _H = "birdRipAuthentication",
            _A = {
               {"none", ""},
               {"plaintext", ""},
               {"md5", ""},
            },
            _D = "none"
         },
         honor = s.choiceType{
            _H = "birdRipAuthentication",
            _A = {
               {"always", ""},
               {"neighbor", ""},
               {"never", ""},
            },
            _D = "never"
         },
         port = s.integerType{
            _H = "birdRipPort",
            _A = {0, 65535},
         },
         infinity = s.integerType{
            _H = "birdRipInfinity",
            _A = {0, 65535},
            _D = 16
         },
         period = s.integerType{
            _H = "birdRipPeriod",
            _A = {0, 65535},
            _D = 30
         },
         ["timeout-time"] = s.integerType{
            _H = "birdRipTimeout",
            _A = {0, 65535},
            _D = 120
         },
         ["garbage-time"] = s.integerType{
            _H = "birdRipGarbage",
            _A = {0, 65535},
            _D = 300
         },
         interface = s.newScheme{
            _NAME = s.newScheme{
               mode =  s.choiceType{
                  _H = "birdRipInterfaceMode",
                  _A = {
                     {"multicast", ""},
                     {"broadcast", ""},
                     {"quiet", ""},
                     {"nolisten", ""},
                     {"version1", ""},
                  },
                  _D = "multicast"
               },
               metric = s.integerType{
                  _H = "birdRipInterfaceMetric",
                  _A = {0, 65535},
                  _D = 1
               },
            },
         },
      },
   },
   static = s.newScheme{
      _H = "birdStatic",
      _LIST = protcolScheme{
         _H = "birdStaticList",
         interface = s.nilType,
         ["router-id"] = s.nilType,
         route =  s.newScheme{
            _P = 1000,
            _H = "birdStaticRoute",
            _NAME_TYPE = s.ipprefixVal,
            _NAME_H = "birdStaticRoutePrompt",
            _NAME =  s.newScheme{
               _H =  "birdStaticRouteName",
               _N =  function (sch, conf)
                        return {
                           true,
                           action = true,
                           gateway = conf.action=="via-gateway",
                           interface = conf.action=="via-interface",
                        }
                     end,
               action  = s.choiceType{
                  _H = "birdStaticRouteAction",
                  _A = {
                     {"via-interface", ""},
                     {"via-gateway", ""},
                     {"drop", ""},
                     {"reject", ""},
                     {"prohibit", ""},
                  },
                  _DEP = 1,
               },
               gateway = s.ipAddressType{
                  _H = "birdStaticRouteGateway",
               },
               interface = s.stringType{
                  _H = "birdStaticRouteInterface",
               },
            },
         },
      },
   },
   bgp = s.newScheme{
      _H = "birdBGP",
      _LIST = protcolScheme{
         _H = "birdBGPList", 
         ["local-as"] =s.integerType{
            _H = "birdBGPLocalas",
            _A = {0, 4294967295},
            _P = 95,
            _D =0,
         },
         ["neighbor-as"] =s.integerType{
            _H = "birdBGPNeighboras",
            _A = {0, 4294967295},
            _P = 95,
            _D =0,
         },
         ["neighbor-ip"] =s.ipAddressType{
            _H = "birdBGPNeighborip",
            _P = 95,
            _D = "0.0.0.0",
         },
         multihop =  s.newScheme{
            _H = "birdBGPMultihop",
            _N = function (sch, conf)
                    return {
                       [1] = nil,
                       ["number"] = conf.enable ~= "on" or nil,
                       ["via"]  = conf.enable ~= "on" or nil,
                    }
                 end,
            enable = s.choiceType{
               _H = "birdBGPMultihopEnable",
               _A = {
                  {"on",""},
                  {"off",""},
               },
               _D = "off",
               _DEP = 1,
            },
            ["number"] = s.integerType{
               _H = "birdBGPMultihopNumber",
               _A = {1, 65535},
               _D = 1,
            },
            ["via"] = s.ipAddressType{
               _H = "birdBGPMultihopVia",
            },
         },
         ["next-hop-self"] = s.choiceType{
            _H = "birdBGPNexthopself",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "off"
         },
         ["missing-lladdr"] = s.choiceType{
            _H = "birdBGPmissinglladdr",
            _A = {
               {"self",""},
               {"drop",""},
               {"ignore",""},
            },
         },
         ["source-address"] = s.ipAddressType{
            _H = "birdBGPSourceaddress",
         },
         password = s.stringType{
            _H = "birdBGPPassword",
         },
         passive  = s.choiceType{
            _H = "birdBGPPassive",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "off"
         },
         ["rr-client"]  = s.choiceType{
            _H = "birdBGPRrclient",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "off"
         },
         ["rr-cluster-id"] = s.ipAddressType{_H = "birdBGPRrclusterid",},
         ["rs-client"]  = s.choiceType{
            _H = "birdBGPRsclient",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "off"
         },
         ["enable-route-refresh"]  = s.choiceType{
            _H = "birdBGPEnablerouterefresh",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "on"
         },

         ["interpret-communities"]  = s.choiceType{
            _H = "birdBGPInterpretcommunities",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "on"
         },
         ["enable-as4"]   = s.choiceType{
            _H = "birdBGPEnableas4",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "on"
         },
         capabilities  = s.choiceType{
            _H = "birdBGPCapabilities",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "on"
         },
         ["advertise-ipv4"]  = s.choiceType{
            _H = "birdBGPAdvertiseipv4",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "on"
         },
         ["route-limit"] = s.integerType{
            _H = "birdBGPRoutelimit",
            _A = {1}
         },
         ["disable-after-error"] = s.choiceType{
            _H = "birdBGPDisableaftererror",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "off"
         },
         ["hold-time"] = s.integerType{
            _H = "birdBGPHoldtime",
            _A = {1,65536}
         },
         ["startup-hold-time"] = s.integerType{
            _H = "birdBGPStartholdtime",
            _A = {1,65536}
         },
         ["keepalive-time"] = s.integerType{
            _H = "birdBGPKeepalivetime",
            _A = {1,65536}
         },
         ["connect-retry-time"] = s.integerType{
            _H = "birdBGPConnectretrytime",
            _A = {1,65536}
         },
         ["start-delay-time"] = s.integerType{
            _H = "birdBGPStartdelaytime",
            _A = {1,65536}
         },
         ["error-wait-time-min"] = s.integerType{
            _H = "birdBGPErrorwaittimemin",
            _A = {1,65536},
            _D = 60,
         },
         ["error-wait-time-max"] = s.integerType{
            _H = "birdBGPErrorwaittimemax",
            _A = {1,65536},
            _D = 300,
         },
         ["error-forget-time"] = s.integerType{
            _H = "birdBGPErrorforgettime",
            _A = {1,65536}
         },
         ["path-metric"] = s.choiceType{
            _H = "birdBGPPathmetric",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "on"
         },
         ["prefer-older"] = s.choiceType{
            _H = "birdBGPPreferolder",
            _A = {
               {"on",""},
               {"off",""},
            },
            _D = "off"
         },
         ["default-bgp_med"] = s.integerType{
            _H = "birdBGPDefaultbgpmed",
            _A = {0, 65535},
            _D = 0,
         },
         ["default-bgp_local_pref"] = s.integerType{
            _H = "birdBGPDefaultbgplocalpref",
            _A = {0, 65535},
            _D = 100,
         },
      },
   },
   ospf = s.newScheme{
      _H = "birdOSPF",
      _LIST = protcolScheme{
         _H = "birdOSPFList",
         interface = s.nilType,
         rfc1583compat = s.booleanType{ 
            _H = "birdOSPFRfc1583compat",
            _D = false,
         },
         tick = s.integerType{
            _H = "birdOSPFTick",
            _A = {0, 65535},
            _D =1,
         },
         area = s.newScheme{
            _H = "birdOSPFArea",
            _NAME_H =  "birdOSPFAreaIdPrompt",
            _NAME_TYPE = s.ipaddressVal,
            _NAME =  s.newScheme{
               _H = "birdOSPFAreaId",
               stub =  s.booleanType{
                  _H = "birdOSPFAreaStub",
                  _D = false,
               },
               ["stub-cost"] =s.integerType{
                  _H = "birdOSPFStubcost",
                  _A = {0, 65535},
               },
               networks =  s.newScheme{
                  _H = "birdOSPFAreaNetworks",
                  _NAME_TYPE = s.ipprefixVal,
                  _NAME = s.newScheme{
                     hidden =  s.booleanType{
                        _H = "birdOSPFAreaNetworksHidden",
                        _D = false,
                     },
                  },
               },
               stubnet =  s.newScheme{ 
                  _H = "birdOSPFAreaStubnet",
                  _NAME_TYPE = s.ipprefixVal,
                  _NAME = s.newScheme{
                     _H = "birdOSPFAreaStubnetPrefix",
                     hidden =  s.booleanType{
                        _H = "birdOSPFAreaStabnetHidden",
                        _D = false,
                     },
                     summary =  s.booleanType{
                        _H = "birdOSPFAreaStubnetSummary",
                        _D = false,
                     },
                  },
               },
               interface =  s.newScheme{
                  _H = "birdOSPFAreaInterface",
                  _LIST = s.newScheme{
                     name  =  s.stringType{
                        _H = "birdOSPFAreaInterfaceName",
                     },
                     cost = s.integerType{
                        _H = "birdOSPFAreaInterfaceCost",
                        _A = {0, 65535},
                        _D = 10,
                     },
                     stub =  s.booleanType{
                        _H = "birdOSPFAreaInterfaceStub",
                        _D = false,
                     },
                     hello = s.integerType{
                        _H = "birdOSPFAreaInterfaceHello",
                        _A = {0, 65535},
                        _D = 10,
                     },
                     poll = s.integerType{
                        _H = "birdOSPFAreaInterfacePoll",
                        _A = {0, 65535},
                        _D = 20,
                     },
                     retransmit = s.integerType{
                        _H = "birdOSPFAreaInterfaceRetransmit",
                        _A = {0, 65535},
                        _D = 5,
                     },
                     priority = s.integerType{
                        _H = "birdOSPFAreaInterfacePriority",
                        _A = {0, 65535},
                        _D = 1,
                     },
                     wait = s.integerType{
                        _H = "birdOSPFAreaInterfaceWait",
                        _A = {0, 65535},
                        _D = 40,
                     },
                     ["dead-count"] = s.integerType{
                        _H = "birdOSPFAreaInterfaceDeadcount",
                        _A = {0, 65535},
                     },
                     dead = s.integerType{
                        _H = "birdOSPFAreaInterfaceDead",
                        _A = {0, 65535},
                     },
                     ["rx-buffer"] = s.integerType{
                        _H = "birdOSPFAreaInterfaceRxbuffer",
                        _A = {0, 65535},
                     },
                     type = s.choiceType{
                        _H = "birdOSPFAreaInterfaceType",
                        _A = {
                           {"broadcast", ""},
                           {"nonbroadcast", ""},
                           {"pointopoint", ""},
                        },
                     },
                     ["strict-nonbroadcast"] =  s.booleanType{
                        _H = "birdOSPFAreaInterfaceStrictnonbroadcast",
                        _D = false,
                     },
                     authentication = s.choiceType{
                        _H = "birdOSPFAreaInterfaceAuthentication",
                        _A = {
                           {"none", ""},
                           {"simple", ""},
                           {"cryptographic", ""},
                        },
                        _D = "none",
                     },

                     password =  s.stringType{
                        _H = "birdOSPFAreaInterfacePassword"
                     },
                     neighbors = s.newScheme{
                        _H = "birdOSPFAreaInterfaceNeigborsName",
                        _NAME_H = "birdOSPFAreaInterfaceNeigborsPrompt",
                        _NAME_TYPE = s.ipaddressVal,
                        _NAME = s.newScheme{
                           eligible = s.booleanType{
                              _H = "birdOSPFAreaInterfaceNeighbors",
                              _D = false,
                           },
                        },
                     },
                  },
               },
               ["virtual-link"] =  s.newScheme{
                  _H = "birdOSPFAreaVirtuallink",
                  _NAME_H = "birdOSPFAreaVirtuallinkPrompt",
                  _NAME_TYPE = s.ipaddressVal,
                  _NAME = s.newScheme{
                     hello = s.integerType{
                        _H = "birdOSPFAreaVirtuallinkHello",
                        _A = {0, 65535},
                        _D = 10,
                     },
                     retransmit = s.integerType{
                        _H = "birdOSPFAreaVirtuallinkRetransmit",
                        _A = {0, 65535},
                        _D = 5,
                     },
                     wait = s.integerType{
                        _H = "birdOSPFAreaVirtuallinkWait",
                        _A = {0, 65535},
                        _D = 40,
                     },
                     ["dead-count"] = s.integerType{
                        _H = "birdOSPFAreaVirtuallinkDeadcount",
                        _A = {0, 65535},
                     },
                     dead = s.integerType{
                        _H = "birdOSPFAreaVirtuallinkDead",
                        _A = {0, 65535},
                     },
                     authentication = s.choiceType{
                        _H = "birdOSPFAreaVirtuallinkAuthentication",
                        _A = {
                           {"none", ""},
                           {"simple", ""},
                           {"cryptographic", ""},
                        },
                        _D = "none",
                     },
                     password =  s.stringType{
                        _H = "birdOSPFAreaVirtuallinkPassword",
                     },
                  },
               },
            },
         },
      },
   },
   device = s.newScheme{
      _H = "birdDevice",
      ["scan-time"] = s.integerType{
         _H = "birdDeviceScantime",
         _A = {0, 65535},
      },
      primary = s.newScheme{
         _H = "birdPrimary",
         _NAME_TYPE = s.ipprefixOrAddressVal,
         _NAME_H = "birdDevicePrefixPrompt",
         _NAME = s.newScheme{
            _H = "birdDevicePrefix",
            interface = s.stringType{
               _H = "birdDeviceInterface"
            },
         },
      },
   },
   kernel = s.newScheme{
      _H = "birdKernel",
      _LIST = protcolScheme{
         interface = s.nilType,
         ["router-id"] = s.nilType,
         ["scan-time"] = s.integerType{
            _H = "birdKernelScantime",
            _A = {0, 65535},
         },
         learn = s.booleanType{ 
            _H = "birdKernelLearn",
         },
         ["device-routes"] = s.booleanType{ 
            _H = "birdKernelDeviceroutes",
         },
         ["kernel-table"] = s.integerType{
            _H = "birdKernelKerneltable",
            _A = {0, 65535},
         },
      },
   },
   direct = s.newScheme{
      _H = "birdDirect",
      _LIST = protcolScheme{
         _H = "birdDirectList",
         ["router-id"] = s.nilType,
         interface =  s.newScheme{
            _H = "birdDirectInterface",
            _LIST = s.stringType{
               _H = "birdDirectInterfaceName",
            },
         },
      },
   },
   pipe = s.newScheme{
      _H = "birdPipe",
      _LIST = protcolScheme{
         _H = "birdPipeList",
         ["router-id"] = s.nilType,
         interface = s.nilType,
         preference = s.nilType,
         ["peer-table"] = s.stringType{
            _H = "birdPipePeertable",
         },
         mode = s.choiceType{
            _H = "birdPipeMode",
            _A = {
               {"opaque", ""},
               {"transparent", ""},
            },
            _D = "opaque",
         },
      },
   },
   table = s.newScheme{
      _H = "birdTable",
      _LIST = s.stringType{_H = "birdTableName"}
   },
}
