module(..., package.seeall)
local socket = require'socket'
local nsg = require'nsg'
--local serialize = require'serialize'
local ser = nsg.serialize

local MESSAGE_DELIMITER = '\003'

local function sendMessage(self, mes)
  --print('sendMessage', mes)
  mes = mes:gsub('\\','\\\\'):
  gsub(MESSAGE_DELIMITER, '\\'..MESSAGE_DELIMITER)..MESSAGE_DELIMITER

  self.sock:send(mes)
end

local function sockRead(sock, pattern, prefix)
  pattern = pattern or 2048
  local data, emsg, partial = sock:receive(pattern, prefix)
  if data then
    return data
  end
  if partial and #partial > 0 then
    return partial
  end
  return nil, emsg
end

local function findDelim(buf)
  local s
  local e = 0
  repeat
    s, e = buf:find("\\*"..MESSAGE_DELIMITER..'$', e + 1)
    if s and (e-s)%2==0 then
      return e
    end
  until not s
  return nil
end

local function getMessage(self, timeout)
  while true do
    socket.select({self.sock}, nil, timeout)
    local data, err = sockRead(self.sock, 8192)
    if not data then return nil, err end
    if data:sub(-1) == MESSAGE_DELIMITER then
      self.buf = self.buf..table.concat(self.packets)..data
      self.packets = {}
      while true do
        local d = findDelim(self.buf)
        if not d then break end
        local mes = self.buf:sub(1, d-1):gsub("\\(.)", "%1")
        self.buf = self.buf:sub(d + 1)
        --print('getMessage', mes)
        return mes
      end
    else
      self.packets[#self.packets+1] = data
    end
  end
end

local function send(self)
end

local function clean(...)
  local n = select('#', ...)
  if (...) then
    return select(2, ...)
  else
    return nil, select(2, ...)
  end
end
local function receive(self, timeout)
  local mes, err = self:getMessage(timeout)
  if not mes then return nil, "nsgconf session failed: "..err end
  local res, err = loadstring('return '..mes)

  if not res then
    return nil, err
  else
    setfenv(res, {})
    return clean(pcall(res))
  end
end

local function closecon(self)
  self.sock:close()
end

function open(args)
  args = args or {}
  local sesnum = args.sesnum or 0
  local address = args.address or '127.0.0.1'
  local port = args.port or 50007
  local locaddr = args.locaddr or '127.0.0.1'
  local locport = args.locport or 0

  local sock, err = socket.connect(address, port, locaddr, locport)
  if not sock then
    return nil, err
  end
  sock:settimeout(0)

  local session = {
    timeout = 7,
    buf = "",
    packets = {},
    sock = sock,
    sesnum = sesnum,
    sendMessage = sendMessage,
    getMessage = getMessage,
    send = send,
    receive = receive,
    closecon = closecon,
    __index = function(t, com) 
      return function(self, ...) 
        local par = {}
        for i, v in ipairs({...}) do
          par[i] = ser(v)
        end
        local mes = ("S[%s]:%s(%s)"):format(tostring(self.sesnum), com,
          table.concat(par,','))
        while self:getMessage(0) do end -- purge receive buffer
        self:sendMessage(mes)
        return self:receive(self.timeout)
      end
    end

  }
  setmetatable(session, session)

  local n, err
  if sesnum == 0 then
    args.mode = args.mode or"text"
    session:sendMessage("S[0]:init("..ser(args)..")")
    n, err = session:receive(session.timeout)
    --print('receive', n, err)
    if not n then
      session:closecon()
      return nil, err
    end
    session.sesnum = n
  else
    local res, err = session:isSession()
    return res and session, err
  end

  return session, err
end
