module(..., package.seeall)
local s = require'scheme'

local clonePath = s.clonePath

local function checkName(val)
   if  nil == string.match(val, "^[A-Za-z][A-Za-z0-9-_]*$") or #val > 30 then
      return nil, 'Incorrect config slot name'
   end
   return val
end

local function confToTable(str)
   local confTable, res
   local func, err = loadstring('setfenv(1, {}) '..str, "")
   if not func then
      return {_err = err}
   end
   local res, confTable = pcall(func)
   return confTable
end

local function getCurrentConfigs()
   local configs={}
   local conf = C.getConfByPath({"system","configurations"}) or {}
   local sch  = C.getSchemeByPath({"system","configurations"})or {}
   for k in pairs(conf) do
      if ( not (sch[k] or (tostring(k)):find('^[_].*$'))) then
         configs[k] = true
      end
   end
   return configs
end

local function restoreCurrentConfigs(configs)
   for k in pairs(configs) do
      C.setConfigValue({"system", "configurations", k, "_keep"}, true)
   end
end


configScheme = s.newScheme{
   _H = "configurations",
   _HTML = "configweb",
   _Z = function(path, sch, conf, pconf)
           local fpath = "/etc/"
           if C.BI and C.BI.env and C.BI.env.safe_mode then
              fpath = "/etc/etc.sav/"
           end
           local files = s.readpopen("ls "..fpath.."nsgconfig.bu.* 2>/dev/null")
           for name in files:gmatch('[^A-Za-z0-9-_./]*([A-Za-z0-9-_./]+)[^A-Za-z0-9-_./]*') do
              local shortName = name:gsub("^.*/nsgconfig%.bu%.","")
              if name and #name>5 and (conf == nil or conf[shortName]==nil) then
                 os.execute("rm -f "..name)
              end
           end
           return "OK"
        end,
   ["clear-config"] = s.newScheme{
      _H = 'loadDefault',
      _X = function()
              local old = getCurrentConfigs()
              C.setConfigValue({}, {})
              restoreCurrentConfigs(old)
              return "OK"
           end,
      _A = {"Default, tested and FLASHED.\n"},
      _R = 'admin',
      _P = 100,
      _DEP = 10,
   },
   ["reload-saved"] = s.newScheme{
      _H = 'reloadSaved',
      _X = function()
              local old = getCurrentConfigs()
              C.loadConfig()
              restoreCurrentConfigs(old)
              return "OK"
           end,
      _A = {"Default, tested and FLASHED.\n"},
      _R = 'admin',
      _P = 100,
      _DEP = 10,
   },
   _NAME_TYPE= checkName,
   _NAME = s.newScheme{
      _R = 'admin',
      _H = "configN",
      _N = function(sch, conf, path)
              local fpath = "/etc/"
              if C.BI and C.BI.env and C.BI.env.safe_mode then
                 fpath = "/etc/etc.sav/"
              end
              local fname = fpath.."nsgconfig.bu."..tostring(path[#path])
              local f = io.open(fname)
              if f then
                 f:close()
                 f = true
              end
              return {
                 save = f,
                 show = not f or nil,
                 clear = not f or nil,
                 load = not f or nil,
              }
              end,
      save = s.newScheme{
         _R = 'admin',
         _H = 'configNSave',
         _DEP = 1,
         _X = function(path)
                 local conf = C.getConfByPath({})
                 local fpath = "/etc/"
                 if C.BI and C.BI.env and C.BI.env.safe_mode then
                    fpath = "/etc/etc.sav/"
                 end
                 local fname = fpath.."nsgconfig.bu."..tostring(path[#path - 1])
                 local f, err = io.open(fname, 'w')
                 if f then
                    f:write("return", s.ser(conf))
                    f:close()
                    --os.execute'savecfg 2>/dev/null' -- TODO
                    local ox = require'posix'
                    ox.chmod(fname, 'a-rwx,u+wr')
                    return 'Configuration saved'
                 else
                    return false, strf(
                       "Can't write configuration to file %s(%s)", fname, err)
                 end
              end,
      },
      load = s.newScheme{
         _R = 'admin',
         _H = 'configNLoad',
         _X = function(path)
                 local old = getCurrentConfigs()
                 local fpath = "/etc/"
                 if C.BI and C.BI.env and C.BI.env.safe_mode then
                    fpath = "/etc/etc.sav/"
                 end
                 local fname = fpath.."nsgconfig.bu."..tostring(path[#path - 1])
                 local confstr = s.readpopen("cat "..fname.." 2>&1")
                 C.setConfigValue({}, confToTable(confstr))
                 restoreCurrentConfigs(old)
                 return "OK"
              end,
         _DEP = 10,
      },
      clear = s.newScheme{
         _H = 'configNClear',
         _R = 'admin',
         _DEP = 1,
         _X = function(path)
                 local fpath = "/etc/"
                 if C.BI and C.BI.env and C.BI.env.safe_mode then
                    fpath = "/etc/etc.sav/"
                 end
                 local fname = fpath.."nsgconfig.bu."..tostring(path[#path - 1])
                 return s.readpopen("rm -f "..fname.." 2>&1")
              end,
      },
      show = s.newScheme{
         _H = "configNShow",
         _R = 'admin',
         _P = 75,
         _X = function(path)
                 local fpath = "/etc/"
                 if C.BI and C.BI.env and C.BI.env.safe_mode then
                    fpath = "/etc/etc.sav/"
                 end
                 local fname = fpath.."nsgconfig.bu."..tostring(path[#path - 1])
                 return  s.ser(confToTable( s.readpopen("cat "..fname.." 2>&1")), 1)
             end
      },
   },
   
}
