function RootOrAdminScheme(sch)
  return function(_, path)
    return (C.session and 
      (C.session.user == 'root' or C.session.rights == 'admin')
    ) and 
    sch
  end
end

local s = require'scheme'
--local ppp = require'nsgpppconf'
local sys = require'nsgsysconf'
local ser = require'nsgserconf'
local tools = require'nsgtoolsconf'
local oneWire = require'nsg1wireconf'
local rs232 = require'nsgrs232conf'
local ipnod = require'nsgipconf'
local eth = require'nsgethconf'
local port = require'nsgportconf'
local WiFi = require'nsgwifi'
local ipsec = require'nsgipsecconf'
local uitcp = require'nsguitcpconf'
--local nf = require'nsgnfconf'
local pppoe = require'nsgpppoeconf'
local pptp = require'nsgpptpconf'
local stunnel = require'nsgstunnel'
local gre = require'nsggre'
local usb = require'nsgusbconf'
local search = require'nsgsearch'

local portListScheme = s.newScheme{
  _H = "portList",
  ["3g"] = BI['3g'] and function(_, path) 
    C.setOption(path, 'wirelessType', "3g")
    C.setOption(path, 'simNumber', "2")
    C.setOption(path, 'controlledPower', "pin")
    C.setOption(path, 'smsType', "aux")
    return port.portScheme
  end,
  edge = BI.edge and function(_, path) 
    C.setOption(path, 'wirelessType', "edge")
    C.setOption(path, 'simNumber', "2")
    C.setOption(path, 'controlledPower', "pin")
    C.setOption(path, 'smsType', "+++")
    return port.portScheme
  end,

  cdma = BI.cdma and function(_, path) 
    C.setOption(path, 'wirelessType', "cdma")
    C.setOption(path, 'simNumber', "1")
    C.setOption(path, 'controlledPower', "pin")
    return port.portScheme
  end,

  ["1-wire"] = BI["1-wire"] and function(_, path) 
    usb.probeUSBDeviceScheme(nil, path, 2) -- second iface !
    oneWire.setDevName()
    return oneWire.oneWireScheme
  end,

  ["rs-232"] = BI.rs232 and rs232.rs232Scheme,
  eth0 = eth.ethScheme{_H = "portEth0"},
  eth1 = BI.eth and BI.eth > 1 and eth.ethScheme{_H = "portEthN"},
  eth2 = BI.eth and BI.eth > 2 and eth.ethScheme{_H = "portEthN"},
  eth3 = BI.eth and BI.eth > 3 and eth.ethScheme{_H = "portEthN"},
  eth4 = BI.eth and BI.eth > 4 and eth.ethScheme{_H = "portEthN"},
  wlan0 = BI.wifi and WiFi.WiFiScheme,

  s1 = BI.s1 and  usb.portTemplateFunction,   

  s2 = BI.s2 and  usb.portTemplateFunction,   

  m1 = BI.m1 and  usb.portTemplateFunction,   

  a1 = BI.a and BI.a > 0 and function(_, path)
    usb.port2tty["a1"] = "/dev/ttyS2" -- TEMPO!
    return rs232.rs232Scheme
  end,
  a2 = BI.a and BI.a > 1 and function(_, path)
    usb.port2tty["a2"] = "/dev/ttyS3" -- TEMPO!
    return  rs232.rs232Scheme
  end,
  a3 = BI.a and BI.a > 2 and function(_, path)
    usb.port2tty["a3"] = "/dev/ttyS4" -- TEMPO!
    return  rs232.rs232Scheme
  end,

}

local tunnelListScheme= s.newScheme{
  _H = "tunnelList",
  ipsec = ipsec.ipsecScheme,
  pppoe = pppoe.pppoeScheme,
  pptp = pptp.pptpScheme,
  uitcp = uitcp.uitcpScheme,
  stunnel = stunnel.stunnelScheme,
  gre = gre.greScheme,
}

return s.newScheme{
  _H = "root",
  _Z = function() return true end,
  bridge = BI.bridge and eth.bridgeScheme,
  ip = ipnod.ipScheme,
  port = portListScheme,
  system = sys.systemScheme,
  services = ser.servicesScheme,
  tools = tools.toolsScheme,
  tunnel = tunnelListScheme,
  ["ethernet-switch"] = (
    BI.arch == 'nsg600' and BI.eth and BI.eth > 1 and
    eth.ethernetSwitchScheme
  ),
  search = search.searchHelpScheme,
}
