module(..., package.seeall)
local s = require'scheme'

local addDynNameserverValErr = "table in format {'<ifname>'[, 'DNS address', ...]}"
local function addDynNameserverVal(val)
  if type(val) == 'table' and type(val[1]) == 'string' then
    for i = 2, 4 do
      if val[i] and not s.ipaddressVal(val[i]) then
        return nil, addDynNameserverValErr
      end
    end
    return val
  end
  return nil, addDynNameserverValErr
end

local function getResolvConfNameservers(dnsList, src)
  local data = ''
  if dnsList then
    for k, v in ipairs(dnsList) do
      data = data..'nameserver '..v..'\t# From '..src..'\n'
      if k >= 3 then break end
    end
  end
  return data
end

local function createResolvConfAux(conf, dynDns)
  conf = conf or {}
  local data = {}
  for k, v in pairs(dynDns) do
    data[1] = getResolvConfNameservers(v, k)
    break
  end
  if not data[1] then
    data[1] = getResolvConfNameservers(conf["name-server"], "configuration")
  end
  if conf.domain then
    data[#data+1] = 'domain '..conf.domain..'\n'
  end
  if conf.search then
    data[#data+1] = 'search '..conf.search..'\n'
  end
  if conf.timeout then
    data[#data+1] = 'options timeout:'..conf.timeout..'\n'
  end
  if conf.attempts then
    data[#data+1] = 'options attempts:'..conf.attempts..'\n'
  end
  if conf.rotate then
    data[#data+1] = 'options rotate\n'
  end
  if not conf["check-names"] then
    data[#data+1] = 'options no-check-names\n'
  end


  local f = io.open('/etc/resolv.conf', 'w')
  if f then
    f:write(table.concat(data))
    f:close()
  end
end

local function createResolvConf(path, sch, conf, prevConf)
  --if not s.confCmp(conf, prevConf) then  
  path = s.clonePath(path)
  conf = s.addDefault(path, sch, conf or {}, true)
  --[[
    iface_mod.setAliasHook(ipRouteAliasHookKey, 
      function()
        return  ipRouteApplyAux(path, sch, conf)
      end
    )
    ]]
  createResolvConfAux(conf, C.getOption(path, 'dyndns') or {})
  --end
  return true

end

local function addDynNameserver(path, sch, val)
  path = s.clonePath(path, -1)
  local dynDns = C.getOption(path, 'dyndns') or {}
  if val[2] then
    dynDns[val[1]] = {val[2], val[3], val[4]}
  else
    dynDns[val[1]] = nil
  end
  C.setOption(path, 'dyndns', dynDns)
  -- setAliasHook ?
  createResolvConfAux(C.getConfByPath(path), dynDns)
end

dnsClientScheme = s.newScheme{
  _H = 'dnsClient',
  _Z = createResolvConf,
  ["name-server"] = s.newScheme{
    _H = 'dnsClientNameserverList',
    _LIST = s.ipAddressType{_H = "dnsClientNameserver"},
    _P = 100,
  },

  domain = s.stringType{
    _H = 'dnsClientDomain',
  },

  search = s.stringType{
    _H = 'dnsClientSearch',
  },

  timeout = s.integerType{
    _H = 'dnsClientTimeout',
    _A = {1, 50}
  },

  attempts = s.integerType{
    _H = 'dnsClientAttempts',
    _A = {1, 5}
  },

  rotate = s.booleanType{
    _H = 'dnsClientRotate',
  },

  ["check-names"] = s.booleanType{
    _H = 'dnsClientCheckNames',
    _D = true,
  },

  ['add-dyn-nameservers'] = function(_, path)
    return (C.session and C.session.script) and 
    s.newScheme{
      _H = 'dnsClientAddDynNameserver',
      _T = 'string',
      _V = addDynNameserverVal,
      _X = addDynNameserver,
      _R = 'user',
      _P = -90,
    }
  end,
  show  = s.newScheme{
    _H = "dnsClientShow",
    _X = function()
      return('\n'..s.readpopen("cat /etc/resolv.conf 2>&1"))
    end,
    _R = 'user',
    _P = -100,
  },
}
