module(..., package.seeall)
local s = require'scheme'
local sock = require'socket'
local i = require'nsgifaceconf'
local e = require'nsgethconf'

local clonePath = s.clonePath

local function clearConfForGre(conf)
   conf = s.deepcopy(conf)
   local allowKey = {
      encapsulation = 1,
      source = 1,
      device = 1,
      destination = 1,
      key = 1,
      checksum = 1,
      keepalive = 1,
      ["keepalive-interval"] = 1,
      ["keepalive-retry"] = 1,
      ["adm-state"] = 1,
   }
   for k, v in pairs(conf) do
      if nil == allowKey[k]
      then
         conf[k] = nil
      end
   end
   return conf
end


local greDevConfigs = {

}

local greNeteventKey = {}
local greUpdateTunnelState =
   function(msg)
      if msg and msg.ifname and greDevConfigs[msg.ifname] and (msg.type=='dellink' or  msg.type=='newlink') then
         local sock=require"socket"
         local udp=sock:udp()
         if udp then
            udp:settimeout(2)
            for k,v in pairs(greDevConfigs[msg.ifname]) do
               local cmd
               if msg.type=='newlink' and msg.ifup then
                  cmd = 'addTunnel("'..tostring(k)..'",'..s.ser(clearConfForGre(v))..')'
               elseif msg.type=='dellink' or (msg.type=='newlink' and not msg.ifup)  then
                  cmd = 'delTunnel("'..tostring(k)..'")'
               end
               C.log_debug("send cmd:"..cmd)
               udp:sendto(cmd,"127.0.0.1",50009)
               local res,ip,port = udp:receivefrom()
               if res then
                  if res:find("^Error") then
                     C.log_warn(res)
                  else
                     C.log_debug(res)
                  end
               end
            end
            udp:close()
         end
      end
   end

local function createGreConf(conf)
   --C.log_warn(s.ser(conf))
   greDevConfigs = {}
   local greConf
   for k,v in pairs(conf or {}) do
      if(tostring(k):find("^gre%d+$")) then
         if v["adm-state"] == "up" and v.destination and v.source then
            greConf = (greConf or "")..tostring(k).."="..s.ser(clearConfForGre(v))..",\n"
            if v.device then
               greDevConfigs[v.device] = greDevConfigs[v.device] or {}
               greDevConfigs[v.device][k] = v
            end
         end
      end
   end
   if greConf then
      greConf = "return {\n"..greConf.."}\n"
   end
   return greConf
end

local function restartGre(path, sch, conf, prevConf)
   local ret
   local defConf = s.addDefault(path, sch, conf, true)
   local grePath = s.deepcopy(path)
   grePath[#grePath + 1 ] = "gre" --avoiding to be in parent node of the 'ifAddress' and 'link' daemons
   C.restartDaemon(grePath)
   local greConf = createGreConf(defConf)
   if greConf and not s.confCmp(defConf, prevConf)
   then
      local fname = os.tmpname()
      local f = io.open(fname, 'w')
      if f then
         f:write(greConf)
         f:close()
         C.restartDaemon(
            grePath,
            {"/usr/bin/lua", "/usr/share/lua/5.1/gre.lua", fname},
            {noRestart = false,
             usePopen = true,
             atExit = C.increasingSleep,
             tmpFileList = {fname},
             }
         )
         C.updateNeteventCallback(
            greNeteventKey,
            greUpdateTunnelState
         )
      else
         return false, "can`t write config file"
      end
   end
   return true
end

local greTunnelScheme =
   function (_,path)
      local encapsulation = C.getConfByPath(clonePath(path, nil, "encapsulation")) or 'ip'
      local greSchemes = {
         eth = e.ethScheme,
         ip = i.interfaceScheme,
         x3 = s.newScheme,
      }
 
      local greScheme ={
         _H = "greTun",
         encapsulation = s.choiceType{
            _H = "greTunEncapsulation",
            _A = {
               {"ip", ""},
               {"eth", ""},
            },
            _D = "ip",
            _DEP = 1,
            _P = 100,
         },
         show = s.newScheme{
            _H = "ifShow",
            _X = function(path)
                    return i.showInterfaceStat(tostring((path[#path-1])))
                 end,
            _R = 'user',
         },
         description = s.stringType{_H = "greTunDescription", _P = 100},
         _N = function(sch, conf)
                 if conf and (conf.keepalive ~= "yes") then
                    return {nil, ["keepalive-interval"] = true, ["keepalive-retry"] = true,}
                 end
                 return {}
              end,
         source = s.ipAddressType{
            _H = "greTunSource",
            _D = "0.0.0.0",
         },
         device = s.stringType{
            _H = "greTunDev",
            _D = "lo", --lo is always active device
         },
         destination = s.ipAddressType{
            _H = "greTunDestination",
         },
         key = s.integerType{
            _H = "greTunKey",
            _A = {0, 2^32 -1},
         },
         checksum = s.choiceType{
            _H = "greTunChecksum",
            _A = {
               {"yes", ""},
               {"no", ""},
            },
            _D = "no",
         },
         keepalive = s.choiceType{
            _H = "greTunKeepalive",
            _A = {
               {"yes", ""},
               {"no", ""},
            },
            _D = "no",
             _DEP = 1,
         },
         ["keepalive-interval"] = s.integerType{
            _H = "greTunKAInterval",
            _A = {1, 2^32 - 1},
            _D = 5,
         },
         ["keepalive-retry"] = s.integerType{
            _H = "greTunKARetry",
            _A = {1, 2^32 - 1},
            _D = 3,
         },
      }

      return greSchemes[encapsulation or 'x3'](greScheme)
   end

local function greTunnelVal(val)
   val = (tostring(val)):match("^gre(%d+)$") or val
   local n, err = s.integerVal(tonumber(val), 1, 255)
   if n then
      return "gre"..tostring(n)
   else
      return nil, err
   end
end

greScheme = s.newScheme{
   _H = "gre",
   _Z = restartGre,
   _NAME = greTunnelScheme,
   _NAME_TYPE = greTunnelVal,
   ["log"] =  s.newScheme{
      _H = "greShowLog",
      _X = function(path)
              path = s.clonePath(path)
              path[#path] = "gre"
              return C.getDaemonLog(path)
           end,
      _R = 'user',
   },
}
