module(..., package.seeall)
local s = require'scheme'
local ppp = require'nsgpppconf'
--------------------------
local pppoeDiscoveryScan = function(path, sch, conf, sesNum) 
  local iface = C.getConfOrDefault(s.clonePath(path, -2, "iface")) or "eth0"
  local com = {"/usr/sbin/pppoe", "-A", "-I", iface, }

  C.restartDaemon(path, com, {sesNum = sesNum, noTimestamp = true, noRestart = true, }) 
end


local pppoeDiscovery = s.newScheme {
   _H = "pppoeDiscovery",
   _A = {"\nSearching PPPoE-servers in the LAN segment.\n".. "It takes a few time...\n"},
   _X = pppoeDiscoveryScan,
   _XI = true,
}

local pppoeOneShow = s.newScheme {
   _H = "pppoeOneShow",

   ["log"] =  s.newScheme{
      _H = "pppoeOneShowLog",
      _X = function(path)
	      path = s.clonePath(path, -2)
	      return C.getDaemonLog(path)
	   end,
      _R = 'user',
   },  

   discovery = function(_, path)
		  local mode = C.getConfOrDefault(s.clonePath(path, -2, "mode"))
		  return mode == 'client' and pppoeDiscovery
	       end,
}
--------------------------
local function AddOptions(target, opt, value) 
   if value then
      target[#target + 1] = opt
      target[#target + 1] = value
   end
end

local function restartPppoe(path, sch, conf, prevConf)
   local conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
   local ApCmdStart = {}
   --clone PPP-conf with defaults (only MAIN)
   local pppconf = s.addDefault(path, sch, conf, true) 
   local mainCom = ppp.getPPPParams(path, pppconf, 'main', {})
   local ConfText = table.concat(mainCom, '\n')..'\n'
   local filename = os.tmpname()
   local f = io.open(filename, 'w')
   if f then
      f:write(ConfText)
      f:close()
   end

   if conf.mode and conf.mode == "server" then
      ApCmdStart = {'/usr/sbin/pppoe-server', '-s', '-k', '-F',  }

      AddOptions(ApCmdStart, '-I', conf.iface or 'eth0')
      AddOptions(ApCmdStart, '-C', conf.name)
      AddOptions(ApCmdStart, '-N', conf.limit)
      AddOptions(ApCmdStart, '-L', conf["ip-address-pool"].from)
      AddOptions(ApCmdStart, '-R', conf["ip-address-pool"].to)
      AddOptions(ApCmdStart, '-T', conf.idle)
      AddOptions(ApCmdStart, '-O', filename)
   else
      local subcmd = "pppoe -s " .. "-I " .. (conf.iface or 'eth0')
      if conf.name then
	 subcmd = subcmd .. ' -C ' .. conf.name
      end
      ApCmdStart = {'/usr/sbin/pppd', 'pty', subcmd, 'nodetach', 'sync', 'file', filename }
   end


   if conf["adm-state"] == 'up' then
      C.restartDaemon(path, ApCmdStart, { tmpFileList = {filename}, atExit = C.increasingSleep})
      return true
   end
   C.restartDaemon(path)
   return true
end

--------------------------
local pppoeOne = s.newScheme {

   _Z = restartPppoe,

   _H = "pppoeOne",

   _N = function(sch, conf)
	   local need = {
	      true,
	      mode = true,
	      iface = true,
	      name = true,
	      ppp = true,
	      show = true,
	      ["adm-state"] = true,
	   }
	   if conf and conf.mode then
	      if conf.mode == "server" then
		 need["limit"] = true
		 need["idle"] = true
		 need["ip-address-pool"] = true
	      else
-- needless for client 
	      end
	   end
	   return need
	end,

-- Common Lines 

   mode = s.newScheme {
      _H = "pppoxOneMode",
      _T = 'choice',
      _A = { "client", "server"},
      _V = s.choiceVal,
      _D = "client",
      _DEP = 1,
   },

   ["adm-state"] = s.newScheme{
      _T = "choice",
      _A = {"up", "down"},
      _V = s.choiceVal,
      _H = "pppoxOneAdmState",
      _D = "down",
   },

   iface = s.stringType { _H = "pppoeOneIface", },

   name = s.stringType { _H = "pppoeOneName", },

   ppp = ppp.pppScheme,

   show = pppoeOneShow,

-- PPPoE-server params
   
   
   limit = s.integerType { _H = "pppoeOneLimit", _A = {1, 100}, _D = 10, },
   
   ["ip-address-pool"] = s.IpAddressPool,   
   
   idle = s.integerType { _H = "pppoeOneIdle", _A = {30, 3600}, _D = 600, }, 

-- PPPoE-client params

}

pppoeScheme = s.newScheme {
   _H = "pppoeScheme",

   _NAME = pppoeOne,
}
