module(..., package.seeall)
local s = require'scheme'
local ppp = require'nsgpppconf'
local usb = require'nsgusbconf'

local function restartRawTcpDaemon(path, sch, conf)
  local conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
  local rt = conf["raw-tcp"] or {}
  local ttyDev = usb.port2tty[path[#path]]
  local com = {
    "dcdstarter", ttyDev,
    "-s", tostring(conf.baudrate) or "9600",
    "-t", (
      (tostring(conf.databits) or "8")..
      ((conf.parity or 'N'):sub(1,1):upper())..
      (tostring(conf.stopbits) or "1")
    ),
    "-i", rt["ip-address"] or "127.0.0.1",
    "-p", rt["tcp-port"] or 50002,
    "-k", rt["dcd-inquiry-time"] or 1000,
    "-c", rt["reconnect-delay"] or 30,
    "-r", rt["hangup-time"] or 2000,
  }
  if rt["server-mode"] then
    com[#com+1] = "-a"
  end
  local dl = rt["debug-level"]
  if dl > 0 then
    com[#com+1] = "-lstdout"
    if dl > 1 then
      com[#com+1] = "-"..("v"):rep(dl-1)
    end
  end
  C.restartDaemon(path, com, {noTimestamp = true, atExit = C.increasingSleep}) 
end

local function restartAsyncPort(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    if conf and conf["adm-state"] ~= 'down' then
      local encap = conf.encapsulation
      if encap == "raw-tcp" then
        restartRawTcpDaemon(path, sch, conf)
        return true
      elseif encap == 'ppp' then
        ppp.restartPPPDaemon(path, sch, conf)
        return true
      end
    end
    C.restartDaemon(path)
  end
  return true
end

local rawTcpScheme = s.newScheme{
  _H = "rs232Rawtcp",
  ["dcd-inquiry-time"] = s.integerType{
    _H = "rs232RawtcpDcdinquirytime",
    _A = {0, 60000},
    _D = 1000,
  }, 
  ["debug-level"] = s.integerType{
    _H = "rs232RawtcpDebuglevel",
    _A = {0, 4},
    _D = 1,
  }, 
  ["hangup-time"] = s.integerType{
    _H = "rs232RawtcpHanguptime",
    _A = {-1, 60000},
    _D = 2000,
  }, 
  ["ip-address"] = s.ipAddressType{
    _H = "rs232RawtcpIpaddress",
    _D = "127.0.0.1"
  },
  ["reconnect-delay"]  = s.integerType{
    _H = "rs232RawtcpReconnectdelay",
    _A = {1, 86400},
    _D = 30,
  }, 
  ["server-mode"] = s.booleanType{_H = "rs232RawtcpServer"},
  ["tcp-port"] = s.tcpPortType{
    _H = "rs232RawtcpTcpport",
    _D = 50002,
  },
}

local baudrateScheme = s.newScheme{
  _T = 'choice',
  _A = {
    --50,
    --75,
    --110,
    --134,
    --150,
    --200,
    300,
    600,
    1200,
    --1800,
    2400,
    4800,
    9600,
    19200,
    38400,
    57600,
    115200,
    230400,
  },
  _V = s.choiceVal,
  _H = "rs232Baudrate",
  _D = 9600,
}

startRawAccess = function(path, sch, conf, sesNum)
  local ttyDev = usb.port2tty[path[#path-2]]
  C.restartDaemon(s.clonePath(path, -2)) -- kill ppp daemon if it is
  local portConf = C.getConfWithDefault(s.clonePath(path, -2))
  C.restartDaemon(
    path, {'nsgcu', ttyDev, 
      "-s", tostring(portConf.baudrate) or "9600",
      "-t", (
        (tostring(portConf.databits) or "8")..
        ((portConf.parity or 'N'):sub(1,1):upper())..
        (tostring(portConf.stopbits) or "1")
      ),
    },
    {
      sesNum = sesNum, noTimestamp = true, noRestart = true,
    }
  ) 
end

local rawAccessScheme = s.newScheme{
  _H = "portRawAccessNode",
  launch = s.newScheme{
    _H = "portRawAccess",
    _X = startRawAccess,
    _XI = true,
  },
}

rs232Scheme = s.newScheme{
  _H = "rs232",
  _Z = restartAsyncPort,
  --[[
  _N = function(sch, conf)
    local need = {
      true,
      ["adm-state"] = true,
      encapsulation = true,
      baudrate = true,
      databits = true,
      stopbits = true,
      parity = true,
      ["log"] = true,
    }
    if conf and conf.encapsulation then
      need[conf.encapsulation] = true
    end
    return need
  end,
  ]]
  ["adm-state"] = s.newScheme{
    _T = "choice",
    _A = {"up", "down"},
    _V = s.choiceVal,
    _H = "rs232AdmState",
    _D = "up",
  },
  encapsulation = s.newScheme{
    _T = "choice",
    _A = {"ppp", "raw-tcp", "none"},
    _V = s.choiceVal,
    _H = "rs232Encapsulation",
    _D = "none",
    _DEP = 1,
  },
  baudrate = baudrateScheme,
  databits = function(_, path)
    local enc = C.getConfOrDefault(s.clonePath(path, -1, "encapsulation"))
    return enc ~= 'ppp' and s.newScheme{
      _T = 'choice',
      _A = {
        5,
        6,
        7,
        8,
      },
      _V = s.choiceVal,
      _H = "rs232Databits",
      _D = 8,
    }
  end,
  stopbits = function(_, path)
    local enc = C.getConfOrDefault(s.clonePath(path, -1, "encapsulation"))
    return enc ~= 'ppp' and s.newScheme{
      _T = 'choice',
      _A = {
        1,
        2,
      },
      _V = s.choiceVal,
      _H = "rs232Stopbits",
      _D = 1,
    }
  end,
  parity = function(_, path)
    local enc = C.getConfOrDefault(s.clonePath(path, -1, "encapsulation"))
    return enc ~= 'ppp' and s.newScheme{
      _T = 'choice',
      _A = {
        "none",
        "even",
        "odd",
      },
      _V = s.choiceVal,
      _H = "rs232Parity",
      _D = "none",
    }
  end,

  ["raw-tcp"] = function(_, path)
    local enc = C.getConfOrDefault(s.clonePath(path, -1, "encapsulation"))
    return enc == 'raw-tcp' and rawTcpScheme{_P = -90}
  end,

  ppp = function(_, path)
    local enc = C.getConfOrDefault(s.clonePath(path, -1, "encapsulation"))
    return enc == 'ppp' and ppp.pppScheme{_P = -90}
  end,

  ["log"] =  s.newScheme{
    _H = "rs232Log",
    _X = function(path)
      path = s.clonePath(path, -1)
      return C.getDaemonLog(path)
    end,
    _R = 'user',
    _P = -100,
  },

  ["raw-access"] = function(_, path)
    local enc = C.getConfOrDefault(s.clonePath(path, -1, "encapsulation"))
    local adm = C.getConfOrDefault(s.clonePath(path, -1, "adm-state"))
    return (enc == 'none' or adm == 'down') and rawAccessScheme
  end,
}
