module(..., package.seeall)
local s = require'scheme'

local clonePath = s.clonePath

local function createStunnelConfFile(conf)
   local ret = false
   local active = false
   local stunConf = {}
   conf = conf or {}
   stunConf[#stunConf + 1] ="# NSG autogenerated configure file for stunnel"
   stunConf[#stunConf + 1] = "foreground=yes"
   stunConf[#stunConf + 1] = "pid=/var/run/stunnel.pid"
  -- if tostring(conf["enable"]) == "true" then
      for k, v in pairs(conf) do
         if k ~= "extra-options" and k ~= "enable" and k ~= "show" and k ~= "tunnels" and nil==(k:match("^_.*$")) then
            stunConf[#stunConf + 1] = k.."="..tostring(v)
         end
      end
      for k, v in pairs(conf["extra-options"] or {}) do
         stunConf[#stunConf + 1] = k.."="..tostring(v)
      end
      for tun,tconf in pairs(conf.tunnels or {}) do
         if tostring(tconf["enable"]) == 'true' then
            active = true
            stunConf[#stunConf + 1] = "["..tun.."]"
            stunConf[#stunConf + 1] = "#"..(tconf.description or "")
            for k, v in pairs(tconf) do
               if ({["extra-options"]=1,["enable"]=1,description=1,[k:match("^_.*$") or ""]=1})[k]==nil then
                  stunConf[#stunConf + 1] = k.."="..tostring(v)
               end
            end
            for k, v in pairs(tconf["extra-options"] or {}) do
               stunConf[#stunConf + 1] = k.."="..tostring(v)
            end
         end
      end
      --end
   if active then
   local f = io.open("/etc/stunnel/stunnel.conf", 'w')
   if f then
      f:write(table.concat(stunConf,"\n").."\n")
      f:close()
      ret = true
   else
      ret = false
   end
   end
   return ret and active
end

local function restartStunnel(path, sch, conf, prevConf)
   local ret
   local defConf = s.addDefault(path, sch, conf, true)

   C.restartDaemon(path)
   ret = createStunnelConfFile(defConf)
   if ret and not s.confCmp(defConf, prevConf)
   then
      if conf -- and tostring(conf["enable"])=="true" 
      then
         C.restartDaemon(
            path,
            {"/usr/sbin/stunnel"},
            {noRestart = false,
             usePopen = true,
             atExit = C.increasingSleep,}
         )
      end
   end
   return true
end

local hostPortValidate = function(val)
                            return ((string.match(val,"^%d+$")) or (string.match(val, "^[%w.-%%]+:%d+$")) or nil) , "[host:]port"
                         end

local stunnelTunnelHttpsScheme = s.newScheme{
   _H = "stunnelTunHTTPS",
   description = s.constantType{_H = "stunnelTunDescription", _P = 100,
                              _D = "autogenerated HTTPS redirector for WEB configuration interface"},
   enable = s.constantType{_H = "stunnelTunnelEnable", _D = true},
   accept = s.constantType{_H = "stunnelTunAccept", _D = "443"},
   client = s.constantType {_H = "stunnelTunClient", _D = "no"},
   connect = s.constantType{_H = "stunnelTunConnect", _D = "80"},
   sslVersion = s.choiceType{
      _H = "stunnelTunVersion",
      _A = {
         {"all", ""},
         {"SSLv2", ""},
         {"SSLv3", ""},
         {"TLSv1", ""},
      },
      _D = "SSLv3",
   },
   cert = s.constantType{_H = "stunnelTunCert", _D = "/etc/stunnel/https_auto.cert"},
   key = s.constantType{_H = "stunnelTunKey", _D = "/etc/stunnel/https_auto.key"},
}

local stunnelTunnelScheme = s.newScheme{
 --[[  _N = function(sch, conf)
           if (conf.client == "yes") then
              return { transparent = 1}
           else
              return { }
           end
        end,
        --]]
   description = s.stringType{_H = "stunnelTunDescription", _P = 100},
   _H = "stunnelTun",
   accept = s.stringType{
      _H = "stunnelTunAccept",
      _V = hostPortValidate,
      _P = 90,
   },
   connect = s.stringType{
      _H = "stunnelTunConnect",
      _V = hostPortValidate,
      _P = 90,
   },
   ["local"] = s.stringType{_H = "stunnelTunLocal"},
   CApath = s.stringType{_H = "stunnelTunCApath"},
   CAfile = s.stringType{_H = "stunnelTunCAfile"},
   CRLpath = s.stringType{_H = "stunnelTunCRLpath"},
   CRLfile = s.stringType{_H = "stunnelTunCRLfile"},
   OCSP = s.stringType{_H = "stunnelTunOCSP"},
   cert = s.stringType{_H = "stunnelTunCert"},
   ciphers = s.stringType{_H = "stunnelTunCiphers"},
   key = s.stringType{_H = "stunnelTunKey"},
   sslVersion = s.choiceType{
      _H = "stunnelTunVersion",
      _A = {
         {"all", ""},
         {"SSLv2", ""},
         {"SSLv3", ""},
         {"TLSv1", ""},
      },
      _D = "SSLv3",
   },
   verify = s.choiceType{
      _H = "stunnelTunVerify",
      _A = {
         {1, ""},
         {2, ""},
         {3, ""},
      },
   },
--[[   transparent = s.newScheme{
      _H = "stunnelTunTransparent",
      _T = 'choice',
      _A = {
         {"yes", ""},
         {"no", ""},
      },
      _D = "no",
      _V = s.choiceVal,
   },
--]]
   client = s.choiceType{
      _H = "stunnelTunClient",
      _A = {
         {"yes", ""},
         {"no", ""},
      },
      _D = "no",
      _P = 90,
      _DEP = 1,
   },
  
   ["enable"] = s.booleanType{
      _H = "stunnelTunnelEnable",
      _D = false,
      _P = 100
   },
  
   ["extra-options"] = s.newScheme{
      _H = "stunnelTunExtraoptions",
      _P = -100,
      _NAME = s.stringType{
         _H = "stunnelTunExtraoptionsOption",
      }
   },
}


stunnelScheme = s.newScheme{
   _H = "stunnel",
   _Z = restartStunnel,
--[[   ["enable"] = s.newScheme{
      _H = "stunnelEnable",
      _T = 'choice',
      _A = {
         {"up", ""},
         {"down", ""},
      },
      _V = s.choiceVal,
      _D = "down",
      _P = 100
   },
--]]
   tunnels = s.newScheme{
      _N = function(sch, conf)
              local httpConf = C.getConfByPath({"services", "http"}) or {}
              if (C.getConfOrDefault({"services", "http","https"}) ~= "enable" or not C.getConfOrDefault({"services", "http","enable"}) ) then
                 return { ["~HTTPS~"] = 1}
              else
                 return { }
              end
           end,
      _H = "stunnelTunnels",
      _P = -100,
      _NAME = stunnelTunnelScheme,
      ["~HTTPS~"] =  stunnelTunnelHttpsScheme,
   },
   debug = s.integerType{
      _H = "stunnelDebug",
      _A = {0,7},
      _D = 5,
   },
   show = s.newScheme{
      _H = "stunnelShow",
      config = s.newScheme{
         _H = "stunnelShowConfig",
         _X = function()
                 return('\n'..s.readpopen("cat /etc/stunnel/stunnel.conf 2>&1"))
              end,
         _R = 'user',
      },
      ["log"] =  s.newScheme{
         _H = "stunnelShowLog",
         _X = function(path)
                 path = s.clonePath(path, -2)
                 return C.getDaemonLog(path)
              end,
         _R = 'user',
      },
   },
}
