module(..., package.seeall)
local s = require'scheme'
local WiFiCommon = require'nsgwifico'
--[[
##################################################################
#####  ADD routines for insertion parameters to hostapd.conf  ####
##################################################################
--]]
----------------------------------------
local function writeOneServer(conf, target) 
   if not conf.address then
      return
   end
   target[#target+1] = "auth_server_addr="..conf.address   
   target[#target+1] = "auth_server_port="..conf.port   
   target[#target+1] = "auth_server_shared_secret=" .. (conf.secret or '""')
end

local function AddWiFiRadius(conf, target)
   if conf and target then
      writeOneServer(conf, target)
      for _, v in ipairs(conf) do
	 writeOneServer(v, target)
      end
   end
end
----------------------------------------
local function AddWiFiWPA(conf, target)
   if conf and target then
      target[#target+1] = '\n'.."#### WPA/IEEE 801.11i configuration ###"

      if conf.wpa2 then
	 wpa = 3 -- both
      else
	 wpa = 1 -- only Draft 3.0
      end

      -- Form WPA parameters --
      
      if conf.method == "personal" then
	 if conf.passphrase and string.len(conf.passphrase) then
	    target[#target+1] = ("%s=%d"):format('wpa', wpa) -- only if pass exist
	    target[#target+1] = "wpa_key_mgmt=WPA-PSK"
	    target[#target+1] = ("%s=%s"):format('wpa_passphrase', conf.passphrase)
	 end
      else
	 target[#target+1] = ("%s=%d"):format('wpa', wpa) 
	 target[#target+1] = "wpa_key_mgmt=WPA-EAP"
      end

      if conf.wpa2 then
	 target[#target+1] = "wpa_pairwise=CCMP TKIP"
      end

      target[#target+1] = "#### End of section ####" .. '\n'
   end
end
----------------------------------------
local function AddWiFiDetail(conf, target)
   if conf and target then
      target[#target+1] = '\n'.."#### IEEE 801.11 related configuration ###"
      for k, v in pairs(conf or {}) do
	 target[#target+1] = 
	    ("%s=%s"):format(k, tostring(v))
      end
      target[#target+1] = "#### End of section ####" .. '\n'
   end
end

----------------------------------------
--[[
##################################################################
#####    CREATE  routines for construction conf-files         ####
##################################################################
--]]
local function createWiFiACLFiles(conf)
   if conf then
      local fa = io.open("/etc/hostapd.accept", 'w')      
      local fd = io.open("/etc/hostapd.deny", 'w')      
      for k, v in pairs(conf or {}) do 
	 if v == 'accept' then
	    fa:write(k .. '\n')
	 else
	    fd:write(k .. '\n')
	 end
      end
      fa:close()
      fd:close()
   end
end
----------------------------------------
function createWiFiConfFile(conf, wlandev)
  local wifiConf = {
     "\n####### hostapd configuration file ########\n",
     "interface=" .. wlandev,
     "driver=nl80211",
     "ctrl_interface=/var/run/hostapd",
     "auth_algs=1",
  }
  wifiConf[#wifiConf+1] = '\n'.."#### configurable AP parameters ###"..'\n'
     
  if conf.ssid then
     wifiConf[#wifiConf+1] = 'ssid='..conf.ssid
  end
  -- Detail parameters of IEEE 802.11 --
  AddWiFiDetail(conf.detail, wifiConf)
  
  if conf and conf.security then 
     -- Security WEP
     wifiConf[#wifiConf+1] = '\n'.."#### Static WEP key configuration ###"
     WiFiCommon.AddWepKeys(conf.security.wep, wifiConf, "wep_default_key")
     wifiConf[#wifiConf+1] = '\n'.."#### End of section ####" .. '\n'
     -- Access Control List (MAC)
     wifiConf[#wifiConf+1] = '\n'.."#### Station MAC address -based authentication ###"
     createWiFiACLFiles(conf.security["mac-address-list"].rules)
     if conf.security["mac-address-list"].policy == 'accept' then
	wifiConf[#wifiConf+1] = "macaddr_acl=0"
     else
	wifiConf[#wifiConf+1] = "macaddr_acl=1"
     end
     if conf.security["mac-address-list"].rules then
	wifiConf[#wifiConf+1] = "accept_mac_file=/etc/hostapd.accept"
	wifiConf[#wifiConf+1] = "deny_mac_file=/etc/hostapd.deny"
     end
     wifiConf[#wifiConf+1] = "#### End of section ####" .. '\n'

     -- Open System Authentication (802.1X)
     wifiConf[#wifiConf+1] ='\n'.."#### IEEE 802.1X and related configuration ###"
     if conf.security.radius.address then -- if where is at least one server ! 
	wifiConf[#wifiConf+1] = "ieee8021x=1"
	if conf.security.radius.own_ip_addr then
	   wifiConf[#wifiConf+1] = "own_ip_addr=" .. conf.security.radius.own_ip_addr
	end
	AddWiFiRadius(conf.security.radius, wifiConf)
     end
     wifiConf[#wifiConf+1] ='\n'.."#### End of section ####" .. '\n'
     -- WPA (IEEE 802.11i/D3.0/RSN)
     AddWiFiWPA(conf.security.wpa, wifiConf)
  end

  local wifiConfText = table.concat(wifiConf, '\n')..'\n'
  local f = io.open("/etc/hostapd.conf", 'w')
  if f then
    f:write(wifiConfText)
    f:close()
  end
end
---------------------------------------- 
local WiFiAccessControlList = s.newScheme {
   _H = "WifiApSecurityAcl",
   
   policy = s.newScheme{
      _H = "WifiApSecurityAclPolicy",
      _T = "choice",
      _A = {
	 {"accept", "allow unless in deny list"},
	 {"deny", "forbidden unless in accept list"},
      },
      _D = 'accept',
      _V = s.choiceVal,
   }, 
   rules = s.newScheme{
      _H = "WifiApSecurityAclRules",
      _NAME = s.newScheme{
	 _H = "WifiApSecurityAclRulesAccess",
	 _T = "choice",
	 _A = {
	    {"accept", "allow"},
	    {"deny", "forbidden"},
	 },
	 _V = s.choiceVal,
      },
      _NAME_TYPE = s.macaddressVal,
   },
}
----------------------------------------
local WiFiRadiusParams = s.newScheme {
   address = s.ipAddressType {
      _H = "WifiApSecurityOpenRadiusParamsAddress",
   },
   port = s.integerType {
      _H = "WifiApSecurityOpenRadiusParamsPort",
      _A = {1000, 65535},
      _D = 1812, },
   secret = s.stringType {
      _H = "WifiApSecurityOpenRadiusParamsSecret",
   },
}

local WiFiRadiusClient = WiFiRadiusParams {
   _H = "WifiApSecurityOpenRadius",
   _LIST = WiFiRadiusParams {
      _H = "WifiApSecurityOpenRadiusParams",
   },

   own_ip_addr = s.ipAddressType{_H = "WifiApSecurityOpenOwnIpAddr", _P = -1, },

}
----------------------------------------
----------------------------------------

local WiFiOpenAuthentication = s.newScheme {
   _H = "WifiApSecurityOpen",

   eapol_version = s.integerType{
      _H = "WifiApSecurityOpenEapolVersion",
      _A = {1, 2},
      _D = 2,
   },


}
local WiFiWPA = s.newScheme {
   _N = function(sch, conf)
	   local mgmt = conf and conf.method or sch.method._D
	   local need = {true, method = true, wpa2 = true, }
	   if mgmt == "personal" then
	      need.passphrase = true
	   end
	   return need
	end,
   _H = "WifiApSecurityWpa",

   method = s.newScheme {
      _H = "WifiStSecurityWpaKeyManagement",
      _T = 'choice',
      _A = {
	 {"personal", "Preshared Key"},
	 {"enterprise", "Extansable Authentication Protocol"},
      },
      _V = s.choiceVal,
      _D = "personal",
      _DEP = 1,
   },

   passphrase = s.stringType{
      _H = "WifiCoSecurityWpaPassphrase",
      _V = WiFiCommon.CheckPSKphrase,
   },
   wpa2 = s.booleanType { _H = "WifiCoSecurityWpaWpa2", _D = false, },
}
----------------------------------------
----------------------------------------

WiFiSecurity = s.newScheme {
   _H = "WifiApSecurity",
   _P = 10,

   ["mac-address-list"] = WiFiAccessControlList,

   wep = WiFiCommon.WiFiSecurityWep,

   open = WiFiOpenAuthentication,

   wpa = WiFiWPA,

   radius = WiFiRadiusClient,
}

----------------------------------------
WiFiGetConnected = s.newScheme {
   _H = "WifiApConnected",
   _P = 10,
   _X = function()
      return(s.readpopen"hostapd_cli all 2>&1")
    end,
}
----------------------------------------
WiFiDetail = s.newScheme {
   _H = "WifiApDetail",
   _P = 5,
   hw_mode = s.newScheme {
      _H = "WifiApDetailHwMode",
      _T = 'choice',
      _A = {
	 {"b", "IEEE 802.11b (up to 11 Mb)",},
	 {"g", "IEEE 802.11g (up to 54 Mb)",},  
      },
      _D = "g",
      _P = 40,
      _V = s.choiceVal,
   },

   max_num_sta = s.integerType{
      _H = "WifiApDetailMaxNumSta",
      _A = {1, 2007},
      _D = 255,
      _P = 30,
   }, 

   channel = WiFiCommon.WiFiDetailChannels,

   beacon_int = s.integerType{
      _H = "WifiApDetailBeaconInt",
      _A = {15, 65535},
      _D = 100,
      _P = 10,
   }, 

}
----------------------------------------
