module(..., package.seeall)
local s = require'scheme'
local WiFiCommon = require'nsgwifico'
local usb = require'nsgusbconf'

----------------------------------------
local function InsertString(target, prefix, object, suffix)
   if object and string.len(object) > 0 then
      target[#target+1] = prefix .. object .. suffix
   else
      target[#target+1] = "# Empty field! >>" .. prefix 
   end
end

local function AddWiFiWPAOpen(conf, target) 
      InsertString(target, "ca_cert=\"/etc/cert/", conf.ca_cert, "\"")
   if conf.eap == "tls" then
      InsertString(target, "eap=", "TLS", "")
      InsertString(target, "client_cert=\"/etc/cert/", conf.client_cert, "\"")
      InsertString(target, "private_key=\"/etc/cert/", conf.private_key, "\"")
      InsertString(target, "private_key_passwd=\"", conf.private_key_passwd, "\"")
   end
   if conf.eap == "peap" then
      InsertString(target, "eap=", "PEAP", "")
      InsertString(target, "password=\"", conf.password, "\"")
      InsertString(target, "phase1=\"", conf.phase1, "\"")
      InsertString(target, "phase2=\"", conf.phase2, "\"")
   end
   if conf.eap == "ttls" then
      InsertString(target, "eap=", "TTLS", "")
      InsertString(target, "anonymous_identity=\"", conf.anonymous_identity, "\"")
      InsertString(target, "password=\"", conf.password, "\"")
      InsertString(target, "phase2=\"", conf.phase2, "\"")
   end
end
----------------------------------------
function createWiFiClientConfFile(conf)
  local wifiConf = {
     "\n####### supplicant configuration file ########\n",
     "ctrl_interface=/var/run/wpa_supplicant\n",
  }
  if conf.connection then     
     for n, v in pairs(conf.connection or {}) do
	local pfx = ""
	if not v["active"] then
	   pfx = "#"
	end
	if v["mode"] == "ad-hoc" then -- before "network"-section
	   wifiConf[#wifiConf+1] = pfx .. "ap_scan=2" -- nessessary to Ad-Hoc! 
	end
	   
	wifiConf[#wifiConf+1] = "\n###### network \"" .. n .. "\" #######"  
	wifiConf[#wifiConf+1] = pfx .. "network={"
	wifiConf[#wifiConf+1] = pfx .. 'ssid="' .. v["ssid"] .. '"'
	wifiConf[#wifiConf+1] = pfx .. ("%s=%d"):format('priority', v["priority"])
	if v["mode"] == "ad-hoc" then
	   wifiConf[#wifiConf+1] = pfx .. "mode=1" -- "0" is default (infrastructure)
	end
-- Use security parameters only "active" connection !
	if v.active and v.security then
	   if v.security == "none" then
	      wifiConf[#wifiConf+1] = "key_mgmt=NONE"
	   end
	   if v.security == "wep" then
	      wifiConf[#wifiConf+1] = "key_mgmt=NONE" -- also NONE !
	      WiFiCommon.AddWepKeys(v.wep, wifiConf, "wep_tx_keyidx")
	   end
	   if v.security == "wpa" and v.wpa then
	      local conf = v.wpa
	      if conf.method == "personal" then
		 wifiConf[#wifiConf+1] = "key_mgmt=WPA-PSK"
		 if conf.passphrase and string.len(conf.passphrase) then
		    wifiConf[#wifiConf+1] = 'psk="' ..  conf.passphrase .. '"'
		 end
	      else
		 wifiConf[#wifiConf+1] = "key_mgmt=WPA-EAP"
		 if conf.identity and string.len(conf.identity) then
		    wifiConf[#wifiConf+1] = 'identity="' ..  conf.identity .. '"'
		 end
		 if conf.open then
		    AddWiFiWPAOpen(conf.open, wifiConf)
		 end
	      end
	      if conf.wpa2 then
		 wifiConf[#wifiConf+1] = "pairwise=CCMP TKIP"
		 wifiConf[#wifiConf+1] = "group=CCMP TKIP"
	      end
	   end
	end
	wifiConf[#wifiConf+1] = pfx .. "}"
     end
  end
  
  local wifiConfText = table.concat(wifiConf, '\n')..'\n'
  local f = io.open("/etc/supplicant.conf", 'w')
  if f then
    f:write(wifiConfText)
    f:close()
  end
end
----------------------------------------
local WiFiWPAOpen = s.newScheme {
   _H = WifiStSecurityWpaOpen,
   _N = function(sch, conf)
	   local eap = conf and conf.eap
	   local need = {true, eap = true, ca_cert = true}
	   if eap == "tls" then
	      need.client_cert = true
	      need.private_key = true
	      need.private_key_passwd = true
	   else
	      need.phase2 = true
	      need.password = true
	      if eap == "peap" then
		 need.phase1 = true
	      else
		 if eap == "ttls" then
		    need.anonymous_identity = true
		 end
	      end
	   end
	   return need
	end,
   
   eap = s.newScheme {
      _H = "WifiStSecurityWpaEap",
      _T = 'choice',
      _A = {
	 {"tls", "TLS"},
	 {"peap", "PEAP"},
	 {"ttls", "TTLS"}, 
      },
      _V = s.choiceVal,
      _D = "tls", 
   },
   
   ca_cert = s.stringType{ _H = "WifiStSecurityWpaOpenCaCert", },

   client_cert = s.stringType{ _H = "WifiStSecurityWpaOpenClientCert", },

   private_key = s.stringType{ _H = "WifiStSecurityWpaOpenPrivateKey", },

   private_key_passwd = s.stringType{ _H = "WifiStSecurityWpaOpenPrivateKeyPasswd", },

   phase1 = s.stringType{ _H = "WifiStSecurityWpaOpenPhase1", },

   phase2 = s.stringType{ _H = "WifiStSecurityWpaOpenPhase2", },

   password = s.stringType{ _H = "WifiStSecurityWpaOpenPassword", },

   anonymous_identity = s.stringType{ _H = "WifiStSecurityWpaOpenAnonymousIdentity", },


}
----------------------------------------
local WiFiWPA = s.newScheme {
   _N = function(sch, conf)
	   local mgmt = conf and conf.method or sch.method._D
	   local need = {true, method = true, wpa2 = true, }
	   if mgmt == "personal" then
	      need.passphrase = true
	   else
	      need.identity = true
	      need.open = true
	   end
	   return need
	end,

   _H = "WifiStSecurityWpa",
   
   method = s.newScheme {
      _H = "WifiStSecurityWpaKeyManagement",
      _T = 'choice',
      _A = {
	 {"personal", "Preshared Key"},
	 {"enterprise", "Extansable Authentication Protocol"},
      },
      _V = s.choiceVal,
      _D = "personal",
      _DEP = 1,
   },
   
   passphrase = s.stringType{
      _H = "WifiCoSecurityWpaPassphrase",
      _V = WiFiCommon.CheckPSKphrase,
   },

   identity = s.stringType{
      _H = "WifiStSecurityWpaIdentity",
      _V = CheckPSKphrase,
   },

   wpa2 = s.booleanType { _H = "WifiCoSecurityWpaWpa2", _D = false, },
   
   open = WiFiWPAOpen,

}
----------------------------------------

WiFiConnection = s.newScheme {
   _H = "WifiStConnection",
   _P = 16,
   _NAME = s.newScheme {
      _N = function(sch, conf) 
	      local need = {
		 true, 
		 active = true, 
		 priority = true, 
		 ssid = true, 
		 mode = true, 
		 security = true,
	      }
	      if conf.security and conf.mode then
		 if conf.mode == "infrastructure" then
		    need[conf.security] = true -- get according "security"
		 else
		    need["wep"] = true   -- only WEP enable !
		 end
	      end
	      return need
	   end,
      
      active = s.booleanType { _H = "WifiStConnectionActivity", _D = true, },

      priority = s.integerType { _H = "WifiStConnectionPriority", _D = 0, _A = {0, 256}, _P = -5, },

      ssid = s.stringType{_H = "WifiStConnectionSsid",  _P = 92, _D = "", },
 
      mode = s.newScheme {
	 _H = "WifiStConnectionMode",
	 _T = 'choice',
	 _A = {
	    {"infrastructure", ""},
	    {"ad-hoc", ""},
	 },
	 _V = s.choiceVal,
	 _D = "infrastructure",
	 _DEP = 1,
      },
      
      security = s.newScheme {
	 _H = "WifiStConnectionSecurity",
	 _T = 'choice',
	 _A = {
	    {"none", "Without any authentication/authorization",},
	    {"wep", "Wireless Equivalent Security (Static WEP)"},
	    {"wpa", "Wi-Fi Protect Access (IEEE-802.11i)"},
	 },
	 _V = s.choiceVal,
	 _D = "none",
	 _DEP = 1,
      },
      
      wep = WiFiCommon.WiFiSecurityWep,

      wpa = WiFiWPA, -- temporary ? (may be combined with AP into "common" ?)
      
   },

}
----------------------------------------
local function WifiScanAtStart(oper)
   C.threadSleep(1)  -- Now is simple delay   
end

local launchWiFiClScan = function(path, sch, conf, sesNum) 
  local sch --, conf = getParent(path, sesNum)
  local path = s.clonePath(path, -1)
  local wlandev = usb.port2netdev[path[#path]]
  local com = {"iw", wlandev, "scan", }

  if not wlandev then
     return "No WLAN device"
  end

  s.readpopen("ifconfig " .. wlandev .. " up 2>&1")

  C.restartDaemon(path, com,
    {sesNum = sesNum, noTimestamp = true, noRestart = true, atStart = WifiScanAtStart, }) 
end

----------------------------------------

WiFiScan = s.newScheme {
   _H = "WifiStScan",
   _P = 15,
   _X = launchWiFiClScan,
   _XI = true,
}

