#!/usr/bin/lua
local uitcpVersion = '1.01' -- Don't change - automodified string
package.path = "/etc/nsg/share/lua/5.1/?.lua;"..package.path
package.cpath = "/etc/nsg/lib/lua/5.1/?.so;"..package.cpath

local nsg = require'nsg'
local progname = arg[0]
local optsDescr = {
  ("%s [OPTION ...] CONFIG_FILE"):format(arg[0]),
  "",
  "OPTIONS:",
  help = {0, "display this help and exit"},
  h = 'help',
  n = {0, "Run from nsgconfd"},
}
local arg = nsg.getopt({...}, optsDescr)
arg[0] = progname

require'signal'.signal('TERM', function()signal.raise'INT'end)
require"nsginit"
require"serialize"
local deepcopy = serialize.deepcopy
local serialize = serialize.serialize
local os = os
local type = type
local io = io
local string = string
local pcall = pcall
local tostring = tostring
local math = math
local ipairs = ipairs
local pairs = pairs
local table = table
local coroutine = coroutine
local loadstring = loadstring
local unpack = unpack

local RECEIVE_BLOCK_SIZE = 2^12
local MAX_RECEIVE_BLOCKS = 2
local WINDOW_SIZE = 8
local MultiTunnelLC = 255

local uiError
local countedPcall do
  local function countHook(...)
    local info = debug.getinfo(2)
    if info then 
      uiError(("May be deadlock in %s:%d"):format(info.short_src, info.currentline))
    end
    error("May be deadlock in function")
  end
  function countedPcall(func, ...)
    debug.sethook(countHook, '', 10*10^6)
    local res, err =  pcall(func, ...)
    debug.sethook()
    return res, err
  end
end

-- Find uitcp path
local f = loadfile'/etc/nsgSysConfig'
local nsgc = {}
if f then
  local res, config = countedPcall(f)
  if res and type(config) == 'table' then
    nsgc = config
  end
end

local function splitFileName( filename )
  local fn = string.gsub( filename, "%\\", "/" )
  if not string.find( fn, "/", 1, true ) then return "", fn end
  local s, e, p, n = string.find( fn, "(.-%/)([^%/]-)$" )
  return p, n
end
local uitcpPath, uitcpName = splitFileName( arg[0] )

local uitcpBinPath = (nsgc.uitcpBinPath or uitcpPath)..'/'
package.path = uitcpBinPath..'?.lua;'..package.path

local conf -- config table
local mainTunnel

local function readpopen(prog)
  local f = io.popen(prog)
  local s
  if f then
    s = f:read("*a")
    f:close()
  end
  return s
end
--------------------------------------------------------------------------------
-- log and debug stuff

local logFileMaxSize = 1024 * 1024
local log = 4 -- initial log level = INFO 
local trace
local logFileName = '/var/log/uitcp'
local logFileSize
local logFile
if not arg.n then
  logFile = io.open(logFileName, "a+")
end
if logFile then
  logFileSize = #logFile:read("*a")
else
  logFileName = false
  logFile = io.output()
end

local function saveLogFile(len)
  if logFileName then
    logFileSize = logFileSize + len
    if logFileSize > logFileMaxSize then
      os.rename(logFileName, logFileName..".old")
      logFile = io.open(logFileName, "w")
      if logFile then
        logFileSize = 0
      else
        log, trace = nil, nil
      end
    end
  end
end
local function DBG(...) if dbg then print(debug.getinfo(2,'l').currentline, ...) end
end

local logLevelToNum = {
  ['fatal'] = 1,
  ['error'] = 2, 
  ['warn'] = 3, 
  ['info'] = 4,
  ['debug'] = 5,
}

local logLevelString = {
  'FATAL',
  'ERROR', 
  'WARN', 
  'INFO',
  'DEBUG',
}

local sqlEnvironment
local sqlLogConnection
local envHostname = readpopen"hostname":sub(1, -2)

local LOG, uiFatal, --[[uiError,]] uiWarn, uiInfo, uiDebug,
uiFatalT, uiErrorT, uiWarnT, uiInfoT, uiDebugT do

  local function sqlLog(timestamp, level, tunnel, logmes)
    local ret, err = sqlLogConnection{
      date = timestamp,
      level = level,
      hostname = envHostname,
      tunnelname = tunnel,
      message = logmes,
    }
    if not ret then
      local sqlLogConnectionSav = sqlLogConnection
      sqlLogConnection = nil
      uiError(err)
      sqlLogConnection = sqlLogConnectionSav
    end
    return ret
  end

  local function uiLog(level, tunnel, ...) 
    if log and level <= log then -- collectgarbage()--print("LOG",...)
      local e, logmes = pcall(string.format, ...)
      if not e then
        logmes = logmes.."\n"..tostring(...).."\n"
      end
      local timestamp = os.date('%F %T')
      if sqlLogConnection and sqlLogConnection:isReady() and level <= 5 then
        if sqlLog(timestamp, logLevelString[level], tunnel, logmes) then
          return
        end
      end
      local tunstr = tunnel and '['..tunnel..']' or ''
      local s = string.format("%s %-5s:%s%s\n",
        timestamp, --collectgarbage("count"),
        logLevelString[level],
        tunstr,
      logmes)
      logFile:write(s)
      logFile:flush()
      saveLogFile(#s)
    end 
  end
  function LOG(...)
    uiLog(4, nil, ...)
  end

  function uiFatal(...)
    uiLog(1, nil, ...)
  end
  function uiError(...)
    uiLog(2, nil, ...)
  end
  function uiWarn(...)
    uiLog(3, nil, ...)
  end
  function uiInfo(...)
    uiLog(4, nil, ...)
  end
  function uiDebug(...)
    uiLog(5, nil, ...)
  end

  function uiFatalT(tunnel, ...)
    uiLog(1, tunnel, ...)
  end
  function uiErrorT(tunnel, ...)
    uiLog(2, tunnel, ...)
  end
  function uiWarnT(tunnel, ...)
    uiLog(3, tunnel, ...)
  end
  function uiInfoT(tunnel, ...)
    uiLog(4, tunnel, ...)
  end
  function uiDebugT(tunnel, ...)
    uiLog(5, tunnel, ...)
  end
end

local function hex_dump(tunName, buf, ...)
  if trace then
    if type(trace) ~= "number" then trace = #buf end
    local n = math.min(#buf, trace)
    local e, logmes = pcall(string.format, ...)
    if not e then
      logmes = logmes.."\n"..tostring(...).."\n"
    end
    local s = string.format("%s TRACE:[%s]%s(%d bytes)\n",
      os.date('%F %T'), tunName or "NONE", logmes, #buf)
    logFile:write(s)
    for byte=1, n, 16 do
      local chunk = buf:sub(byte, byte+15)
      logFile:write(string.format('%08X  ',byte-1))
      chunk:gsub('.', function (c) logFile:write(string.format('%02X ',string.byte(c))) end)
      logFile:write(string.rep(' ',3*(16-#chunk)))
      logFile:write(' ',chunk:gsub('[%c\255]','.'),"\n")
    end
    logFile:flush()
    saveLogFile(#s + math.ceil(n/16)*75)
  end
end

--------------------------------------------------------------------------------
-- Read scheme and config

uiInfo("============= uiTCP-%s beginning =============", uitcpVersion or 'work')
local scheme = require("scheme")
if arg.n then
  local origConfCmp = scheme.confCmp
  scheme.confCmp = function(s, a, b)
    return origConfCmp(a, b)
  end
  uitcpScheme = {__index = function()return uitcpScheme end}
  setmetatable(uitcpScheme, uitcpScheme)
else
  uitcpScheme = assert(loadfile(uitcpBinPath..'uitcpScheme.lua'))()
end

local configFile = arg[1] or 
(nsgc.uitcpConfigPath and nsgc.uitcpConfigPath..'/config') or 
"/etc/uitcp/config"
local f, e = loadfile(configFile)
if not f then
  uiError("Error loading config file %s: %s", configFile, e)
else
  local res
  res, conf = countedPcall(f)
  if not res then
    uiError("Error in config file %s: %s", configFile, conf)
  else
    if type(conf) ~= 'table' then
      uiWarn("May be obsolete format of config file")
      conf = uitcpConf or {}  -- table uitcpConf may be defined in configFile 
      -- (version <= 0.10)
    end
  end
end
if not conf then
  uiWarn("Can't find config. Start with default config")
  conf = {}
end

if not arg.n then
  scheme.addDefault(uitcpScheme, conf)
  if not uitcpConf then -- obsolete ver < 0.11
    scheme.stripConfig(uitcpScheme, conf)
    local res, err = scheme.confValidate(uitcpScheme, conf)
    if not res then
      uiError("Bad configuration\n%s", serialize(err, 0))
      uiWarn("Bad values will replace by default")
      scheme.confValidate(uitcpScheme, conf, true)
      --error("Bad configuration")
    end
  end
end

require"classlib"
--------------------------------------------------------------------------------
-- Logging
class.SqlConnection()

function SqlConnection:__init(env, param)
  self.env = env
  self.sourcename = param.sourcename
  self.username = param.username
  self.password = param.password
  self.hostname = param.hostname
  self.port = param.port
  self.tablename = param.tablename
  self.readyTime = os.time()
  self.readyTimeInterval = 1
  self.key = {}
  local fields = {}
  if type(param.fields) == 'table' then
    for k, v in pairs(param.fields) do
      self.key[#self.key+1] = k
      fields[#fields+1] = v
    end
  end
  if self.tablename and #fields > 0 then
    local flds = table.concat(fields, ',')
    local ptrn = (",'%s'"):rep(#fields):sub(2)
    self.insert  = string.format("INSERT INTO %s (%s) VALUES (%s)",
    self.tablename, flds, ptrn)
    self:connect()
  else
    uiError"No tablename or fields define for insert to SQL database"
  end
end

function SqlConnection:connect()
  if not self.env then
    self.readyTime = os.time() + 60 * 60 * 24
    return nil, "SQL:Can't connect: no sql enviroment"
  end
  local err
  self.conn, err = self.env:connect(self.sourcename, self.username, 
  self.password, self.hostname, self.port)
  if not self.conn then
    self.readyTime = os.time() + self.readyTimeInterval
    self.readyTimeInterval = math.min(self.readyTimeInterval * 2, 512)
    return nil, 'sql:connect: '..err
  end
  self.readyTime = os.time()
  self.readyTimeInterval = 1
  return true
end

function SqlConnection:__call(t)

  if not self.insert then
    return nil, "No fields for insert"
  end
  local list = {}
  for k, v in ipairs(self.key) do
    list[k] = (t[v] or ''):gsub("[\\']","\\%1")
  end
  local ret, err
  if self.conn then
    ret, err = self.conn:execute(string.format(self.insert, unpack(list)))
    if ret then
      return true
    end
    self.conn = nil
  end
  ret, err = self:connect()
  if ret and self.conn then
    ret, err = self.conn:execute(string.format(self.insert, unpack(list)))
    return ret, err
  else
    return nil, err
  end
end

function SqlConnection:isReady()
  return self.readyTime <= os.time()
end

function SqlConnection:close()
  if self.con then self.conn:close() end
end

-- Create SQL environment

local dbToServer = {
  PostgreSQL = 'postgres',
  ODBC = 'odbc',
  MySQL = 'mysql',
  SQLite = 'sqlite',
  Oracle = 'oci8',
  ADO= 'ado',
}
local function createSqlEnvironment(sql)
  if sql and sql.database and sql.sourcename then
    local srvname = dbToServer[sql.database]
    if not srvname then
      uiError("SQL:Bad database server %s", sql.database)
    else
      local res, luasql = pcall(require, 'luasql.'..srvname)
      if not res then
        uiError("SQL:Can't find module %s", 'luasql.'..srvname)
      else
        local res, env = pcall(luasql[srvname])
        if not res then
          uiError("SQL:Can't create environment for %s", srvname)
        else
          sqlEnvironment = env
          uiInfo("%s: Create environment", sql.database)
        end
      end
    end
  end
end


--------------------------------------------------------------------------------



-- reopen log file
local function reOpenLogFile()
  uiInfo("Reopen log file. Next output may be in other file")
  if sqlLogConnection then 
    sqlLogConnection:close() 
    sqlLogConnection = nil
  end
  if sqlEnvironment then 
    sqlEnvironment:close() 
    sqlEnvironment = nil
  end
  if logFile ~= io.output() then logFile:close();  logFile = nil end
  trace = nil
  log = nil
  logFile = nil
  if conf.log and type(conf.log) ~= 'table' then -- for old config VER <= 0.10
    conf.log = {level = 'info'}
  end
  if conf.log then
    log = logLevelToNum[conf.log.level] or 4 -- INFO by default
    trace = conf.log.trace
    logFileMaxSize = conf.log.logFileMaxSize or 1024*1024
    if type(conf.log.logFile) == "string" then
      if conf.log.logFile ~= 'stdout' then
        logFile = io.open(conf.log.logFile, "a+")
      end
      if logFile then
        logFileName = conf.log.logFile
        logFileSize = #logFile:read("*a")
      end
    end
    if not logFile then logFile = io.output() end
    if conf.log.sql then
      createSqlEnvironment(conf.log.sql)
      sqlLogConnection = SqlConnection(sqlEnvironment, conf.log.sql)
    end
  end
  if not logFile then log, trace = nil, nil end
end
reOpenLogFile()
uiInfo("============== uiTCP-%s started ==============", uitcpVersion or 'work')
--------------------------------------------------------------------------------
local socket = require"socket"
local ssl = require"ssl"
local task = require'task'

--------------------------------------------------------------------------------
local hdr = {
  lenLow = 1,
  lenHigh = 2,
  dstLC = 3,
  srcLC = 4,
  sn = 5,
  rn = 6,
  len = 6,
}

local threads = {}
local sleepThr = {}
local tunnels = {}
local stats = {}
-- local function definition
local createConnection
local openTcpSocket
local createRecvDataThread
local connectionCreated
local startSslHandshake
local createLocalListener
local createPipe
local SHUTDOWN
--------------------------------------------------------------------------------

local ip = "/usr/sbin/ip"
local function isInterfaceUp(int)
  if type(int) == 'string' then
    return string.find(readpopen(ip.." link show "..int), "UP")
  else
    return true
  end
end

--------------------------------------------------------------------------------
-- init timer
local timerList = {}

local function addTimer(tmo, func, ...)
  tmo = os.time() + tmo
  local timer = {time = tmo, func = func, arg = {...}}
  for i, t in ipairs(timerList) do
    if t.time > tmo then
      table.insert(timerList, i, timer)
      return timer
    end
  end
  table.insert(timerList, timer)
  return timer
end

local function delTimer(timer)
  if timer then
    for i, t in ipairs(timerList) do
      if t == timer then table.remove(timerList, i) end
    end
  end
end

local function isTimer(timer)
  if timer then
    for i, t in ipairs(timerList) do
      if t == timer then return true end
    end
  end
  return false
end

local function getTimer()
  return timerList[1]
end

local function popTimer()
  return table.remove(timerList, 1)
end

local function execTimer(timer)
  timer.func(unpack(timer.arg))
end

local function timerHandler()
  local t = os.time()
  while true do
    local timer = getTimer()
    if not timer then return end
    if timer.time <= t then
      execTimer(popTimer())
    else return end
  end
end

local MAX_TMO = 5
local function nextTmo()
  local t = os.time()
  local timer = getTimer()
  if not timer or timer.time - t > MAX_TMO then 
    return MAX_TMO
  elseif timer.time - t < 0 then
    return 0
  end
  return timer.time - t
end

--------------------------------------------------------------------------------
local numOf15MinIntervalInDay = 4*24
local sizeOf15MinInterval = 60*15

local function getCurrentTimeInterval(time)
  return math.floor(time / sizeOf15MinInterval)
end

local function timeIntervalToggle(stat, i, time)
  for k, v in pairs(stat) do
    if type(k) == "number" and k < i - numOf15MinIntervalInDay then
      stat[k] = nil
    end
  end
  stat.currentTimeInterval = i
end

local function getTimeInterval(stat)
  local time = os.time()
  local i = getCurrentTimeInterval(time)
  if not stat[i] or stat.currentTimeInterval ~= i then
    timeIntervalToggle(stat, i, time)
    stat[i] = {
      recvBytes = 0,
      sendBytes = 0,
      recvPack = 0,
      sendPack = 0,
      sendErr = 0,
    }
  end
  return i
end

local function initTunnelStatsEntry(tun, mode, conf)
  tun.mode = mode
  tun.currState = "dead"
  tun.currStateTime = os.time()
  tun.state = {}
  tun.links = {}
  tun.peerData = {}
  tun.recvBytes = 0
  tun.sendBytes = 0
  tun.recvPack = 0
  tun.sendPack = 0
  tun.sendErr = 0
  if conf then 
    tun.description = conf.description
  end
  local time = os.time()
  tun.currentTimeInterval = getCurrentTimeInterval(time) - 1
end

local function addTunnelStatsEntry(name, mode, conf)
  if type(name) ~= "string" or #name == 0 then
    --DBG("Tunnel name not a string or empty string", tostring(name))
  elseif not stats.tun[name] then
    stats.tun[name] = {}
    initTunnelStatsEntry(stats.tun[name], mode, conf)
  elseif conf then
    stats.tun[name].description = conf.description
  end
end

local function initTunnelStats()
  stats = {}
  stats.startTime = os.time()
  stats.tun = {}

  if conf.tunnel then
    addTunnelStatsEntry(conf.tunnel.name, "client", conf.tunnel)
  end
  if conf.tunnels then
    for k, v in pairs(conf.tunnels) do
      addTunnelStatsEntry(k, "client", v)
    end
  end
  if conf.clients then
    for k, v in pairs(conf.clients) do
      addTunnelStatsEntry(k, "server", v)
    end
  end
end

--------------------------------------------------------------------------------
local function stopRecv(th)
  if th then sleepThr[th] = true end
end

local function startRecv(th)
  sleepThr[th] = nil
  if coroutine.status(th) ~= "dead" then
    for i, t in ipairs(threads) do
      if t == th then
        --DBG"allready exist"
        return
      end
    end
    table.insert(threads, th)
  end
end

--------------------------------------------------------------------------------
class.Fifo()

function Fifo:__init()
  self.first = 0
  self.last = -1
end

function Fifo:push(v)
  local last = self.last + 1
  self.last = last
  self[last] = v
  return last - self.first + 1
end

function Fifo:pushFirst(v)
  local first = self.first - 1
  self.first = first
  self[first] = v
end

function Fifo:pop()
  local first = self.first
  if first > self.last then -- FIFO is empty
    return nil, 0
  end
  local v = self[first]
  self[first] = nil
  self.first = first + 1
  return v, self.last - first
end

function Fifo:isEmpty()
  if self.first > self.last then
    return true
  else return false end
end

--------------------------------------------------------------------------------
-- Auxiliary functions

local function x(s, nodebug)
  local ret = type(s) == 'string' and os.execute("("..s..")>/dev/null 2>&1")
  if ret ~= 0 and not nodebug then
    uiError("Error executing script:\n%s", tostring(s))
  end
  return ret
end

local ipAddrTable = {}
local function ipAddrAdd(addr)
  local n = ipAddrTable[addr]
  if n then
    ipAddrTable[addr] = n + 1
  else
    ipAddrTable[addr] = 1
  end
  -- in any case add address to lo
  local xcom = ip.." addr add "..addr.."/32 dev lo"
  uiDebug(xcom)
  x(xcom, n)
end
local function ipAddrDel(addr)
  local n = ipAddrTable[addr]
  if n then
    if n <= 1 then
      ipAddrTable[addr] = nil
    else
      ipAddrTable[addr] = n - 1
      return
    end
  end
  local xcom = ip.." addr del "..addr.."/32 dev lo"
  uiDebug(xcom)
  x(xcom)
end


--------------------------------------------------------------------------------
class.Connection()

function Connection:__init(initial)
  --DBG"Connection __init"
  self.sock = initial.sock
  self.server = initial.server
end

function Connection:__call(recvData)
  --DBG("receive data", self.sock, recvData)
end

function Connection:getSock()
  return self.sock
end

function Connection:setSock(s)
  self.sock = s
end

function Connection:isServer()
  --DBG("isServer()", self.server)
  return self.server
end

function Connection:setThread(co)
  self.thread = co
end

function Connection:getThread(co)
  return self.thread
end

function Connection:closeSock()
  if self.sock then
    self.sock:close()
    --self.sock = nil -- TODO
  end
end

--------------------------------------------------------------------------------
local function elems (l)
  local n = 0
  return function (l)
    n = n + 1
    if n <= #l then
      return l[n]
    end
  end,
  l, true
end

local function createPemFileName(mode, key)
  if key == 'key' then
    return mode..'key.pem'
  elseif key == 'certificate' then
    return mode..'.pem'
  elseif key == 'cafile' then
    return 'root.pem'
  else
    return nil
  end
end

local dfltssl = {path = '/etc/uitcp/certs'}
local function findPemFile(ssl, key, name)
  local pemFileName, pemFilePath
  for sslconf in elems{ssl, conf.ssl or {}, dfltssl} do
    if sslconf[key] then
      return sslconf[key]
    elseif sslconf.path then
      pemFileName = createPemFileName(ssl.mode, key)
      if not pemFileName then return nil end
      for pemFilePath in elems{sslconf.path..'/'..name..'/'..pemFileName,
      sslconf.path..'/'..pemFileName} do
        local f = io.open(pemFilePath)
        --print (pemFilePath)
        if f then
          f:close()
          return pemFilePath
        end
      end
    end
  end
  return nil
end

local function set2list(t) -- work with new(nsgxonfd) and old foemat
  uiDebug(serialize(t))
  local res = {}
  for k, v in pairs(t) do
    if type(k) == 'number' then
      res[#res] = v
    end
    if v == true then
      res[#res+1] = k
    end
  end
  uiDebug(serialize(res))
  return res
end

local function createSslParamTable(ssl, name)
  ssl = ssl or {}
  if type(ssl) ~= 'table' or type(name) ~= 'string' then return nil end
  local auxssl
  if type(conf.ssl) == 'table' then
    auxssl = conf.ssl
  else
    auxssl = {}
  end
  local ssl = deepcopy(ssl)
  ssl.mode = conf.mode
  ssl.protocol = ssl.protocol or auxssl.protocol or "sslv3"
  ssl.key = findPemFile(ssl, 'key', name)
  ssl.certificate = findPemFile(ssl, 'certificate', name)
  ssl.cafile = findPemFile(ssl, 'cafile', name)
  ssl.capath = ssl.capath or auxssl.capath or 
  (ssl.path and ssl.path.."/capath") or
  (auxssl.path and auxssl.path.."/capath") or 
  '/etc/uitcp/certs/capath'
  ssl.password = ssl.password or auxssl.password
  ssl.verify = set2list(ssl.verify or auxssl.verify)
  ssl.options = set2list(ssl.options or auxssl.options)
  ssl.ciphers = ssl.ciphers or auxssl.ciphers
  ssl.depth = ssl.depth or auxssl.depth
  ssl.certcheck = ssl.certcheck or auxssl.certcheck
  uiDebugT(name, "SSL parameters:\n%s", serialize(ssl, 0))
  return ssl
end

local function getTunnelConfig(name)
  if conf.tunnel and conf.tunnel.name == name then
    return conf.tunnel
  else
    return conf.tunnels and conf.tunnels[name]
  end
end

local function getHostPort(hp)
  if hp then
    return hp:match("(.*):(.*)")
  end
end
--------------------------------------------------------------------------------
class.TunnelConnection(Connection)

function TunnelConnection:__init(initial)
  self.Connection:__init(initial)
  --DBG"TunnelConnection __init"
  self.state = "init"
  self:initTunnelConfig(initial)
  self.ssl = createSslParamTable(initial.ssl, self.name)
  self.nat = initial.nat
  self.listenerSocks = {}
  self:init()
  if self:isServer() then
    local c = self:getSock()
    local h, p = c:getsockname()
    local ph, pp = c:getpeername()
    uiInfo("Accept connection %s:%d>>%s:%d", ph, pp, h, p)
    self:updateKeepTimer(30, self.initialWatchdog, self)
  elseif self.name and initial.links then
    if tunnels[self.name] then
      uiWarn("Tunnel "..self.name.." already exist. Destroy it")
      tunnels[self.name]:destroy()
    end
    tunnels[self.name] = self
    self:initTunnelLinks(initial.links)
    if not self.linksPriority then self.linksPriority = initial.linksPriority end -- obsolete 0.12
    self:connect()
  end
  self.conType = 'tunnel'
end

function TunnelConnection:initTunnelConfig(initial)
  self.name = initial.name
  self.host = initial.host -- TODO obsolete
  self.port = initial.port -- TODO obsolete
  self.servers = initial.servers
  if not self.servers then -- TODO when no host or port
    if self.host and self.port then
      self.servers = {}
      self.servers[1] = self.host..':'..self.port
    end
  end
  self.numOfServers = initial.numOfServers or 1
  self.currentServer = 1
  self.window = initial.window or WINDOW_SIZE
  if self.window < 1 or self.window > 128 then self.window = WINDOW_SIZE end
  self.keepTmo = initial.keepalive
  self.keepAttempt = initial.attempt
  self.connectTmo = initial.connectTmo
  self.connectAttempt = initial.connectAttempt
  self.multiTunnelSocket = initial.multiTunnelSocket
end

function TunnelConnection:initTunnelLinks(linksConf)
  self.links = deepcopy(linksConf)
  for _, l in ipairs(self.links) do

    if not l.servers then --TODO obsolete
      if l.host or l.port then
        l.host = l.host or self.host
        l.port = l.port or self.port
        l.servers = {}
        l.servers[1] = l.host..':'..l.port
      end
    end

    l.servers = l.servers or self.servers
    l.description = l.description or l.dev or "UNKNOWN"
    l.currDescr = l.description
  end
  self.linkN = #self.links
  self.linkLast = self.linkN
  self.linkCurrent = self.linkN
  self.linksPriority = self.links.priority
  self.linksPriorityTmo = tonumber(self.linksPriority) or math.huge
  self.linksPriorityIdleTime = self.links["idle-time"]
  self.linksPriorityTmo = math.abs(self.linksPriorityTmo)
end

function TunnelConnection:init()
  self.state = "init"
  self.s = {} -- table for slave connection
  self.sended = {}
  self.pending = Fifo()
  self.recvBuf = ""
  self.sn = 0
  self.rn = 0
  self.an = 0
  self.rr = 0
  self.connectErr = 0

  self.lastRecvTime = os.time()
  self:stateChanged"dead"
end

function TunnelConnection:getLinkMode()
  if self.links[self.linkCurrent].dualSim then
    local procf = io.open("/proc/sys/nsg/port/"..self.links[self.linkCurrent].dev.."/.mode")
    if procf then
      local s = procf:read("*a")
      procf:close()
      return '-'..s
    end
  end
  return ""
end

local function verToString()
  return uitcpVersion and string.format('%q', uitcpVersion) or 'nil'
end

function TunnelConnection:updateKeepTimer(...)
  if self.keepTimer then delTimer(self.keepTimer ) end
  self.keepTimer = addTimer(...)
end

function TunnelConnection:delKeepTimer()
  if self.keepTimer then
    delTimer(self.keepTimer )
    self.keepTimer = nil
  end
end

function TunnelConnection:sendCommandNew()
  local com = string.format(
    'new(%q,%q,%d,%d,%d,%d,%s)',
    self.name,
    tostring(self.links[self.linkCurrent].currDescr)..
    self:getLinkMode(),
    self.window,
    self.keepTmo,
    self.keepAttempt,
    self.connectTmo * (self.connectAttempt+1),
    verToString()
  )
  self:send(com)
  self:updateKeepTimer(self.keepAttempt * self.keepTmo,
    self.newack,
  self, nil, "Timeout exhausted")

end

function TunnelConnection:reinit()
  self:allSlave('destroy', true)
  self:init()
  if not self:isServer() then
    self:sendCommandNew()
    self:initSession()
  end
end

function TunnelConnection:initSession()
  self.state = "waitcon"
  self.keepSeq = 0
  self.keepErr = 0
end

function TunnelConnection:new(name, link, w, tmo, attempt, watchdog, ver)
  if SHUTDOWN then
    self:send('newack()')
    uiWarnT(name, "New tunnel rejected: uiTCP in SHUTDOWN mode")
    return
  end
  if uitcpVersion and ver ~= uitcpVersion then
    uiInfoT(name, "Client has different from server version(%s)", ver)
  end
  if self.state == "init" or self.state == "disc" then
    if tunnels[name] then
      uiWarnT(name, "Tunnel already exist. Destroy it")
      tunnels[name]:destroy()
    end
    if conf.clients and conf.clients[name] then
      tunnels[name] = self
      self.name = name
      self.nat = conf.clients[name].nat
      self.ssl = createSslParamTable(conf.clients[name].ssl, name)
      createLocalListener(self, conf.clients[name].localListener)
      self:initRawSockets(conf.clients[name].socket)
      self.multiTunnelSocket = conf.clients[name].multiTunnelSocket
      multiTunnelSockets:addTunnel(self)
      self.link = link
      self.window = w
      self.keepTmo = tmo
      self.keepAttempt = attempt
      self.watchdogTmo = watchdog
      self:send(string.format('newack(%q,%s)', self.name, verToString()))
      self.host, self.port = self:getSock():getpeername()
      self:stateChanged"dead"
      uiInfoT(name, "New tunnel initiated from %s", self.host)
      self:opened()
    else
      self:send('newack()')
      uiInfoT(name, "New tunnel rejected: no such client in configuration")
    end
  else
    error("Get new() in not init or disc state")
  end
end

local updateConf = {
  port = conf.update and conf.update.port or 50003,
  action = "exit",
}
local runUpdateProcedure do
  local updateTaskControl
  local updateAttempt
  local taskId
  local maxUpdateAttempt = 10
  local retryTmo = 60
  local startTmo = 10
  local tmo = 1
  local function startUpdateTask(ver)
    if not taskId or not task.isrunning(taskId) then
      taskId = task.create(uitcpBinPath..'tools/update.lua',
        {task.id(), 'uitcp-'..ver, "127.0.0.1", updateConf.port})
      if taskId <= 0 then
        uiError("Can't create update task: %d", taskId)
        taskId = nil
      else
        uiInfo('Update task is created: %d', taskId)
        addTimer(startTmo, updateTaskControl, ver)
      end
    end
  end
  function updateTaskControl(ver)
    if taskId and task.isrunning(taskId) then
      addTimer(tmo, updateTaskControl, ver)
    else
      if updateAttempt then
        local mes = task.receive(0)
        if mes and mes == "updated" then
          uiInfo("Update is finished")
          taskId, updateAttempt = nil, nil
          --TODO restart
          error("Restart")
        else
          uiError("Update error: "..(mes or "unknown").."(attempt %d)", updateAttempt)
          if updateAttempt >= maxUpdateAttempt then
            uiWarn("There are %d unsuccessful attempts to update client. "..
              "Stop updating.",
            maxUpdateAttempt)
            taskId, updateAttempt = nil, nil
          else
            updateAttempt = updateAttempt + 1
            addTimer(retryTmo, startUpdateTask, ver)
          end
        end
      else  -- First call
        updateAttempt = 1
        startUpdateTask(ver)
      end
    end
  end
  function runUpdateProcedure(ver)
    if updateAttempt then
      updateAttempt = 1
    else
      addTimer(10, updateTaskControl, ver)
    end
  end
end

function TunnelConnection:newack(name, ver)
  self:delKeepTimer()
  if self.state == "waitcon" then
    if name then
      if ver and uitcpVersion and ver ~= uitcpVersion then
        uiWarnT(name, "Server has different from client version(%s)", ver)
        --[[
        if self == mainTunnel and self.currentServer == 1 and 
        updateConf.action ~= "no" then
          runUpdateProcedure(ver)
        end
        ]]
      end
      uiInfoT(self.name, "New tunnel initiated")
      self:initRawSockets(getTunnelConfig(self.name).socket)
      multiTunnelSockets:addTunnel(self)
      self:opened()
    else
      uiWarnT(self.name, "Can't initiate new tunnel")
      self.state = "init"
      self:closed()
    end
  else
    local e = "get newack() in not waitcon state"
    uiFatal(e)
    error(e)
  end
end

function TunnelConnection.connected(c, argt)
  --DBG("Tunnel Connected")
  local self = argt.self
  self.linkCurrent = self.linkN

  if c then
    self.connectCycleStart = nil
    if self.linksPriority and self.linkN ~= 1 then 
      self.linkN = #self.links 
    end
    self.linkLast = self.linkN
    --if self.examineMorePriorityLink and 
    --    self.examineMorePriorityLink == self.linkCurrent then
    --  self.examineMorePriorityLink = nil
    --  c:close()
    --else
    self:setSock(c)
    createRecvDataThread(self)
    if self.state == "init" then
      self:sendCommandNew()
      uiInfoT(self.name, "Init new tunnel")
    else
      self:send('resume("'..self.name..'","'..
        tostring(self.links[self.linkCurrent].currDescr)..
        self:getLinkMode()..'")')
      self:updateKeepTimer(self.keepAttempt * self.keepTmo,
        self.resnack,
      self, self.name, "Timeout exhausted")

      uiInfoT(self.name, "Resume tunnel")
    end
    self:initSession()
    --end
  else
    local h, p
    local lnk = self.links and self.links[self.linkN]
    if lnk then
      if lnk.scriptServers then
        h, p = getHostPort(lnk.scriptServers[self.currentServer])
      else
        h, p = getHostPort(lnk.servers[self.currentServer])
      end
    end
    uiWarnT(self.name, "Can't connect %s:%s(%s)",
    h or "?", p or "?", argt.status or "")
    if self.linkLast == self.linkN then
      self.connectErr = self.connectErr + 1
      --DBG("connectErr", self.connectErr)
      local prevServ
      if self.connectErr >= self.connectAttempt then
        uiWarnT(self.name, 
        "Can't restore tunnel, all local TCP connection will be closed")
        self:allSlave('destroy', true)
        self:init()
        -- Try next server
        prevServ = self.currentServer
        self.currentServer = self.currentServer + 1
        if self.currentServer > self.numOfServers then
          self.currentServer = 1
        end
      end
      if prevServ and prevServ ~= self.currentServer then
        -- other server
        self.linkN = #self.links 
        self:connect()
      else
        local delta = self.connectCycleStart
        and self.connectTmo - (os.time() - self.connectCycleStart)
        or self.connectTmo
        self.connectCycleStart = nil
        --if self.examineMorePriorityLink then
        --  self.examineMorePriorityLink = nil
        --else
        if delta > 0 then
          self.connectTimer = addTimer(delta, self.connect, self)
        else
          self:connect()
        end
        --end
      end
    else
      self:connect()
    end
  end

end

local function getIfaceName(port)
  if port then
    return
    (readpopen("/usr/sbin/nsgsh -q system.get-iface-name."..port) or port):
    match"^%s*(.-)%s*$"
  end
end

function TunnelConnection:connect()

  local link = self.links[self.linkCurrent] 
  if link and (not link.restart or #link.restart == 0) then -- TODO obsolete
    if link.restartScript then
      if type(link.restartScript) == "string" then
        link.restartScript = {link.restartScript}
      end
      link.restart = {timeout = link.restartScriptTmo}
      for k, v in ipairs(link.restartScript) do
        link.restart[k] = {script = v}
      end
    end
  end
  if link and not link.disable and link.restart then
    if not link.lastScriptMark then -- first time
      link.lastScriptMark = os.time()
      if type(link.restart) ~= "table" then
        link.restartScript = nil -- TODO obsolete
        link.restart = nil
      else
        link.scriptCurr = link.restart.primary or 0
        link.scriptNum = #link.restart
        if link.scriptNum == 0 then
          link.restartScript = nil -- TODO obsolete
          link.restart = nil
        end
      end
    else
      local restartTimeout = link.restart[link.scriptCurr] and 
      link.restart[link.scriptCurr].timeout or 
      link.restart.timeout
      if not restartTimeout or
      os.time() - link.lastScriptMark >= restartTimeout then
        link.scriptCurr =   link.scriptCurr + 1
        if link.scriptCurr > link.scriptNum then link.scriptCurr = 1 end
        uiInfoT(self.name, "Start script: %s",
          link.restart[link.scriptCurr].script)
        x(link.restart[link.scriptCurr].script)
        link.lastScriptMark = os.time()
        link.scriptServers = link.restart[link.scriptCurr].servers
        link.currDescr = link.restart[link.scriptCurr].description or 
        link.description
      end
    end
  end

  if not self.connectCycleStart then self.connectCycleStart = os.time() end
  self.linkN = self.linkN + 1; if self.linkN > #self.links then self.linkN = 1 end

  local newlink = self.links[self.linkN]
  if newlink.disable then
    self.connected(false, {self = self, status = "Disabled administratively"})
    return
  end
  local dev  = getIfaceName(newlink.dev)
  if not isInterfaceUp(dev) then
    self.connected(false, {self = self, status = "Netdev "..(dev or "")..
    " is DOWN or not exist"})
    return
  end
  local host, port
  if newlink.scriptServers then
    host, port = getHostPort(newlink.scriptServers[self.currentServer])
  else
    host, port = getHostPort(newlink.servers[self.currentServer])
  end
  if not(host and port) then
    self.connected(false, {self = self, 
        status = ("No hostname for server %d, link %d, script %d"):
        format(self.currentServer, self.linkN, link.scriptCurr)})
    return
  end
  local gw   = newlink.gw
  local defaultRoute = ((dev and " dev "..dev or "")..
    (gw and " via "..gw or "")
  )
  uiInfoT(self.name, "Try connect %s:%s%s", host, port, defaultRoute)
  x(ip..' route del '..host..'/32', true)
  if defaultRoute ~= "" then
    local res  = x(ip..' route add '..host..'/32'..defaultRoute)
    uiInfoT(self.name, "ip route res = %s", tostring(res))
  end
  local c = openTcpSocket(host,
    port,
    TunnelConnection.connected,
    {self = self, keepTmo = self.connectTmo * 0.5})
end

function TunnelConnection:resume(name, link)
  --DBG("resume", name, link)
  if SHUTDOWN then
    self:send('resnack("'..name..'")')
    uiWarnT(name, "Can't resume tunnel, uiTCP in SHUTDOWN mode")
  elseif tunnels[name] then
    tunnels[name]:resumeConnection(self:getSock(), self.an, link)
    self.dead = true
  else
    self:send('resnack("'..name..'")')
    uiWarnT(name, "Can't resume tunnel, no such tunnel")
  end
  --if self.keepTimer then delTimer(self.keepTimer ) end
  self:delKeepTimer()
end

function TunnelConnection:resumeConnection(sock, an, link)
  --self:updateAn(an)
  self:closed()
  self.host, self.port = sock:getpeername()
  uiInfoT(self.name, "Tunnel resumed from %s", self.host)
  self:setSock(sock)
  self.link = link
  self:send('resack("'..self.name..'")')
  createRecvDataThread(self)
  self:opened()
  --self:rej(self.an)
end

function TunnelConnection:resack(name)
  self:delKeepTimer()
  --DBG("resack", name)
  if self.state == "waitcon" then
    uiInfoT(name, "Tunnel resumed")
    self:opened()
    --self:rej(self.an)
  else
    error("get resack() in not waitcon state")
  end
end

function TunnelConnection:resnack(name, err)
  self:delKeepTimer()
  uiWarnT(name, "Tunnel not resumed%s, try reinit tunnel", 
  err and '('..err..')' or "")
  self:reinit()
end

function TunnelConnection:updateAn(newAn)
  local n = self.an
  if (newAn - n) % 256 > (self.sn - n) % 256 then
    return false
  end
  while n ~= newAn do
    self.sended[n] = nil
    n = (n + 1) % 256
  end
  self.an = newAn
  return true
end

function TunnelConnection:__call(recvData)
  hex_dump(self.name, recvData, "receive data from tunnel")
  --socket.sleep(0.1)
  self:recvBytes(#recvData)
  self.recvBuf = self.recvBuf..recvData
  self.lastRecvTime = os.time()
  while true do
    local buf = self.recvBuf
    if #buf < hdr.len then break end
    local len = buf:byte(hdr.lenLow) + buf:byte(hdr.lenHigh) * 256
    if len > #buf - hdr.len then break end -- too many data, wait

    self:recvPack()
    self.recvBuf = self.recvBuf:sub(len + hdr.len + 1) -- remove head
    local dst, src = buf:byte(hdr.dstLC), buf:byte(hdr.srcLC)
    local data = buf:sub(hdr.len + 1, len + hdr.len)
    if not self:updateAn(buf:byte(hdr.rn)) then
      if not self.name then
        -- 
      else
        uiWarnT(self.name, "Bad acknowledge number in receive packet(%d)", 
          buf:byte(hdr.rn))

        self:destroy()
        return false
      end
    end
    local rej = false
    if not (dst == 0 and src == 0) then -- numbered packet
      local bufSn = buf:byte(hdr.sn)
      if bufSn ~= self.rn then
        self.recvBuf = ""
        self:send('rej('..self.rn..')')
        rej = true
      else
        self.rn = (self.rn + 1) % 256
      end
    end
    if not rej then
      if dst == 0 then -- command
        if len > 0 then
          local f, e = loadstring("local m=...;m:"..data)
          if not f then 
            uiErrorT(self.name, e) 
          else
            res, err = countedPcall(f, self)
            if not res then 
              uiErrorT(self.name, err) 
            end
          end
        end
      elseif dst == MultiTunnelLC then
        multiTunnelSockets:receiveData(self, data)
      else -- data
        local slave = self.s[dst]
        if slave then
          slave:send(data)
        else
          uiWarnT(self.name, "Receive data with unexistent LC: %d", dst)
        end
      end
    end
  end -- while
  self:send() -- may be window has opened
end

function TunnelConnection:realSend(data, remoteLC, localLC, sn)
  local sock = self:getSock()
  if not sock then return nil end

  if self.packetRemainder then
    local packet = self.packetRemainder
    self.packetRemainder = nil
    uiDebugT(self.name, "Packet remainder send, len = %d", #packet)
    local n, e, s = sock:send(packet)
    if not n or #packet ~= n then
      uiDebugT(self.name, "Packet remainder send error: %s, len = %d, n = %s s = %d",
        tostring(e), #packet, tostring(n), s)
      local rem = (n or s)
      --if rem > 0 then -- partial send
      self.packetRemainder = packet:sub(rem + 1)
      --end
      return nil
    end
  end

  local len = #data
  local hd = string.char(len % 256, math.floor(len / 256),
  remoteLC, localLC, sn, self.rn)
  self.rr = self.rn
  local packet =  hd..data
  hex_dump(self.name, packet, "Send to tunnel")

  local n, e, s = sock:send(packet)
  if not n or #packet ~= n then
    self:sendErr() -- statistic
    uiDebugT(self.name, "Send error: %s, len = %d, n = %s s = %d", 
      tostring(e), #packet, tostring(n), s)
    local rem = (n or s)
    if rem > 0 then -- partial send
      self.packetRemainder = packet:sub(rem + 1)
      n = #packet
    else
      return nil
    end
  end

  self:sendPack()
  self:sendBytes(n) -- statistic
  return n
end

function TunnelConnection:sendRR()
  if (self.rn - self.rr) % 256 >= self.window/2 then
    self:realSend("", 0, 0, self.sn) -- TODO(if return nil)
  end
end

function TunnelConnection:send(data, remoteLC, localLC)
  if data and not remoteLC then -- send unnumbered packet
    return self:realSend(data, 0, 0, self.sn)
  else -- send numbered packet
    if (self.sn - self.an) % 256 >= self.window or -- no window
    self.state ~= "data" then
      --DBG("no window")
      if data then
        --if self.pending:isEmpty() then self:allSlave('sleep') end
        self:allSlave('sleep')
        self.pending:push{data, remoteLC, localLC}
      end
      self:sendRR()
    elseif data and not self.pending:isEmpty() then
      --DBG"pending"
      self.pending:push{data, remoteLC, localLC}
      self:send()
    else
      --DBG"data"
      if not self.pending:isEmpty() then
        data, remoteLC, localLC = unpack((self.pending:pop())) --(())need only first result
        if self.pending:isEmpty() then self:allSlave('wake') end
      end
      if data then
        local res, n = self:realSend(data, remoteLC, localLC, self.sn)
        if res then
          self.sended[self.sn] = {data, remoteLC, localLC, self.sn}
          self.sn = (self.sn + 1) % 256
          self:send()
        else
          self:allSlave('sleep')
          self.pending:pushFirst{data, remoteLC, localLC}
          if not isTimer(self.sendAfterErrorTimer) then
            self.sendAfterErrorTimer = addTimer(1, self.send, self)
          end
          return true
        end
      end
      self:sendRR()
    end
  end
  return true
end

function TunnelConnection:rej(n)
  uiDebugT(self.name, "Rej n = %d, sn = %d", n, self.sn)
  while n ~= self.sn do
    local pack = self.sended[n]
    if pack then
      self:realSend(unpack(pack))
    else
      uiWarnT(self.name, "Can't resend packet(an = %d)", n)
      self:destroy()
      return false -- do not send the rest
    end
    n = (n + 1) % 256
  end
  self:send()
  return true
end

local  minLCforTunnel = 101
local  maxLCforTunnel = 200
local function getFreeLC(s)
  local last = s.last or minLCforTunnel - 1
  local next = last + 1
  if next > maxLCforTunnel then next = minLCforTunnel end
  while s[next] and next ~= last do
    next = next + 1
    if next > maxLCforTunnel then next = minLCforTunnel end
  end
  if next ~= last then
    s.last = next
    return next
  end
  return nil
end
function TunnelConnection:addSlave(s)
  --DBG("addSlave", self, s)
  local lc
  if s.sockType == 'tcp' then
    if self.state == "init" or self.dead then
      s:destroy()
      return
    end
    lc = getFreeLC(self.s)
  else
    lc = s.localLC
    if  lc%1~=0 or lc <= 0 or lc > 255 or self.s[lc] then lc = nil end
  end
  if not lc then
    uiWarnT(self.name, "Bad logical channel")
    s:destroy(true)
    return
  end
  self.s[lc] = s
  s.localLC = lc
  if s.sockType ~= 'net' then
    s.stat.peerHost, s.stat.peerPort = s:getSock():getpeername()
    s.stat.host, s.stat.port = s:getSock():getsockname()
    s.stat.direction = "from"
  end
  if s:isServer() then
    s.stat.direction = "to"
    uiInfoT(self.name, "Accept local TCP connection(src = %s:%d, dst = %s:%d), lc = %d",
    s.stat.peerHost, s.stat.peerPort, s.stat.host, s.stat.port, lc)
    local id = type(s.id) == 'string' and ("%q"):format(s.id) or "nil"
    local com = 'con('..lc..',"'..s.stat.peerHost..'",'..s.stat.port..','..id..')'
    if not self:send(com, 0, lc) then
      s:destroy()
      self.s[lc] = nil
    end
  end
end

function TunnelConnection:getSlave(lc)
  return self.s[lc]
end

function TunnelConnection.conCreated(c, conf)
  --DBG("conCreated", c)
  local s = connectionCreated(c, conf)
  local lc = 0
  if s then
    lc = s.localLC
    uiInfoT(conf.master.name, "ConAck lc = %d, remote lc = %d", lc, conf.remoteLC)
  else
    uiWarnT(conf.master.name, "Can't connect %s:%d", conf.host, conf.port)
    if conf.srcHost then
      ipAddrDel(conf.srcHost)
    end
  end
  local com = 'conack('..conf.remoteLC..','..lc..')'
  conf.master:send(com, 0, 255)
end

local allLists do
  local function allListsAux(...)
    for i = 1, select('#', ...) do
      local t = select(i, ...)
      if type(t) == 'table' then
        for _, v in ipairs(t) do
          coroutine.yield(v)
        end
      else
        if t ~= nil then coroutine.yield(t) end
      end
    end
  end
  function allLists(...)
    local t = {...}
    return coroutine.wrap(function() allListsAux(unpack(t)) end)
  end
end

function TunnelConnection:con(remoteLC, inSrcAddr, inDstPort, inId)
  --DBG("con:", remoteLC, inSrcAddr, inDstPort)
  uiInfoT(self.name, "Local connection request from tunnel, inSrcAddr = %s,"
    .." inDstPort = %d, inId = %s"
    ,inSrcAddr, inDstPort, tostring(inId))
  local host
  local port = inDstPort
  local srcHost
  if type(conf.localCon) == "table" then -- TODO obsolete parameter
    host = conf.localCon.outDstAddr
    port = conf.localCon.outDstPort
    srcHost = conf.localCon.outSrcAddr
  end
  if inDstPort == updateConf.port then
    host = "127.0.0.1"
    port = inDstPort
    srcHost = nil
  end
  local tcpKeepaliveInterval, tcpKeepaliveRetry
  for v in allLists(self.nat, conf.nat) do
    --DBG(v.inSrcAddr, inDstPort, v.outDstAddr, v.outDstPort, v.outSrcAddr)
    if  (not v.inSrcAddr or v.inSrcAddr == inSrcAddr)
    and (not v.inDstPort or v.inDstPort == inDstPort) 
    and (not v.inId or v.inId == inId) then
      host = v.outDstAddr or host
      port = v.outDstPort or port
      srcHost = v.outSrcAddr or srcHost
      tcpKeepaliveInterval = v.tcpKeepaliveInterval
      tcpKeepaliveRetry = v.tcpKeepaliveRetry
      break
    end
  end
  if not host or not port then
    uiWarnT(self.name, "Can't find parameters for setting local connection")
    local com = 'conack('..remoteLC..','..'0'..')'
    self:send(com, 0, 255)
    return
  end
  uiInfoT(self.name, "Set local connection to %s:%d, src = %s",
  host, port, srcHost or "")
  local arg = {
    host = host,
    port = port,
    srcHost = srcHost,
    remoteLC = remoteLC,
    conClass = LocalConnection,
    master = self,
    remoteLC = remoteLC,
    func = TunnelConnection.conCreated,
    tcpKeepaliveInterval = tcpKeepaliveInterval,
    tcpKeepaliveRetry = tcpKeepaliveRetry,
  }
  createConnection(arg)
end

function TunnelConnection:conack(dst, src)
  if not self.s[dst] then
    uiWarnT(self.name, "Connect ack: No such local connection, lc = %d", dst)
    return
  end
  if src == 0 then
    uiWarnT(self.name, "Not connected lc = %d", dst)
    self.s[dst]:destroy()
    self.s[dst] = nil
  elseif self.s[dst].mustBeClosed then
    uiWarnT(self.name, "Connect ack: Already closed lc = %d", dst)
    local com = 'close('..src..')'
    self:send(com, 0, dst)
    self.s[dst]:destroy()
    self.s[dst] = nil
  else
    uiInfoT(self.name, "Connected lc = %d. remote lc = %d", dst, src)
    self.s[dst]:setRemoteLC(src)
  end
end

function TunnelConnection:closeReq(s)
  local num = false
  for k, v in pairs(self.s) do
    if v == s then num = k; break end
  end
  if num then
    uiInfoT(self.name, "LC=%d closed", s.localLC)
    if s.sockType == 'tcp' then
      local rlc = s.remoteLC
      if rlc and rlc ~= 0 then
      local com = 'close('..s.remoteLC..')'
      self:send(com, 0, s.localLC)
      else
        s.mustBeClosed = true
      end
    end
  end
end

function TunnelConnection:close(n)
  uiInfoT(self.name, "LC=%d closed by request from tunnel", n)
  if self.s[n] then
    if self.s[n].sockType ~= 'tcp' then
      uiErrorT(self.name, "%s socket can't be closed", self.s[n].sockType)
      return
    end
    local com = 'closeack('..self.s[n].remoteLC..')'
    self:send(com, 0, self.s[n].localLC)
    self.s[n]:destroy()
    self.s[n] = nil
  end
end

function TunnelConnection:closeack(n)
  self.s[n] = nil
end

function TunnelConnection:flowControlReq(s, xonxoff)
  local num = false
  for k, v in pairs(self.s) do
    if v == s then num = k; break end
  end
  if num then
    local com = ('%s(%d,%q)'):format(xonxoff, s.remoteLC, xonxoff)
    self:send(com, 0, s.localLC)
    uiDebugT(self.name, "LC=%d %s send", s.localLC, xonxoff)
  end
end

function TunnelConnection:xoff(n)
  uiDebugT(self.name, "LC=%d xoff recv", n)
  local slave = self.s[n]
  if slave then
    slave:sleep()
  end
end

function TunnelConnection:xon(n)
  uiDebugT(self.name, "LC=%d xon recv", n)
  local slave = self.s[n]
  if slave then
    slave:wake()
  end
end

function TunnelConnection:tryReOpen()
  if self.linksPriority then
    local tm = self.linksPriorityTmo
    if (
      tm and os.time() - self.sessionStartTime >= tm or 
      not tm and self.keepSeq >=10
    ) then
      if self.linkCurrent > 1 then
        for n, link in ipairs(self.links) do
          if n >= self.linkCurrent then
            return false
          end
          if not link.disable and isInterfaceUp(link.dev) then
            return true
          end
        end
        return false
      else -- self.linkCurrent == 1
        local primary = self.links[1].restart and tonumber(self.links[1].restart.primary)
        if primary and primary ~= self.links[1].scriptCurr then
          return true
        else
          return false
        end
      end
    end
  end
end

function TunnelConnection:keepalive()
  local keepTmo = self.keepTmo
  local prioClose = self:tryReOpen()
  if prioClose and self.linksPriorityIdleTime then
    keepTmo = self.linksPriorityIdleTime
  end

  local currClock = os.time()
  local delta = currClock - self.lastRecvTime
  if delta >= keepTmo then
    if prioClose then
      uiInfoT(self.name, "Time to examine more priority link")
      self.noRecv = true
      self:closeSock()
      return
    end
    self.keepSeq = self.keepSeq + 1
    self.keepErr = self.keepErr + 1
    local isIntUp = isInterfaceUp(self.links[self.linkCurrent].dev)
    if  self.keepErr > self.keepAttempt or not isIntUp then
      if isIntUp then
        uiWarnT(self.name, "Peer not answered")
      else
        uiWarnT(self.name, "Interface %s(link %d) is DOWN", 
          self.links[self.linkCurrent].dev, self.linkCurrent)
      end
      --self:closed()
      self.noRecv = true
      self:closeSock()
      return
    end
    local com = 'k('..self.keepSeq..')'
    self:send(com)
    delta = 0
  end
  local interval = keepTmo - delta
  local restartInterval = self.linksPriorityTmo - (currClock - self.sessionStartTime)
  if restartInterval > 0 then
    interval = math.min(interval, restartInterval)
  end
  self:updateKeepTimer(interval, self.keepalive, self)
end

function TunnelConnection:initialWatchdog()
  uiWarnT(self.name, "New tunnel connection receive neither new() nor resume() command")
  self:destroy(true)
end

function TunnelConnection:watchdog()
  local silence = os.time() - self.lastRecvTime
  if silence >= self.watchdogTmo then
    uiWarnT(self.name, "Can't restore tunnel, all local connection will be closed")
    self:destroy(true)
  else
    self:updateKeepTimer(self.watchdogTmo - silence, self.watchdog, self)
  end
end

function TunnelConnection:watchdogKeepalive()
  local silence = os.time() - self.lastRecvTime
  if silence >= self.keepAttempt * self.keepTmo then
    uiInfoT(self.name, "No keepalive recived for %d sec.",
    self.keepAttempt * self.keepTmo)
    self:closed()
  else
    self:updateKeepTimer(self.keepAttempt * self.keepTmo - silence,
    self.watchdogKeepalive, self)
  end
end

function TunnelConnection:k(n)
  if not SHUTDOWN then
    local com = 'a('..tostring(n)..')'
    self:send(com)
  end
end

function TunnelConnection:a(n)
  if n == self.keepSeq then self.keepErr = 0 end
end

function TunnelConnection:closed()
  uiInfoT(self.name or "<NONE>", "Tunnel closed")
  self.noRecv = nil
  if self:stateChanged() == "dead" then
    self.state = "init"
  else
    self.state = "disc"
    self:stateChanged"suspended"
  end
  self:delKeepTimer()
  if self.pipe then self.pipe:closeSock() end
  self:closeSock()
  self:setSock(nil)
  self.recvBuf = ""
  self:allSlave('sleep')
  if not self:isServer() then
    if self.wasDataState then
      self:connect()
    else
      self.connected(false, {self = self})
    end
  else -- server
    if self.watchdogTmo then
      self:updateKeepTimer(self.watchdogTmo, self.watchdog, self)
    end
  end
  self.wasDataState = nil
end

function TunnelConnection:startDataPhase()
  self.wasDataState = true
  self.state = "data"
  self:stateChanged("normal")
  self:allSlave('wake')
  self:rej(self.an)
  if not self:isServer() then
    self.connectErr = 0
    self.sessionStartTime = os.time()
    self:updateKeepTimer(math.min(self.keepTmo, self.linksPriorityTmo), self.keepalive, self)
    local csqr = self.links[self.linkCurrent].csqRequest
    if csqr and csqr.ttydev then
      self.pipe = createPipe(self, "/usr/sbin/at+csq", 
      csqr.ttydev, csqr.interval or 60)
    end
    if self.ssl.expiredTime then    
      local com = ('S{cert=%d}'):format(self.ssl.expiredTime)
      self:send(com)
    end
  else -- server
    self:updateKeepTimer(self.keepAttempt * self.keepTmo,
    self.watchdogKeepalive, self)
  end
end

function TunnelConnection.doneSslHandshake(self, res, err)
  local self = self
  self:delKeepTimer()
  if res then
    uiInfoT(self.name, "SSL handshake successfully done")
    createRecvDataThread(self)
    self:startDataPhase()
  else
    uiInfoT(self.name, "Can't do ssl handshake: %s", err)
    --self:closeSock()--???
    self:closed()
  end
end

local getCertTime do
  local month2num = {
    jan = 1,
    feb = 2,
    mar = 3,
    apr = 4,
    may = 5,
    jun = 6,
    jul = 7,
    aug = 8,
    sep = 9,
    oct = 10,
    nov = 11,
    dec = 12,
  }
  function getCertTime(certFile)
    if type(certFile) ~= 'string' then return nil end
    local str = readpopen("openssl x509 -noout -dates -in "..certFile.." 2>&1")
    uiDebug(tostring(str))
    if not str then return nil end
    local M, D, h, m, s, Y = str:match
    "notAfter%s*=%s*(%a%a%a)%a*%s+(%d%d?)%s+(%d%d):(%d%d):(%d%d)%s+(%d%d%d%d)"
    if Y then
      return os.time{year = Y, month = month2num[M:lower()], 
      day = D, hour = h, min = m, sec = s}
    else
      return nil
    end
  end
end

function TunnelConnection:opened()
  --DBG("Tunnel opened", self)
  if not self.ssl.disable then
    self.ssl.expiredTime = getCertTime(self.ssl.certificate)
    if self.ssl.expiredTime then
      uiInfoT(self.name, "Certificate %s expires in %d days", 
        self.ssl.certificate,
        (1/(60*60*24)) * (self.ssl.expiredTime - os.time()))
    end
    uiInfoT(self.name, "Start SSL handshake")
    local res, err
    local sock = self:getSock()
    --DBG("sock", sock)
    res, sock, err = pcall(ssl.wrap, sock, self.ssl)
    if res and sock then
      self:setSock(sock)
      --DBG("SSL sock", sock, self.keepAttempt * self.keepTmo)
      startSslHandshake(self, TunnelConnection.doneSslHandshake)
      self:updateKeepTimer(self.keepAttempt * self.keepTmo,
        self.doneSslHandshake,
      self, nil, "Timeout exhausted")
    else
      uiErrorT(self.name, "Can't do ssl wrap: %s", sock or err)
      self:closeSock()
    end
    return
  end

  self:startDataPhase()
end

function TunnelConnection:destroy()
  self.noRecv = true
  if not self:isServer() then
    --self:closeSock()
    self:allSlave('destroy', true)
    self:init()
    self:closed()
    return
  end
  self.dead = true
  self:closeSock()
  if self.pipe then self.pipe:closeSock() end
  self:delKeepTimer()
  delTimer(self.connectTimer )
  self:allSlave('destroy', true)
  for _, sock in ipairs(self.listenerSocks) do
    sock:close()
  end
  if tunnels[self.name] == self then
    self:stateChanged"dead"
    tunnels[self.name] = nil
  else
    uiDebugT(self.name, "Attempt to destroy destroied tunnel")
  end
end

function TunnelConnection:allSlave(func, ...)
  for k, v in pairs(self.s) do
    if type(k) == "number" then v[func](v, ...) end
  end
  if self.multiTunnelSocket then
    if func == 'sleep' then
      self.transmitReady = nil
    elseif func == 'wake' then
      self.transmitReady = true
      multiTunnelSockets:tunnelWake(self)
    elseif func == 'destroy' and ({...})[1] then
      multiTunnelSockets:removeTunnel(self)
    end
  end
end

local function setIfAddress(devName, ifAddress)
  if ifAddress and ifAddress.prefix then
    local peer, broadcast, anycast = '', '', ''
    if ifAddress.peer then
      peer = "peer "..ifAddress.peer
    end
    if ifAddress.broadcast then
      broadcast = "broadcast "..ifAddress.broadcast
    end
    if ifAddress.anycast then
      anycast = "anycast "..ifAddress.anycast
    end
    x(("%s addr add dev %s %s %s %s %s"):format(ip, devName, 
    ifAddress.prefix, peer, broadcast, anycast))
  end
end

local function setIfAddresses(devName, ifAddress)
  if ifAddress then
    setIfAddress(devName, ifAddress)
    for _, v in ipairs(ifAddress) do
      setIfAddress(devName, v)
    end
  end
end

local sockType2socketMethod = {
  raw = 'raw',
  udp = 'udp',
  net = 'tun',
}

local function createRawSocket(self, lc, conf, class)
  if conf.disable or self:getSlave(lc) then
    return
  end
  local sockType = conf.type
  local logstr, arg
  local sa, pa, sp, pp
  if sockType == 'raw' then
    arg = conf.raw.protocol
    sa = conf.raw.socketAddress
    pa = conf.raw.peerAddress
    logstr = string.format("raw socket %s, peer = %s, proto = %s, LC = %s",
      tostring(sa), tostring(pa), tostring(arg), tostring(lc))
  elseif sockType == 'net' then
    arg = (conf.net and conf.net.ifName) or
    ("%s.%s"):format(self.name:sub(1, 10), tostring(lc))
    logstr = string.format("net socket %s, LC = %s", tostring(arg), tostring(lc))
  elseif sockType == 'udp' then
    sa = conf.udp.socketAddress
    pa = conf.udp.peerAddress
    sp = conf.udp.socketPort
    pp = conf.udp.peerPort
    logstr = string.format("udp socket %s:%d, peer = %s:%d, LC = %s",
      tostring(sa), sp, tostring(pa), pp, tostring(lc))
  else
    uiErrorT(self.name, "Bad type of socket: %s", sockType)
    return
  end
  local sock, err = socket[sockType2socketMethod[sockType]](arg)
  if not sock then
    uiWarnT(self.name, "Can't open %s, err = %s", logstr, err or "")
    return
  end

  local res, srcHost = true
  if sockType == 'net' then
    if conf.net then
      setIfAddresses(arg, conf.net.ifAddress)
    end
    x(ip.." link set dev "..arg.." up")
  elseif sockType == 'raw' then 
    if sa ~= '*' then
      ipAddrAdd(sa)
      srcHost = sa
    end
    res, err = sock:setsockname(sa)
    if res then
      if conf.raw.peerAddress == '*' then
        res, err = nil, "only connected object supported"
      else
        res, err = sock:setpeername(pa)
      end
      if res then
        res, err = sock:setoption('ip-hdrincl', conf.raw.ipHdrIncl == true)
      end
    end
  else -- udp
    if sa ~= '*' then
      ipAddrAdd(sa)
      srcHost = sa
    end
    res, err = sock:setsockname(sa, sp)
    if res then
      if pa == '*' then
        res, err = nil, "only connected object supported"
      else
        res, err = sock:setpeername(pa, pp)
      end
    end
  end

  if not res then
    uiErrorT(self.name, "Can't open %s, err = %s", logstr, err or "")
    return
  end
  -- OK
  local arg = { sock = sock,
    master = self,
    sockType = sockType,
    ipHdrIncl = conf.ipHdrIncl,
    localLC = lc,
    remoteLC = lc,
    srcHost = srcHost,
    qdisc = conf.qdisc,
  }
  local con = (class or LocalConnection)(arg) 
  createRecvDataThread(con)
  uiInfoT(self.name, "Open %s", logstr)
end

TunnelConnection.createRawSocket = createRawSocket

function TunnelConnection:initRawSockets(conf)
  -- if true then return end --TODO
  for k, v in pairs(self.s) do -- destroy all pvc
    if type(v) == 'table' and v.pvc then
      uiInfoT(self.name, "Destroy %s socket %d", v.sockType, v.localLC)
      v:destroy(true)
      v.alreadyClosed = true
      self.s[k] = nil
    end
  end
  if type(conf) ~= 'table' then return end
  socket.sleep(1) -- TODO 
  for k, v in pairs(conf) do
    if type(k) == 'number' and k%1==0 and k > 0 and k < minLCforTunnel then
      self:createRawSocket(k, v)
    end
  end
end

-- stats stuff
function TunnelConnection:stateChanged(state)
  --DBG("stateChanged", state, self.name)
  local stat 
  if not tunnels[self.name] or tunnels[self.name] == self then -- don't serve dead tunnels
    stat = stats.tun[self.name]
  if not stat then
    addTunnelStatsEntry(self.name, "server")
    stat = stats.tun[self.name]
  end
  else
    uiDebugT(self.name, "Try stateChanged(%s) on inactive tunnel", tostring(state))
  end
  if not state then
    return stat and stat.currState or "dead"
  end
  if stat and stat.currState ~= state then
    --DBG("stateChanged", stat.currState, "->", state)
    local time = os.time()
    local delta = time - stat.currStateTime
    stat.state[stat.currState] = stat.state[stat.currState] or {}
    stat.state[stat.currState].duration =
    (stat.state[stat.currState].duration or 0) + delta
    stat.state[stat.currState].quantity =
    (stat.state[stat.currState].quantity or 0) + 1
    if stat.currState == "normal" and stat.link then
      stat.links[stat.link] = stat.links[stat.link] or {}
      stat.links[stat.link].quantity =
      (stat.links[stat.link].quantity or 0) + 1
      stat.links[stat.link].duration =
      (stat.links[stat.link].duration or 0) + delta
    end
    stat.link = self.links and
    tostring(self.links[self.linkCurrent].currDescr)..self:getLinkMode() or
    self.link
    stat.peer = state == "normal" and
    ( self.links and self.links[self.linkCurrent].host or --???
    self.host) or nil
    stat.currState = state
    stat.currStateTime = time
    stat.peerData.csq = nil
  end
end

function TunnelConnection:isReady()
  return  self:stateChanged() == "normal" and self.transmitReady and
  (self.sn - self.an) % 256
end

local function incrementStat(name, key, n)
  local stat = stats.tun[name]
  if stat then
    stat[key] = stat[key] + (n or 1)
    local i = getTimeInterval(stat)
    stat[i][key] = stat[i][key] + (n or 1)
  end
end
function TunnelConnection:sendBytes(n)
  incrementStat(self.name, "sendBytes", n)
end
function TunnelConnection:recvBytes(n)
  incrementStat(self.name, "recvBytes", n)
end
function TunnelConnection:sendPack(n)
  incrementStat(self.name, "sendPack", n)
end
function TunnelConnection:recvPack(n)
  incrementStat(self.name, "recvPack", n)
end
function TunnelConnection:sendErr(n)
  incrementStat(self.name, "sendErr", n)
end

function TunnelConnection:S(data)
  local stat = stats.tun[self.name]
  if type(data) == 'number' then
    stat.peerData.csq = data
  elseif type(data) == 'table' then
    for k, v in pairs(data) do
      stat.peerData[k] = v
    end
  end
end

local certificateFormat = "-----BEGIN CERTIFICATE-----\n%s\n-----END CERTIFICATE-----\n"
local rsaKeyFormat = "-----BEGIN RSA PRIVATE KEY-----\n%s\n-----END RSA PRIVATE KEY-----\n"
local certificatePattern = "(%-%-%-%-%-BEGIN CERTIFICATE%-%-%-%-%-%s+)(.-)(%s+%-%-%-%-%-END CERTIFICATE%-%-%-%-%-)"
local rsaKeyPattern = "(%-%-%-%-%-BEGIN RSA PRIVATE KEY%-%-%-%-%-%s+)(.-)(%s+%-%-%-%-%-END RSA PRIVATE KEY%-%-%-%-%-)"
local function updateSecret(fn, data, pat, fmt)
  if fn and data then
    local f = io.open(fn, 'r')
    if f then
      local oldData = f:read'*a'
      f:close()
      f = io.open(fn, 'w')
      if f then
        local newData, num = oldData:gsub(pat, "%1"..data.."%3", 1)
        if num ~= 1 then
          newData = fmt:format(data)
        end
        f:write(newData)
        f:close()
        return true
      end
    end
  end
end
local function addSecret(fn, data)
  if fn and data then
    local f = io.open(fn, 'r')
    if f then
      local oldData = f:read'*a'
      f:close()
      f = io.open(fn, 'w')
      if f then
        f:write(data..oldData)
        f:close()
        return true
      end
    end
  end
end
function TunnelConnection:secret(cnf)
  if type(cnf) == 'table' then
    uiInfoT(self.name, "Update secrets")
    if not updateSecret(self.ssl.certificate, cnf.c, 
      certificatePattern, certificateFormat
    ) then 
      uiWarnT(self.name, "Can't update certificate")
    end
    if not updateSecret(self.ssl.key, cnf.k, 
      rsaKeyPattern, rsaKeyFormat 
    ) then 
      uiWarnT(self.name, "Can't update key")
    end
    if not cnf.r 
    or not addSecret(self.ssl.cafile, certificateFormat:format(cnf.r)) then 
      uiWarnT(self.name, "Can't add CA certificate")
    end
    x'savecfg'
  end
end
--------------------------------------------------------------------------------
class.LocalConnection(Connection)

function LocalConnection:__init(initial)
  --DBG"LocalConnection __init"
  self.Connection:__init(initial)
  self.sockType = initial.sockType or 'tcp'
  self.master = initial.master
  if self.sockType == 'tcp' then
    self.id = initial.id
    local sock = self:getSock()
    if sock then
      if initial.tcpKeepaliveInterval then
        sock:setoption('keepalive', true)
        if initial.tcpKeepaliveInterval >= 0 then
          local res, err = pcall(sock.setoption, sock, 
          'keepintvl', initial.tcpKeepaliveInterval)
          if not res then
            uiErrorT(self.master.name, socket._VERSION..
            " don't support option `keepintvl'")
          end
          local res, err = pcall(sock.setoption, sock, 
          'keepidle', initial.tcpKeepaliveInterval)
          if not res then
            uiErrorT(self.master.name, socket._VERSION..
            " don't support option `keepidle'")
          end
          if  initial.tcpKeepaliveRetry then
            res, err = pcall(sock.setoption, sock,
            'keepcnt', initial.tcpKeepaliveRetry)
            if not res then
              uiErrorT(self.master.name, socket._VERSION..
              " don't support option `keepcnt'")
            end
          end
        end
      end
    else
      uiErrorT(self.master.name, "No socket for local connection")
    end
  else
    self.pvc = true
    if self.sockType == 'raw' and not initial.ipHdrIncl then 
      self.ipstrip = true   
    end
  end
  self.localLC = initial.localLC or 0
  self.remoteLC = initial.remoteLC or 0
  self.srcHost = initial.srcHost
  self.pending = Fifo()
  self.outfifo = Fifo()
  self.stat = {}
  self.master:addSlave(self)
  self.window = self.master.window or WINDOW_SIZE
end

function LocalConnection:__call(recvData)
  if not recvData then
    if (not self.pending:isEmpty() and
    self.remoteLC ~= 0 and  self.localLC ~= 0) then
      self.master:send(self.pending:pop(), self.remoteLC, self.localLC)
      self()
    end
  else -- recvData
    --if dbg or true then hex_dump(self.master.name, recvData, "receive data from local") end
    if self.ipstrip then
      recvData = recvData:sub(21)
    end
    if self.remoteLC ~= 0 and self.localLC ~= 0 and self.pending:isEmpty() then
      self.master:send(recvData, self.remoteLC, self.localLC)
    else
      self.pending:push(recvData)
      self()
    end
  end
end

function LocalConnection:send(data)
  if dbg then hex_dump(self.master.name, data, "send to local") end
  local sock = self:getSock()
  local destroyAfterSend
  if sock then
    local num = 0
    if not self.outfifo:isEmpty() then
      if data then
        num = self.outfifo:push(data)
        --uiDebugT(self.master.name, "LC=%d fifo %d(%d)", self.localLC, num, self.window * 8)
        if num > self.window * 16 then
          uiErrorT(self.master.name, "LC=%d out FIFO overflow %d",
            self.localLC , num
          )
          self:closeSock()
          return
        end
        if (not self.xoffFlag and num > self.window) or
        num > self.window * 8 then
          self.master:flowControlReq(self, 'xoff')
          self.xoffFlag = true
        end
      end
      data, num = self.outfifo:pop()
      --uiDebugT(self.master.name, "LC=%d pop %d", self.localLC, num)
      if self.xoffFlag and num < self.window / 2 then
        self.master:flowControlReq(self, 'xon')
        self.xoffFlag = false
      end
      if num == 0 and self.mustBeDestroy then
        destroyAfterSend = true
      end
    end
    if data then
      local n, e, s = sock:send(data)
      if not n or #data ~= n then
        if e ~= "timeout" then 
          uiErrorT(self.master.name, "LC=%d send error: %s", self.localLC , e)
          return 
        end
        if not self.wasTimeoutError then
          uiDebugT(self.master.name, "LC=%d send error: %s", self.localLC, e)
          self.wasTimeoutError = true
        end
        local rem = (n or s)
        if rem > 0 then -- partial send
          data = data:sub(rem + 1)
        end
        self.outfifo:pushFirst(data)
        if not isTimer(self.sendAfterErrorTimer) then
          self.sendAfterErrorTimer = addTimer(5, self.send, self)
        end
      else
        self.wasTimeoutError = false
        if num > 0 then
          self:send()
        elseif destroyAfterSend then
          self:destroy(true)
        end
      end
    end
  end
end

--[[
function LocalConnection:XsendX(data)
  if dbg then hex_dump(self.master.name, data, "send to local") end
  local sock = self:getSock()
  if sock then
    local len, s, n, e = #data, 0
    local debug = true
    local t
    repeat
      n, e, s = sock:send(data, s + 1)
      if not n then
        if e ~= "timeout" then 
          uiErrorT(self.master.name, "LC=%d send error: %s", self.localLC , e)
          return 
        else
          if debug then
            uiDebugT(self.master.name, "LC=%d send error: %s (%d)", self.localLC,
            e, s)
            t = socket.gettime()
            debug = false
          end
          socket.sleep(0.1)
        end -- TODO
      end
    until n or s >= len
    if t then
      uiDebug("%f", socket.gettime()-t)
      t = nil
    end
  end
end
--]]

function LocalConnection:setRemoteLC(lc)
  self.remoteLC = lc
  self()             -- send pended data
end

function LocalConnection:destroy(all)
  if not self.pvc or all then
    if not self.outfifo:isEmpty() then
      self.mustBeDestroy = true
      return
    end
    self:closeSock()
  end
  if self.srcHost then
    ipAddrDel(self.srcHost)
  end
end

function LocalConnection:closed()
  --DBG("closed", self)
  self:destroy(true)
  self.master:closeReq(self)
end

function LocalConnection:sleep()
  stopRecv(self:getThread())
end

function LocalConnection:wake()
  startRecv(self:getThread())
end
--------------------------------------------------------------------------------
local function err2table(err)
  return string.format("{ERR = %q}\n", err)
end

local controlCommands = {}

function controlCommands.conf()
  return serialize(conf)
end

function controlCommands.ver()
  local out = {
    version = uitcpVersion
  }
  return serialize(out)
end

function controlCommands.shutdown(v)
  SHUTDOWN = v
  local mes = SHUTDOWN and 'uiTCP in shutdown mode now' or
  'uiTCP in normal mode now'
  uiInfo(mes)
  return '{"'..mes..'"}'
end

function controlCommands.certs()
  local out = {}
  for k, tun in pairs(tunnels) do
    if tun.ssl and tun.ssl.certificate then
      out[tun.ssl.certificate] = tun.ssl.expiredTime
    end
  end
  for k, v in pairs(stats.tun) do
    out['client '..k] = v.peerData and v.peerData.cert
  end
  return serialize(out)
end

function controlCommands.stat(tun)
  if tun and not stats.tun[tun] then
    return err2table("Tunnel not configured: "..tostring(tun))
  end
  local out = {}
  local function copy(tun)
    local t = {day={}}
    local st = stats.tun[tun]
    getTimeInterval(st) -- to update time interval
    local cti = st.currentTimeInterval
    for k,v in pairs(st) do
      local tk = type(k)
      if tk == "string" then
        t[k] = v
      elseif tk == "number" then
        for fn, fv in pairs(v) do
          t.day[fn] = (t.day[fn] or 0) + fv
        end
        if k == cti or k == cti-1 then
          t[k] = v
        end
      end
    end
    out.tun[tun] = t
  end
  out.startTime = stats.startTime
  out.time = os.time()
  out.tun = {}
  if tun then
    if stats.tun[tun] then
      copy(tun)
    end
  else
    for tun, _ in pairs(stats.tun) do
      copy(tun)
    end
  end
  return serialize(out)
end

function controlCommands.con(tun)
  local out = {}
  local function setCons(k, v)
    out[k] = {}
    out[k].currState = stats.tun[k].currState
    for n, loc in pairs(v.s) do
      if type(loc) == "table" then
        out[k][n] = loc.stat
      end
    end
  end
  if tun then
    if tunnels[tun] then
      setCons(tun, tunnels[tun])
    end
  else
    for k,v in pairs(tunnels) do
      setCons(k, v)
    end
  end
  return serialize(out)
end

function controlCommands.tun()
  local out
  out = {}
  for k,_ in pairs(stats.tun) do
    table.insert(out, k)
  end
  return serialize(out)
end

function controlCommands.restartUitcp(cnf)
  uiDebug("Get new config\n%s", serialize(cnf, 0))
  if  not scheme.confCmp(uitcpScheme.sys, cnf.sys or {}, conf.sys or {}) then
    return err2table("To change sys parameters you must restart uitcp")
  else
    uiDebug("The same config, do nothing")
  end
  return '{}'
end

function controlCommands.restartUitcpSys(cnf)
  uiDebug("Get new config\n%s", serialize(cnf, 0))
  if  not scheme.confCmp(uitcpScheme.sys, cnf or {}, conf.sys or {}) then
    return err2table("To change sys parameters you must restart uitcp")
  else
    uiDebug("The same config, do nothing")
  end
  return '{}'
end

function controlCommands.restartMode(mode)
  uiDebug("Get new mode config\n%s", serialize(mode, 0))
  if not scheme.confCmp(uitcpScheme.mode, mode, conf.mode) then
    return err2table("To change mode you must restart uitcp")
  else
    uiDebug("The same config, do nothing")
  end
  return '{}'
end

local function restartClientTun(clientTun, cnf)
  local tunName = clientTun.name
  uiDebugT(tunName, "Get new config\n%s", serialize(cnf, 0))
  local currCnf
  local currSch
  if clientTun == mainTunnel then
    currCnf = conf.tunnel
    if not arg.n then
      currSch = uitcpScheme.tunnel
    end
  else
    currCnf = conf.tunnels[tunName]
    if not arg.n then
      currSch = uitcpScheme.tunnels._NAME
    end
  end
  local ll = cnf.localListener or {}
  if clientTun == mainTunnel and updateConf.action ~= "no" then
    ll[#ll+1] = {host = '*', port = updateConf.port}
  end
  if not currSch then
    currSch = {__index = function()return currSch end}
    setmetatable(currSch, currSch)
  end
  if scheme.confCmp(currSch, cnf, currCnf) then
    uiDebugT(tunName, "The same config, do nothing")
    return '{}'
  end
  addTunnelStatsEntry(tunName, "client", cnf)
  if clientTun then
    if cnf.name and tunName ~= cnf.name then
      return err2table("To change tunnel name you must restart uitcp")
    end
    cnf.name = tunName
    if cnf.disable then
      clientTun:destroy()
    else
      clientTun:initTunnelConfig(cnf)
      if not scheme.confCmp(currSch.links, cnf.links, currCnf.links) 
      or not scheme.confCmp(currSch.servers, cnf.servers, currCnf.servers)
      or not scheme.confCmp(currSch.host, cnf.host, currCnf.host) -- OBSOLETE +1
      or not scheme.confCmp(currSch.port, cnf.port, currCnf.port) then
        clientTun:initTunnelLinks(cnf.links or {})
      end
      clientTun.nat = cnf.nat
      if not scheme.confCmp(currSch.ssl, cnf.ssl, currCnf.ssl) then
        clientTun.ssl = createSslParamTable(cnf.ssl, tunName)
      end

      if not scheme.confCmp(currSch.localListener, ll, 
      currCnf.localListener) then
        createLocalListener(clientTun, ll)
      end
      if not scheme.confCmp(currSch.socket, cnf.socket, currCnf.socket) then
        clientTun:initRawSockets(cnf.socket)
      end
      --[[    
      if clientTun.pipe then clientTun.pipe:closeSock() end -- TODO
      local csqr = clientTun.links[clientTun.linkCurrent].csqRequest
      if csqr and csqr.ttydev then
        clientTun.pipe = createPipe(clientTun, uitcpBinPath.."/tools/at+csq", 
                               csqr.ttydev, csqr.interval or 60)
      end
      --]]
    end
  end
  if clientTun == mainTunnel then
    conf.tunnel = cnf
  else
    conf.tunnels[tunName] = cnf
  end
  return '{}'
end

function controlCommands.restartMainTun(cnf)
  return restartClientTun(mainTunnel, cnf)
end

local function createClientTun(name, tunConf)
  if tunConf.disable then
    uiInfoT(tostring(name), "Tunnel disabled administratively")
  else
    if mainTunnel and name == mainTunnel.name then
      uiErrorT(name, "Can't create tunnel, name already exist")
    else
      tunConf.name = name
      local tun = TunnelConnection(tunConf)
      createLocalListener(tun, tunConf.localListener)
    end
  end
end

function controlCommands.restartTunnels(cnf)
  if conf.tunnels then
    for cl in pairs(conf.tunnels) do
      if not cnf[cl] then
        conf.tunnels[cl] = nil
        stats.tun[cl] = nil
        if tunnels[cl] then
          tunnels[cl]:destroy()
        end
        uiInfoT(cl, "Tunnel deleted administratively")
      end
    end
  end
  return '{}'
end

function controlCommands.restartClientTun(name, cnf)
  if type(name) == 'string' then
    local tun = tunnels[name]
    if tun then
      return restartClientTun(tun, cnf)
    else
      conf.tunnels[name] = cnf
      createClientTun(name, cnf)
    end
  end
  return '{}'
end

function controlCommands.restartClients(cnf)
  if conf.clients then
    for cl in pairs(conf.clients) do
      if not cnf[cl] then
        conf.clients[cl] = nil
        stats.tun[cl] = nil
        if tunnels[cl] then
          tunnels[cl]:destroy()
        end
        uiInfoT(cl, "Tunnel deleted administratively")
      end
    end
  end
  return '{}'
end

function controlCommands.restartTun(name, cnf)
  if type(name) == 'string' then
    uiDebugT(name, "Get new config\n%s", serialize(cnf, 0))
    if not conf.clients then conf.clients = {} end
    if scheme.confCmp(uitcpScheme.clients._NAME, cnf, conf.clients[name]) then
      uiDebugT(name, "The same config, do nothing")
      return '{}'
    end
    local tun = tunnels[name]
    addTunnelStatsEntry(name, "server", cnf)
    if tun then
      tun.nat = cnf.nat
      if not scheme.confCmp(uitcpScheme.clients._NAME.ssl,
        cnf.ssl, 
        conf.clients[name].ssl) then
        tun.ssl = createSslParamTable(cnf.ssl, name)
      end
      if not scheme.confCmp(uitcpScheme.clients._NAME.localListener,
        cnf.localListener, 
        conf.clients[name].localListener) then
        createLocalListener(tun, cnf.localListener)
      end
      if not scheme.confCmp(uitcpScheme.clients._NAME.socket,
        cnf.socket, 
        conf.clients[name].socket) then
        tun:initRawSockets(cnf.socket)
      end
    end
    conf.clients[name] = cnf
  end
  return '{}'
end

function controlCommands.restartNAT(cnfnat)
  uiDebug("Get new NAT config\n%s", serialize(cnfnat, 0))
  if not scheme.confCmp(uitcpScheme.nat, cnfnat, conf.nat) then
    conf.nat = cnfnat
  else
    uiDebug("The same config, do nothing")
  end
  return '{}'
end

function controlCommands.restartSSL(cnfssl)
  uiDebug("Get new SSL config\n%s", serialize(cnfssl, 0))
  if not scheme.confCmp(uitcpScheme.ssl, cnfssl, conf.ssl) then
    conf.ssl = cnfssl
    for name, tun in pairs(tunnels) do
      if tun.ssl then
        local tunssl
        if tun:isServer() then
          tunssl = conf.clients[name].ssl
        else
          tunssl = getTunnelConfig(self.name).ssl
        end
        tun.ssl = createSslParamTable(tunssl, name)
      end
    end
  else
    uiDebug("The same config, do nothing")
  end
  return '{}'
end

function controlCommands.restartLog(cnflog)
  uiDebug("Get new log config\n%s", serialize(cnflog, 0))
  if not scheme.confCmp(uitcpScheme.log, cnflog, conf.log) then
    conf.log = cnflog
    reOpenLogFile()
  else
    uiDebug("The same config, do nothing")
  end
  return '{}'
end

local createTunnelListener

function controlCommands.restartTL(cnftl)
  uiDebug("Get new tunnelListener config\n%s", serialize(cnftl, 0))
  cnftl = cnftl or {}
  if not scheme.confCmp(uitcpScheme.tunnelListener, cnftl, 
    conf.tunnelListener or {}) then
    conf.tunnelListener = cnftl
    createTunnelListener()
  else
    uiDebug("The same config, do nothing")
  end
  return '{}'
end

function controlCommands.restartMultiTunnelSocket(cnf)
  uiDebug("Get new MultiTunnelSocket config\n%s", serialize(cnf, 0))
  if scheme.confCmp(uitcpScheme.multiTunnelSocket, cnf, conf.multiTunnelSocket) then
    uiDebugT(name, "The same config, do nothing")
    return '{}'
  end
  conf.multiTunnelSocket = cnf
  multiTunnelSockets:initRawSockets(conf.multiTunnelSocket)  
  return '{}'
end

function controlCommands.restart(name, cnf)
  return '{}'
end

local function getPemFile(name, ssl, fn)
  local fullFileName = (fn:sub(1, 1) == '/' or fn:sub(1, 2) == '~/') and fn or (
    (ssl.path or (conf.ssl and conf.ssl.path) or '/etc/uitcp/certs')..
    '/'..name..'/cert-client/'..fn 
  )
  local f, err = io.open(fullFileName, 'r')
  if f then
    local str = f:read'*a'
    f:close()
    return str, fullFileName
  else
    return nil, err
  end
end

local function getSecret(name, ssl, fn, pat)
  local pem, err = getPemFile(name, ssl, fn)
  local _
  if pem then
    _, pem, _ = pem:match(pat)
    if not pem then
      err = err..": not pem format file" 
    end
  end
  return pem, err
end

function controlCommands.secrets(name, cnf)
  local dflt
  if type(cnf) ~= 'table' then
    cnf = {
      c = "client.pem",
      k = "clientkey.pem",
      r = "root.pem"
    }
    dflt = true
  end
  local tun = tunnels[name]
  if not tun then
    return err2table("Tunnel is dead: "..tostring(name))
  end
  if stats.tun[name].currState ~= 'normal' then
    return err2table("Tunnel is not alive: "..tostring(name))
  end
  local cert, key, root
  local ssl = tun.ssl or {}
  local com = "secret{"
  if cnf.c then
    cert, err = getSecret(name, ssl, cnf.c, certificatePattern)
    if cert then
      com = com..("c=%q,"):format(cert)
    elseif not dflt then
      return err2table("Can't find client certificate. ".. err)
    end
  end
  if cnf.k then
    key, err = getSecret(name, ssl, cnf.k, rsaKeyPattern)
    if key then
      com = com..("k=%q,"):format(key)
    elseif not dflt then
      return err2table("Can't find client private key certificate. ".. err)
    end
  end
  if cnf.r then
    root, err = getSecret(name, ssl, cnf.r, certificatePattern)
    if root then
      com = com..("r=%q,"):format(root)
    elseif not dflt then
      return err2table("Can't find CA certificate. ".. err)
    end
  end
  com = com.."}"
  print(com)
  tun:send(com)
  return err2table"OK"
end

class.ControlConnection(Connection)

function ControlConnection:__init(initial)
  --DBG"ControlConnection __init"
  self.Connection:__init(initial)
end

function ControlConnection:__call(recvData)
  --hex_dump(nil, recvData, "receive data from control")
  local srcHost, srcPort = self:getSock():getpeername()
  --DBG(srcHost, srcPort)
  if srcHost ~= "127.0.0.1" then self:send(err2table("Access denied")) end
  local f, err = loadstring("local m=...;return m."..recvData)
  if f then
    f, err = countedPcall(f, controlCommands)
  end
  if type(err) ~= "string" then
    f, err = nil, "Result not a string"
  end
  if f then
    self:send(err.."\n")
  else
    uiError("Error executing control command %q:\n%s", recvData, err)
    self:send(err2table(err))
  end
end

function ControlConnection:send(data)
  --hex_dump(nil, data, "send to control")
  repeat
    local len, s, n, e = #data, 0
    n, e, s = self:getSock():send(data, s + 1)
    if not n then
      uiError("Control socket send error: %s", e)
      if e ~= "timeout" then return end -- TODO
    end
  until n or s >= len
  self:closeSock()
end

function ControlConnection:closed()
  --DBG("ControlConnection:closed", self)
end
--------------------------------------------------------------------------------
--local 
function createPipe(master, path, ...)
  local pipe = socket.pipe(path, ...)
  local arg = { sock = pipe,
    master = master,
  }
  local con = PipeConnection(arg)
  createRecvDataThread(con)
  return con
end



class.PipeConnection(Connection)

function PipeConnection:__init(initial)
  self.Connection:__init(initial)
  self.master = initial.master
end

function PipeConnection:__call(recvData)
  --hex_dump(nil, recvData, "receive data from pipe")
  if self.master then
    local csq = tonumber(recvData)
    if csq then
      local com = ('S(%d)'):format(csq)
      self.master:send(com)
    end
  end
end

function PipeConnection:closed()
end

--------------------------------------------------------------------------------
class.MultiTunnelConnection(Connection)

function MultiTunnelConnection:__init(initial)
  self.Connection:__init(initial)
  self.sockType = initial.sockType
  self.master = initial.master
  self.localLC = initial.localLC
  self.pending = Fifo()
  self.stat = {}
  self.tuns = {}
  self.prevTun = 1
  self.discipline = initial.qdisc
  if not self.discipline or not self[self.discipline] then
    self.discipline = "roundRobin"
  end
  self.master:addMTSocket(self)
end

function MultiTunnelConnection:__call(recvData)
  if not recvData then
    if not self.pending:isEmpty() then
      self:receive((self.pending:pop()))
      self()
    end
  else -- recvData
    --hex_dump(nil, recvData, "receive data from MTC")
    if self.pending:isEmpty() then
      self:receive(recvData)
    else
      self.pending:push(recvData)
      self()
    end
  end
end

function MultiTunnelConnection:roundRobin()
  local tuns = self.tuns
  local maxTun = #tuns
  if maxTun == 0 then return nil end
  if self.prevTun > maxTun then self.prevTun = maxTun end
  local nextTun = self.prevTun
  repeat 
    nextTun = nextTun + 1
    if nextTun > maxTun then nextTun = 1 end
    if tuns[nextTun]:isReady() then
      self.prevTun = nextTun
      return nextTun
    end
  until nextTun == self.prevTun
  return nil
end

function MultiTunnelConnection:fullWindow()
  local tuns = self.tuns
  local maxTun = #tuns
  if maxTun == 0 then return nil end
  if self.prevTun > maxTun then self.prevTun = maxTun end
  local nextTun = self.prevTun
  repeat 
    if tuns[nextTun]:isReady() then
      self.prevTun = nextTun
      return nextTun
    end
    nextTun = nextTun + 1
    if nextTun > maxTun then nextTun = 1 end
  until nextTun == self.prevTun
  return nil
end

--local debugK = 0
function MultiTunnelConnection:weighted()
  --debugK = debugK + 1
  local nextTun, minQueue = nil, 256
  for k, v in ipairs(self.tuns) do 
    local w = v:isReady()
    --    if debugK % 20 == 0 then  
    --      uiDebugT(v.name, "QL = %d", w or 256) 
    --    end
    local kw = w and w/v.window
    if kw and kw < minQueue then
      nextTun, minQueue = k, kw
    end
  end
  return nextTun
end

function MultiTunnelConnection:receive(data)
  local tun = self[self.discipline](self)
  if tun then
    self.tuns[tun]:send(data, MultiTunnelLC, MultiTunnelLC)
  else
    self:sleep()
  end
end 

function MultiTunnelConnection:send(data)
  --hex_dump(nil, data, "send to MTC")
  local sock = self:getSock()
  if sock then
    repeat
      local len, s, n, e = #data, 0
      n, e, s = sock:send(data, s + 1)
      if not n then
        if e ~= "timeout" then 
          uiErrorT(self.master.name, "LC=%d send error: %s", self.localLC , e)
          return 
        end -- TODO
      end
    until n or s >= len
  end
end

function MultiTunnelConnection:destroy()
  self:closeSock()
end

function MultiTunnelConnection:closed()
  self:destroy()
  self.master:removeMTSocket(self)
end

function MultiTunnelConnection:sleep()
  stopRecv(self:getThread())
end

function MultiTunnelConnection:wake()
  startRecv(self:getThread())
end


multiTunnelSockets = {
  name = 'multiTunnelSockets',
  s = {}, -- local socket objects

}

function multiTunnelSockets:getSlave(lc)
  return self.s[lc]
end

multiTunnelSockets.createRawSocket = createRawSocket

function multiTunnelSockets:addTunnel(tun)
  local msName = tun.multiTunnelSocket -- multiSocket name
  if msName then
    local ms = self.s[msName]
    if ms then
      for k, v in ipairs(ms.tuns) do
        if v == tun then
          uiErrorT(tun.name, "Tunnel alredy connected to multitunnel socket %s",
            tostring(msName))
          return
        end
      end
      table.insert(ms.tuns, tun)
    end
  end
end

function multiTunnelSockets:removeTunnel(tun)
  local msName = tun.multiTunnelSocket -- multiSocket name
  if msName then
    local ms = self.s[msName]
    if ms then
      for k, v in ipairs(ms.tuns) do
        if v == tun then
          table.remove(ms.tuns, k)
          return
        end
      end
      uiWarnT(tun.name, "Tunnel not connected to multitunnel socket %s",
        tostring(msName))
    end
  end
end

function multiTunnelSockets:tunnelWake(tun)
  local msName = tun.multiTunnelSocket -- multiSocket name
  if msName then
    local ms = self.s[msName]
    if ms then
      ms:wake()
    end
  end
end

function multiTunnelSockets:receiveData(tun, data)
  local msName = tun.multiTunnelSocket -- multiSocket name
  if msName then
    local ms = self.s[msName]
    if ms then
      ms:send(data)
    end
  end
end

function multiTunnelSockets:addMTSocket(mts)
  local lc = mts.localLC
  if lc and (not self.s[lc]) then
    self.s[lc] = mts
  end
end

function multiTunnelSockets:removeMTSocket(mts)
  for k, v in pairs(self.s) do
    if v == mts then
      self.s[k] = nil
      return
    end
  end
end

function multiTunnelSockets:initRawSockets(conf)
  -- if true then return end --TODO
  for k, v in pairs(self.s) do -- destroy all pvc
    if type(v) == 'table' and v.pvc then
      uiInfo("Destroy multitunnel socket %d", tostring(k))
      v:destroy(true)
      v.alreadyClosed = true
      self.s[k] = nil
    end
  end
  if type(conf) ~= 'table' then return end
  socket.sleep(1) -- TODO 
  for k, v in pairs(conf) do
    --if type(k) == 'number' and k%1==0 and k > 0 and k < minLCforTunnel then
      self:createRawSocket(k, v, MultiTunnelConnection)
    --end
  end
end



---------------Listen stuff-----------------------------------------------------
local function accept(s)
  s:settimeout(0)
  local c, status = s:accept()
  if status == "timeout" then
    coroutine.yield(s)
  end
  return c, status
end

local function listen(conf)
  while true do
    local c, status = accept(conf.serverSock)
    if c then
      conf.sock = c
      local con = conf.conClass(conf)
      --DBG("New connection", con, c)
      createRecvDataThread(con)
    elseif status == "closed" then
      --DBG("listener closed", conf.host, conf.port)
      break
    end
  end
end

local function createListenThread(conf)
  if type(conf.host) ~= 'string' or type(conf.port) ~= 'number' then
    uiError("createListenThread: Bad host or port parameter type %s:%s", 
      tostring(conf.host), tostring(conf.port))
    return nil
  end
  local s, e = socket.bind(conf.host, conf.port)
  if not s then
    uiError("Socket bind to %s:%d error: %s", conf.host, conf.port, e)
    return nil
  end
  local i, p = s:getsockname()
  assert(i, p)
  uiInfo("Waiting connection from talker on %s:%d", i, p)
  conf.serverSock = s
  local co = coroutine.create( 
    function()
      listen(conf)
  end)
  table.insert(threads, co)
  return s
end
-----------------Receive stuf---------------------------------------------------
local function recvData(con, sock)
  --print("recvData, con =", con)
  --print("sock =", sock)
  local n = 0
  local blkSize = con.conType == 'tunnel' and RECEIVE_BLOCK_SIZE + hdr.len or
  RECEIVE_BLOCK_SIZE
  while true do
    sock:settimeout(0)
    local data, status, partial = sock:receive(blkSize)
    if sock ~= con:getSock() then
      sock:close()
      return
    end
    data = data or partial
    if data and #data > 0 then
      --DBG("recvData", con)
      if not con.noRecv then con(data) end
      if con.dead then DBG"dead" return end -- when resume tunnel connection
      n = n + 1
    end
    if status then
      if status == "timeout" or
      status == "wantread" or
      status == "wantwrite" then
        coroutine.yield(sock, status == "wantwrite")
        n = 0
      elseif status == "closed" then break 
      else
        uiInfo("Socket receive ERROR: %s", status)
        sock:close()
        break
      end
    elseif n >= MAX_RECEIVE_BLOCKS then
      coroutine.yield(sock)
      n = 0
    end
  end
  if not con.alreadyClosed then con:closed() end
end

function createRecvDataThread(con)
  local sock = con:getSock()
  --DBG("createRecvDataThread", con, sock)
  local co = coroutine.create(
    function()
      recvData(con, sock)
  end)
  table.insert(threads, co)
  con:setThread(co)
end

--------------------------------------------------------------------------------
local function waitTcpSocket(host, port, func, argt)
  local endtime
  local m = socket.tcp()
  if argt.srcHost then
    ipAddrAdd(argt.srcHost)
    if not m:bind(argt.srcHost, 0) then
      ipAddrDel(argt.srcHost)
      m:close()
      argt.status = "Can't bind"
      func(nil, argt)
      return
    end
  end
  if argt.keepTmo then
    endtime = os.time() + argt.keepTmo
  end
  while true do
    m:settimeout(0)
    local c, status = m:connect(host, port)
    --DBG("waitTcpSocket", c, status)
    if c then
      func(m, argt)
      return
    elseif (status == "timeout" or
    status == "Operation already in progress") and -- TODO (bad criterium)
    (not endtime or os.time() < endtime) then
      coroutine.yield(m, true)
    else
      m:close()
      argt.status = status
      func(nil, argt)
      return
    end
  end
end

function openTcpSocket(host, port, func, argt)
  local co = coroutine.create(
    function()
      waitTcpSocket(host, port, func, argt)
  end)
  table.insert(threads, co)
end

function connectionCreated(m, conf)
  if not m then return nil end
  conf.sock = m
  local con = conf.conClass(conf)
  createRecvDataThread(con)
  return con
end

function createConnection(conf)
  --DBG"createConnection"
  openTcpSocket(conf.host, conf.port, conf.func, conf)
end

--------------------------------------------------------------------------------

local function waitSslHandShake(self, func, sock)
  while true do
    sock:settimeout(0)
    local res, err = sock:dohandshake()
    --DBG("waitSslHandShake", res, err)
    if sock ~= self:getSock() then
      sock:close()
      --DBG("waitSslHandShake closed(no sock)")
      return
    end
    if res then
      func(self, res, err)
      return
    elseif err == "wantread" or err == "wantwrite" or err == "timeout" then
      coroutine.yield(sock,  err == "wantwrite")
    else
      sock:close()
      func(self, res, err)
      return
    end
  end
end

function startSslHandshake(self, func)
  local sock = self:getSock()
  local co = coroutine.create(
    function()
      waitSslHandShake(self, func, sock)
  end)
  table.insert(threads, co)
end
--------------------------------------------------------------------------------
local function dispatch()
  local i = 1
  local connections = {}
  local sndConnections = {}
  local status, res, snd
  while true do
    if threads[i] == nil then
      if threads[1] == nil then break end -- TODO ???
      i = 1
      connections = {}
      sndConnections = {}
    end
    --uiDebug("dispatch thr = %s", tostring(threads[i]))
    if sleepThr[threads[i]] then
      res = nil
    else
      status, res, snd = coroutine.resume(threads[i])
    end
    if not status then
      error("coroutine failed\n"..res)
    end
    if not res then
      table.remove(threads, i)
    else -- timeout
      i = i + 1
      if snd then
        sndConnections[#sndConnections + 1] = res
      else
        connections[#connections + 1] = res
      end
      if #connections + #sndConnections == #threads then
        --DBG("start select", #threads)
        local recvt, sendt, err = socket.select(connections, sndConnections,
          nextTmo()
        )
        --if dbg then print(os.time()) end
        timerHandler()
        --[[
          if dbg then
            for _, v in ipairs(recvt) do print(v) end
            for _, v in ipairs(sendt) do print(v) end
          end
          --]]
      end
    end
  end -- end while
end

--------------------------------------------------------------------------------
function createLocalListener(master, localListenerTable)
  if not master or not localListenerTable then return end
  if master.listenerSocks then
    -- Close all local listener
    for _, sock in ipairs(master.listenerSocks) do
      sock:close()
    end
    master.listenerSocks = {}
  end
  for _, v in ipairs(localListenerTable) do
    local arg = {
      host = v.host or '*',
      port = v.port or 0,
      conClass = LocalConnection,
      master = master,
      server = true,
      tcpKeepaliveInterval = v.tcpKeepaliveInterval,
      tcpKeepaliveRetry = v.tcpKeepaliveRetry,
      id = v.id,
    }
    table.insert(master.listenerSocks, createListenThread(arg))
  end
end

do
  local tunnelListenerTable = {}
  function createTunnelListener()
    for _, sock in ipairs(tunnelListenerTable) do
      sock:close()
    end
    local tl = conf.tunnelListener or {}
    local dfltHost = arg.n and '*' or uitcpScheme.tunnelListener.host._D
    local dfltPort = arg.n and 50005 or uitcpScheme.tunnelListener.port._D
    local arg = {
      host = tl.host or dfltHost,
      port = tl.port or dfltPort,
      conClass = TunnelConnection,
      server = true,
    }
    table.insert(tunnelListenerTable, createListenThread(arg))
    for _, v in ipairs(tl) do
      local arg = {
        host = v.host or dfltHost,
        port = v.port or dfltPort,
        conClass = TunnelConnection,
        server = true,
      }
      table.insert(tunnelListenerTable, createListenThread(arg))
    end
  end
end

local function createControlListener()
  local arg = {
    host = "*",  --conf.tunnelListener.host,
    port = 50006, --conf.tunnelListener.port,
    conClass = ControlConnection,
  }
  createListenThread(arg)
end

--------------------------------------------------------------------------------
local function updateServerTaskControl(taskId)
  for mes in task.receive, 0 do
    uiInfo("Update service message: %s", mes)
  end
  if task.isrunning(taskId) then
    addTimer(10, updateServerTaskControl, taskId)
  else
    uiError("Update service failed")
  end
end


--------------------------------------------------------------------------------
local function startUitcp()
  uiDebug("Current configuration:\n%s", serialize(conf, 0))
  socket.sleep(3) -- to avoide "bind: address already in use" error after restart
  initTunnelStats()

  if conf.multiTunnelSocket then
    multiTunnelSockets:initRawSockets(conf.multiTunnelSocket)
  end

  if conf.mode == 'client' 
  or conf.createTunnel -- TODO obsolete
  then
    uiInfo"Client mode"
    conf.mode = 'client' 
    if conf.tunnel.disable then
      uiInfoT(tostring(conf.tunnel.name), "Tunnel disabled administratively")
    else
      mainTunnel = TunnelConnection(conf.tunnel)
    end
    local ll = conf.tunnel.localListener or {}
    if updateConf.action ~= "no" then
      ll[#ll+1] = {host = '*', port = updateConf.port}
    end
    createLocalListener(mainTunnel, ll)
    if conf.tunnels and type(conf.tunnels) == 'table' then
      for name, tunConf in pairs(conf.tunnels) do 
        createClientTun(name, tunConf)
      end
    end
  else
    uiInfo"Server mode"
    conf.mode = 'server'
    createTunnelListener()
    local pluginFolder = conf.sys and conf.sys.pluginFolder or '/etc/uitcp/plugins'
    local ustf = uitcpBinPath..'tools/update-server.lua'
    local f, err = io.open(ustf)
    if f then
      f:close()
      local updateServerTask = task.create(ustf,
        { task.id(), pluginFolder,})
      if updateServerTask <= 0 then
        uiError("Can't create update task: %d", updateServerTask)
      else
        uiInfo('Update task is created: %d', updateServerTask)
        addTimer(10, updateServerTaskControl, updateServerTask)
      end
    else
      uiError("Error opening update-server: %s", err)
    end

    if socket.poll then
      socket.select = socket.poll
      uiInfo"Use poll instead select function"
      local mfdn = 8192
      socket.setrlimit(nil, mfdn, mfdn)
      uiInfo("Set maximum file descriptor number to %d", mfdn)
    end
  end
  createControlListener()
  uiInfo"Start main loop"
  dispatch()
end

do -- set signal
  local res, signal = pcall(require, 'signal')
  if res then
    local function sigHandler()
      error"Get interrupt signal"
    end
    signal.signal("INT", sigHandler)
  end
end


--require"profiler"
--profiler.start("/var/log/uiprof")
local res, err = pcall(startUitcp)
--profiler.stop()
if not res then
  uiFatal(err)
  if mainTunnel then
    mainTunnel:destroy()
  end
  for name, tun in pairs(tunnels) do
    tun:destroy()
  end
  uiInfo("============== uiTCP-%s exitted ==============", uitcpVersion or 'work')
  error(err)
end

--------------------------------------------------------------------------------
--[[
iptables -A OUTPUT -t nat -p tcp -d 192.168.10.1 --dport 8080 -j DNAT --to-destination 82.138.15.92
--]]
